/*
 * This file is an example of how to embed web-server functionality
 * into existing application.
 * Compilation line:
 * cc example.c shttpd.c -DEMBEDDED -DMT
 */

#ifdef _WIN32
#include <winsock.h>
#define	snprintf		_snprintf
#define	sleep(x)		Sleep((x) * 1000)	
#pragma comment(lib,"ws2_32")
#else
#include <sys/types.h>
#include <sys/select.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <glib/gthread.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include "shttpd.h"

#include "bmpx/dbus.h"
#include "bmpx/metadata.h"

static const gchar *http_ok = "HTTP/1.1 200 OK\r\n";

static const gchar *ctype_xml     = "Content-Type: text/xml\r\n\r\n";
static const gchar *ctype_xhtml   = "Content-Type: application/xhtml+xml\r\n\r\n";
static const gchar *ctype_png     = "Content-Type: image/x-png\r\n\r\n";
static const gchar *ctype_js      = "Content-Type: application/x-javascript\r\n\r\n";

static GMutex *lock;

static gchar* list_tags[] = {
    "artist", 
    "album", 
    "title", 
    "time"
};

enum {
    V_ARTIST,
    V_ALBUM,
    V_TITLE,
    V_TIME
};

G_LOCK_DEFINE(BMPX_DBUS);

static  DBusGConnection *bus;
static  DBusGProxy	*bmpx;

#define DBUS_TYPE_G_STRING_VALUE_HASHTABLE (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))

static GHashTable *action_method_map = NULL;

static gchar* b_names[] = {
	  "prev",
	  "next",
	  "stop",
	  "play",
	  "pause"
};

static gchar* b_methods[] = {
	  "PlayPrev",
	  "PlayNext",
	  "PlayStop",
	  "PlayCurrent",
	  "PlayPause"
};

static int
send_header(struct shttpd_callback_arg *arg)
{
    GError *error = NULL;
    gchar  *mem;
    gsize   size;

    shttpd_push (arg->connection, (const void*)http_ok, strlen(http_ok)); 
    shttpd_push (arg->connection, (const void*)ctype_png, strlen(ctype_png)); 

    g_file_get_contents (DATA_DIR "/bmpx-web/header.png", &mem, &size, &error);

    shttpd_push (arg->connection, mem, size); 

    g_free (mem);
    return (0);
}
 
static int
send_js(struct shttpd_callback_arg *arg)
{
    GError *error = NULL;
    gchar  *mem;
    gsize   size;

    shttpd_push (arg->connection, (const void*)http_ok, strlen(http_ok)); 
    shttpd_push (arg->connection, (const void*)ctype_js, strlen(ctype_js)); 

    g_file_get_contents (DATA_DIR "/bmpx-web/functions.js", &mem, &size, &error);

    shttpd_push (arg->connection, (const void*)mem, size); 

    g_free (mem);
    return (0);
}
 
static int
xmlreply(struct shttpd_callback_arg *arg)
{
  enum {
    XRN_ENTRIES,
    XRN_ENTRY,
    XRN_ARTIST,
    XRN_ALBUM,
    XRN_TITLE,
    XRN_TIME,
    
    XRN_N_NODES
  };

  static xmlChar *xrn_names[] = {
    BAD_CAST "entries",
    BAD_CAST "entry",
    BAD_CAST "artist",
    BAD_CAST "album",
    BAD_CAST "title",
    BAD_CAST "time"
  };

  xmlDocPtr doc;
  xmlNodePtr xrnNode[XRN_N_NODES];
  const gchar *b;
  gchar       *c_str;
  gint	 c, minutes, seconds;

  g_mutex_lock (lock);

  b = shttpd_get_var (arg->connection, "play"); 
  if (b)
    {
      GError *error = NULL;
      gint    track;

      track = strtol (b, NULL, 10);

      G_LOCK (BMPX_DBUS);
      if (!dbus_g_proxy_call (bmpx, "PlayTrack", &error,
			      G_TYPE_INT,
			      track,
			      G_TYPE_INVALID,
			      G_TYPE_INVALID))
      {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete PlayTrack: %s", error->message);
	return (0);
      }
      G_UNLOCK (BMPX_DBUS);
    }

  b = shttpd_get_var (arg->connection, "req"); 
  if ((b) && (!strcmp(b,"tracklist")))
    {
      GError	 *error = NULL;
      gchar	**list;
      xmlChar    *mem;
      gint	  n = 0, size;

      G_LOCK (BMPX_DBUS);
      if (!dbus_g_proxy_call (bmpx, "GetTracklist", &error,
			      G_TYPE_INVALID,
			      G_TYPE_STRV, 
			      &list, 
			      G_TYPE_INVALID))
      {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete GetTracklist: %s", error->message);
	return (0);
      }
      G_UNLOCK (BMPX_DBUS);

      G_LOCK (BMPX_DBUS);
      if (!dbus_g_proxy_call (bmpx, "GetCurrentTrack", &error,
			      G_TYPE_INVALID,
			      G_TYPE_INT, 
			      &c, 
			      G_TYPE_INVALID))
      {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete GetCurrentTrack: %s", error->message);
	return (0);
      }
      G_UNLOCK (BMPX_DBUS);

      doc = xmlNewDoc (BAD_CAST "1.0");
      xrnNode[XRN_ENTRIES] = xmlNewNode (NULL, xrn_names[XRN_ENTRIES]); 
      xmlDocSetRootElement (doc, xrnNode[XRN_ENTRIES]);
      c_str = g_strdup_printf ("%d", c); 
      xmlSetProp (xrnNode[XRN_ENTRIES], BAD_CAST "current", BAD_CAST c_str); 
      g_free (c_str);

      while (list[n]) 
	{
	    GValue      *value;
	    GHashTable	*metadata;
	    const gchar *value_str;
	    gchar	*value_aux;
	    gint	 value_int;

	    xrnNode[XRN_ENTRY] = xmlNewNode (NULL, xrn_names[XRN_ENTRY]); 
	    xmlAddChild (xrnNode[XRN_ENTRIES], xrnNode[XRN_ENTRY]);

	    G_LOCK (BMPX_DBUS);
	    if (!dbus_g_proxy_call (bmpx, "GetMetadataForUri", &error,
			      G_TYPE_STRING,
			      list[n],
			      G_TYPE_INVALID,
			      DBUS_TYPE_G_STRING_VALUE_HASHTABLE, 
			      &metadata, 
			      G_TYPE_INVALID))
	      {
		g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete GetMetadataForUri: %s", error->message);
		continue;
	      }
	    G_UNLOCK (BMPX_DBUS);

	    value = g_hash_table_lookup (metadata, list_tags[V_ARTIST]); 
	    value_str = g_value_get_string (value);
	    xrnNode[XRN_ARTIST] = xmlNewNode (NULL, BAD_CAST list_tags[V_ARTIST]);
	    xmlAddChild (xrnNode[XRN_ARTIST], xmlNewText (BAD_CAST value_str)); 

	    value = g_hash_table_lookup (metadata, list_tags[V_ALBUM]); 
	    value_str = g_value_get_string (value);
	    xrnNode[XRN_ALBUM] = xmlNewNode (NULL, BAD_CAST list_tags[V_ALBUM]);
	    xmlAddChild (xrnNode[XRN_ALBUM], xmlNewText (BAD_CAST value_str)); 

	    value = g_hash_table_lookup (metadata, list_tags[V_TITLE]); 
	    value_str = g_value_get_string (value);
	    xrnNode[XRN_TITLE] = xmlNewNode (NULL, BAD_CAST list_tags[V_TITLE]);
	    xmlAddChild (xrnNode[XRN_TITLE], xmlNewText (BAD_CAST value_str)); 
 

	    value = g_hash_table_lookup (metadata, list_tags[V_TIME]); 
	    value_int = g_value_get_int (value);
            minutes = value_int / 60;
	    seconds = value_int % 60;
	    value_aux = g_strdup_printf ("%d:%-2.2d", minutes, seconds);
	    xrnNode[XRN_TIME] = xmlNewNode (NULL, BAD_CAST list_tags[V_TIME]);
	    xmlAddChild (xrnNode[XRN_TIME], xmlNewText (BAD_CAST value_aux)); 
	    g_free (value_aux);

	    xmlAddChild (xrnNode[XRN_ENTRY], xrnNode[XRN_ARTIST]);
	    xmlAddChild (xrnNode[XRN_ENTRY], xrnNode[XRN_ALBUM]);
	    xmlAddChild (xrnNode[XRN_ENTRY], xrnNode[XRN_TITLE]);
	    xmlAddChild (xrnNode[XRN_ENTRY], xrnNode[XRN_TIME]);

	    n++;
	}

      xmlKeepBlanksDefault (0);
      xmlThrDefIndentTreeOutput (1);
      xmlDocDumpFormatMemory (doc, &mem, &size, 1);

      shttpd_push (arg->connection, (const void*)http_ok, strlen(http_ok)); 
      shttpd_push (arg->connection, (const void*)ctype_xml, strlen(ctype_xml)); 

      shttpd_push (arg->connection, (const void*)mem, size); 
      g_free (mem);
      xmlFreeDoc (doc);
    }

  g_mutex_unlock (lock);

  return (0);
}
 
static int
index_html(struct shttpd_callback_arg *arg)
{
    enum {
	N_XHTML,  
	N_IMG,
	N_BODY,
	N_HEAD,
	N_META,
	N_TR,
	N_TD,
	N_TITLE,
	N_SCRIPT,
	N_DIV_LIST,

	N_NODES
    };

    xmlDocPtr	doc;

    xmlNodePtr	nodes[N_NODES];

    xmlNsPtr	ns;
    xmlDtdPtr	dtd;

    xmlChar    *mem;
    gint	size;

    g_mutex_lock (lock);

      doc = xmlNewDoc(BAD_CAST "1.0");

      nodes[N_XHTML] = xmlNewNode (NULL, BAD_CAST "html"); 
      ns = xmlNewNs (nodes[N_XHTML], BAD_CAST "http://www.w3.org/1999/xhtml", NULL);
      xmlSetNs (nodes[N_XHTML], ns);

      dtd = xmlNewDtd (doc, BAD_CAST "html", BAD_CAST "-//W3C//DTD XHTML 1.0 Strict//EN", BAD_CAST "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd");
      xmlDocSetRootElement (doc, (xmlNodePtr)dtd);

      xmlAddNextSibling ((xmlNodePtr)dtd, nodes[N_XHTML]);	

      nodes[N_HEAD] = xmlNewNode (ns, BAD_CAST "head");

      nodes[N_TITLE] = xmlNewNode (ns, BAD_CAST "title");
      xmlAddChild (nodes[N_TITLE], xmlNewText (BAD_CAST "BMPx WEB"));    

      nodes[N_SCRIPT] = xmlNewNode (ns, BAD_CAST "script");
      xmlSetProp (nodes[N_SCRIPT], BAD_CAST "type", BAD_CAST "text/javascript");
      xmlSetProp (nodes[N_SCRIPT], BAD_CAST "src", BAD_CAST "/functions.js");

      nodes[N_BODY] = xmlNewNode (ns, BAD_CAST "body");
      xmlSetProp (nodes[N_BODY], BAD_CAST "style", BAD_CAST "background-color: white;");
      xmlSetProp (nodes[N_BODY], BAD_CAST "onload", BAD_CAST "loadXMLDoc('/xml?req=tracklist', 0);");

      xmlAddChild (nodes[N_XHTML], nodes[N_HEAD]);
      xmlAddChild (nodes[N_XHTML], nodes[N_BODY]);

      xmlAddChild (nodes[N_HEAD], nodes[N_TITLE]);
      xmlAddChild (nodes[N_HEAD], nodes[N_SCRIPT]);

      nodes[N_DIV_LIST] = xmlNewNode (ns, BAD_CAST "div");
      xmlSetProp (nodes[N_DIV_LIST], BAD_CAST "id", BAD_CAST "div-list");

      nodes[N_IMG] = xmlNewNode (ns, BAD_CAST "img");
      xmlSetProp (nodes[N_IMG], BAD_CAST "src", BAD_CAST "/header.png");
      xmlSetProp (nodes[N_IMG], BAD_CAST "alt", BAD_CAST "BMPx WEB");

      xmlAddChild (nodes[N_BODY], nodes[N_IMG]);
      xmlAddChild (nodes[N_BODY], nodes[N_DIV_LIST]);

      xmlKeepBlanksDefault (0);
      xmlThrDefIndentTreeOutput (1);
      xmlThrDefLoadExtDtdDefaultValue (0);	
      xmlDocDumpFormatMemoryEnc (doc, &mem, &size, "UTF-8", 1);

      shttpd_push (arg->connection, (const void*)http_ok, strlen(http_ok)); 
      shttpd_push (arg->connection, (const void*)ctype_xhtml, strlen(ctype_xhtml)); 
      shttpd_push (arg->connection, (const void*)mem, size); 

      g_free (mem);
      xmlFreeDoc (doc);

      g_mutex_unlock (lock);

      return (0);
}

static gpointer
httpd_thread (gpointer data)
{
    gint      n, sock;
    GError   *error = NULL;

    /* Initialize and setup URLs we gonna serve */
    shttpd_init(NULL);

    /* Let pass 'data' variable to callback, user can change it */
    shttpd_register_url("/", &index_html, NULL); 
    shttpd_register_url("/xml", &xmlreply, NULL); 
    shttpd_register_url("/functions.js", &send_js, NULL); 
    shttpd_register_url("/header.png", &send_header, NULL); 

    /* Open two listening sockets */
    sock = shttpd_open_port(9000);

    action_method_map = g_hash_table_new (g_str_hash, g_str_equal);

    for (n = 0; n < 5 ; n++)
	  g_hash_table_insert (action_method_map, g_strdup(b_names[n]), g_strdup(b_methods[n]));

    G_LOCK (BMPX_DBUS);
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!bus)
	  {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Couldn't connect to session bus: %s", error->message);
	      exit (EXIT_FAILURE);
	  }

    bmpx = dbus_g_proxy_new_for_name (bus,
				      BMP_DBUS_SERVICE,
				      BMP_DBUS_PATH_SYSTEMCONTROL,
				      BMP_DBUS_INTERFACE);
    G_UNLOCK (BMPX_DBUS);


    while (1)
      {
#if 0
          FD_ZERO(&rfds);
	  FD_SET(sock, &rfds);

	  shttpd_merge_fds(&rfds, NULL, &sock);

	  rv = select(sock+1, &rfds, NULL, NULL, NULL);
#endif

	  shttpd_poll (sock, 300);
      }

    /* Probably unreached, because we will be killed by signal */
    shttpd_fini();
}

gint
main(gint argc, gchar **argv)
{
	GMainLoop   *mainloop;
	GThread	    *thread;

	/* Initialize GLib */
	g_type_init ();
	g_thread_init (NULL); 
	dbus_g_type_specialized_init ();

	lock = g_mutex_new ();

	mainloop = g_main_loop_new (NULL, FALSE);

	thread = g_thread_create (httpd_thread,
				  NULL,
				  TRUE,
				  NULL);

	g_main_loop_run (mainloop);
	
	return (EXIT_SUCCESS);
}
