/* gtkpathbar.c
 * Copyright (C) 2004  Red Hat, Inc.,  Jonathan Blandford <jrb@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <string.h>

#include <gtk/gtk.h>

#include "bmp_path_bar.h"
#include "marshalers.h"


enum {
  _CLICKED,
  LAST_SIGNAL
};

typedef enum {
  NORMAL_BUTTON,
  ROOT_BUTTON,
} ButtonType;

#define BUTTON_DATA(x) ((ButtonData *)(x))

#define SCROLL_TIMEOUT           150
#define INITIAL_SCROLL_TIMEOUT   300

static guint path_bar_signals [LAST_SIGNAL] = { 0 };

/* Icon size for if we can't get it from the theme */
#define FALLBACK_ICON_SIZE 16

typedef struct _ButtonData ButtonData;

struct _ButtonData
{
  GtkWidget   *button;
  ButtonType   type;
  gchar	      *name;
  GtkWidget *image;
  GtkWidget *label;
  guint ignore_changes : 1;
  guint file_is_hidden : 1;
};
/* This macro is used to check if a button can be used as a fake root.
 * All buttons in front of a fake root are automatically hidden when in a
 * directory below a fake root and replaced with the "<" arrow button.
 */
#define BUTTON_IS_FAKE_ROOT(button) ((button)->type == NORMAL_BUTTON)

G_DEFINE_TYPE (BmpPathBar,
	       bmp_path_bar,
	       GTK_TYPE_CONTAINER);

static void bmp_path_bar_finalize                 (GObject          *object);
static void bmp_path_bar_dispose                  (GObject          *object);
static void bmp_path_bar_size_request             (GtkWidget        *widget,
						   GtkRequisition   *requisition);
static void bmp_path_bar_unmap                    (GtkWidget          *widget);
static void bmp_path_bar_size_allocate            (GtkWidget        *widget,
						   GtkAllocation    *allocation);
static void bmp_path_bar_add                      (GtkContainer     *container,
						   GtkWidget        *widget);
static void bmp_path_bar_remove                   (GtkContainer     *container,
						   GtkWidget        *widget);
static void bmp_path_bar_forall                   (GtkContainer     *container,
						   gboolean          include_internals,
						   GtkCallback       callback,
						   gpointer          callback_data);
static void bmp_path_bar_scroll_up                (GtkWidget        *button,
						   BmpPathBar       *path_bar);
static void bmp_path_bar_scroll_down              (GtkWidget        *button,
						   BmpPathBar       *path_bar);
static void bmp_path_bar_stop_scrolling           (BmpPathBar       *path_bar);
static gboolean bmp_path_bar_slider_button_press  (GtkWidget        *widget,
						   GdkEventButton   *event,
						   BmpPathBar       *path_bar);
static gboolean bmp_path_bar_slider_button_release(GtkWidget        *widget,
						   GdkEventButton   *event,
						   BmpPathBar       *path_bar);
static void bmp_path_bar_grab_notify              (GtkWidget        *widget,
						   gboolean          was_grabbed);
static void bmp_path_bar_state_changed            (GtkWidget        *widget,
						   GtkStateType      previous_state);
static void bmp_path_bar_style_set                (GtkWidget        *widget,
						   GtkStyle         *previous_style);
static void bmp_path_bar_screen_changed           (GtkWidget        *widget,
						   GdkScreen        *previous_screen);
static void bmp_path_bar_check_icon_theme         (BmpPathBar       *path_bar);
static void bmp_path_bar_update_button_appearance (BmpPathBar       *path_bar,
						   ButtonData       *button_data,
						   gboolean          current_dir);

static GtkWidget *
get_slider_button (BmpPathBar  *path_bar,
		   GtkArrowType arrow_type)
{
  GtkWidget *button;

  gtk_widget_push_composite_child ();

  button = gtk_button_new ();
  gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
  gtk_container_add (GTK_CONTAINER (button), gtk_arrow_new (arrow_type, GTK_SHADOW_OUT));
  gtk_container_add (GTK_CONTAINER (path_bar), button);
  gtk_widget_show_all (button);

  gtk_widget_pop_composite_child ();

  return button;
}

static void
bmp_path_bar_init (BmpPathBar *path_bar)
{
  GTK_WIDGET_SET_FLAGS (path_bar, GTK_NO_WINDOW);
  gtk_widget_set_redraw_on_allocate (GTK_WIDGET (path_bar), FALSE);

  path_bar->spacing = 3;
  path_bar->up_slider_button = get_slider_button (path_bar, GTK_ARROW_LEFT);
  path_bar->down_slider_button = get_slider_button (path_bar, GTK_ARROW_RIGHT);
  path_bar->icon_size = FALLBACK_ICON_SIZE;
  
  g_signal_connect (path_bar->up_slider_button, "clicked", G_CALLBACK (bmp_path_bar_scroll_up), path_bar);
  g_signal_connect (path_bar->down_slider_button, "clicked", G_CALLBACK (bmp_path_bar_scroll_down), path_bar);

  g_signal_connect (path_bar->up_slider_button, "button_press_event", G_CALLBACK (bmp_path_bar_slider_button_press), path_bar);
  g_signal_connect (path_bar->up_slider_button, "button_release_event", G_CALLBACK (bmp_path_bar_slider_button_release), path_bar);
  g_signal_connect (path_bar->down_slider_button, "button_press_event", G_CALLBACK (bmp_path_bar_slider_button_press), path_bar);
  g_signal_connect (path_bar->down_slider_button, "button_release_event", G_CALLBACK (bmp_path_bar_slider_button_release), path_bar);
}

static void
bmp_path_bar_class_init (BmpPathBarClass *path_bar_class)
{
  GObjectClass *gobject_class;
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkContainerClass *container_class;

  gobject_class = (GObjectClass *) path_bar_class;
  object_class = (GtkObjectClass *) path_bar_class;
  widget_class = (GtkWidgetClass *) path_bar_class;
  container_class = (GtkContainerClass *) path_bar_class;

  gobject_class->finalize = bmp_path_bar_finalize;
  gobject_class->dispose = bmp_path_bar_dispose;

  widget_class->size_request = bmp_path_bar_size_request;
  widget_class->unmap = bmp_path_bar_unmap;
  widget_class->size_allocate = bmp_path_bar_size_allocate;
  widget_class->style_set = bmp_path_bar_style_set;
  widget_class->screen_changed = bmp_path_bar_screen_changed;
  widget_class->grab_notify = bmp_path_bar_grab_notify;
  widget_class->state_changed = bmp_path_bar_state_changed;

  container_class->add = bmp_path_bar_add;
  container_class->forall = bmp_path_bar_forall;
  container_class->remove = bmp_path_bar_remove;
  /* FIXME: */
  /*  container_class->child_type = bmp_path_bar_child_type;*/

  path_bar_signals [_CLICKED] =
    g_signal_new ("path-clicked",
		  G_OBJECT_CLASS_TYPE (object_class),
		  G_SIGNAL_RUN_FIRST,
		  G_STRUCT_OFFSET (BmpPathBarClass, path_clicked),
		  NULL, NULL,
		  bmpx_VOID__POINTER_BOOLEAN,
		  G_TYPE_NONE, 2,
		  G_TYPE_POINTER,
		  G_TYPE_BOOLEAN);
}


static void
bmp_path_bar_finalize (GObject *object)
{
  BmpPathBar *path_bar;

  path_bar = BMP_PATH_BAR (object);

  bmp_path_bar_stop_scrolling (path_bar);

  g_list_free (path_bar->button_list);

  if (path_bar->root_path)
    g_free (path_bar->root_path);

  if (path_bar->root_icon)
    g_object_unref (path_bar->root_icon);

  G_OBJECT_CLASS (bmp_path_bar_parent_class)->finalize (object);
}

/* Removes the settings signal handler.  It's safe to call multiple times */
static void
remove_settings_signal (BmpPathBar *path_bar,
			GdkScreen  *screen)
{
  if (path_bar->settings_signal_id)
    {
      GtkSettings *settings;

      settings = gtk_settings_get_for_screen (screen);
      g_signal_handler_disconnect (settings,
				   path_bar->settings_signal_id);
      path_bar->settings_signal_id = 0;
    }
}

static void
bmp_path_bar_dispose (GObject *object)
{
  remove_settings_signal (BMP_PATH_BAR (object),
			  gtk_widget_get_screen (GTK_WIDGET (object)));

  G_OBJECT_CLASS (bmp_path_bar_parent_class)->dispose (object);
}

/* Size requisition:
 * 
 * Ideally, our size is determined by another widget, and we are just filling
 * available space.
 */
static void
bmp_path_bar_size_request (GtkWidget      *widget,
			   GtkRequisition *requisition)
{
  ButtonData *button_data;
  BmpPathBar *path_bar;
  GtkRequisition child_requisition;
  GList *list;

  path_bar = BMP_PATH_BAR (widget);

  requisition->width = 0;
  requisition->height = 0;

  for (list = path_bar->button_list; list; list = list->next)
    {
      button_data = BUTTON_DATA (list->data);
      gtk_widget_size_request (button_data->button, &child_requisition);
      requisition->width = MAX (child_requisition.width, requisition->width);
      requisition->height = MAX (child_requisition.height, requisition->height);
    }

  /* Add space for slider, if we have more than one path */
  /* Theoretically, the slider could be bigger than the other button.  But we're
   * not going to worry about that now.
   */
  path_bar->slider_width = MIN(requisition->height * 2 / 3 + 5, requisition->height);
  if (path_bar->button_list && path_bar->button_list->next != NULL)
    requisition->width += (path_bar->spacing + path_bar->slider_width) * 2;

  gtk_widget_size_request (path_bar->up_slider_button, &child_requisition);
  gtk_widget_size_request (path_bar->down_slider_button, &child_requisition);

  requisition->width += GTK_CONTAINER (widget)->border_width * 2;
  requisition->height += GTK_CONTAINER (widget)->border_width * 2;

  widget->requisition = *requisition;
}

static void
bmp_path_bar_update_slider_buttons (BmpPathBar *path_bar)
{
  if (path_bar->button_list)
    {
      GtkWidget *button;

      button = BUTTON_DATA (path_bar->button_list->data)->button;
      if (gtk_widget_get_child_visible (button))
	{
	  bmp_path_bar_stop_scrolling (path_bar);
	  gtk_widget_set_sensitive (path_bar->down_slider_button, FALSE);
	}
      else
	gtk_widget_set_sensitive (path_bar->down_slider_button, TRUE);

      button = BUTTON_DATA (g_list_last (path_bar->button_list)->data)->button;
      if (gtk_widget_get_child_visible (button))
	{
	  bmp_path_bar_stop_scrolling (path_bar);
	  gtk_widget_set_sensitive (path_bar->up_slider_button, FALSE);
	}
      else
	gtk_widget_set_sensitive (path_bar->up_slider_button, TRUE);
    }
}

static void
bmp_path_bar_unmap (GtkWidget *widget)
{
  bmp_path_bar_stop_scrolling (BMP_PATH_BAR (widget));

  GTK_WIDGET_CLASS (bmp_path_bar_parent_class)->unmap (widget);
}

/* This is a tad complicated
 */
static void
bmp_path_bar_size_allocate (GtkWidget     *widget,
			    GtkAllocation *allocation)
{
  GtkWidget *child;
  BmpPathBar *path_bar = BMP_PATH_BAR (widget);
  GtkTextDirection direction;
  GtkAllocation child_allocation;
  GList *list, *first_button;
  gint width;
  gint allocation_width;
  gint border_width;
  gboolean need_sliders = FALSE;
  gint up_slider_offset = 0;
  gint down_slider_offset = 0;

  widget->allocation = *allocation;

  /* No path is set; we don't have to allocate anything. */
  if (path_bar->button_list == NULL)
    return;

  direction = gtk_widget_get_direction (widget);
  border_width = (gint) GTK_CONTAINER (path_bar)->border_width;
  allocation_width = allocation->width - 2 * border_width;

  /* First, we check to see if we need the scrollbars. */
  if (path_bar->fake_root)
    width = path_bar->spacing + path_bar->slider_width;
  else
      width = 0;

  for (list = path_bar->button_list; list; list = list->next)
    {
      child = BUTTON_DATA (list->data)->button;

      width += child->requisition.width + path_bar->spacing;
      if (list == path_bar->fake_root)
	break;
    }

  if (width <= allocation_width)
    {
      if (path_bar->fake_root)
	first_button = path_bar->fake_root;
      else
	first_button = g_list_last (path_bar->button_list);
    }
  else
    {
      gboolean reached_end = FALSE;
      gint slider_space = 2 * (path_bar->spacing + path_bar->slider_width);

      if (path_bar->first_scrolled_button)
	first_button = path_bar->first_scrolled_button;
      else
	first_button = path_bar->button_list;
      need_sliders = TRUE;
      
      /* To see how much space we have, and how many buttons we can display.
       * We start at the first button, count forward until hit the new
       * button, then count backwards.
       */
      /* Count down the path chain towards the end. */
      width = BUTTON_DATA (first_button->data)->button->requisition.width;
      list = first_button->prev;
      while (list && !reached_end)
	{
	  child = BUTTON_DATA (list->data)->button;

	  if (width + child->requisition.width +
	      path_bar->spacing + slider_space > allocation_width)
	    reached_end = TRUE;
	  else if (list == path_bar->fake_root)
	    break;
	  else
	    width += child->requisition.width + path_bar->spacing;

	  list = list->prev;
	}

      /* Finally, we walk up, seeing how many of the previous buttons we can
       * add */
      while (first_button->next && !reached_end)
	{
	  child = BUTTON_DATA (first_button->next->data)->button;

	  if (width + child->requisition.width + path_bar->spacing + slider_space > allocation_width)
	    {
	      reached_end = TRUE;
	    }
	  else
	    {
	      width += child->requisition.width + path_bar->spacing;
	      if (first_button == path_bar->fake_root)
		break;
	      first_button = first_button->next;
	    }
	}
    }

  /* Now, we allocate space to the buttons */
  child_allocation.y = allocation->y + border_width;
  child_allocation.height = MAX (1, (gint) allocation->height - border_width * 2);

  if (direction == GTK_TEXT_DIR_RTL)
    {
      child_allocation.x = allocation->x + allocation->width - border_width;
      if (need_sliders || path_bar->fake_root)
	{
	  child_allocation.x -= (path_bar->spacing + path_bar->slider_width);
	  up_slider_offset = allocation->width - border_width - path_bar->slider_width;
	}
    }
  else
    {
      child_allocation.x = allocation->x + border_width;
      if (need_sliders || path_bar->fake_root)
	{
	  up_slider_offset = border_width;
	  child_allocation.x += (path_bar->spacing + path_bar->slider_width);
	}
    }

  for (list = first_button; list; list = list->prev)
    {
      child = BUTTON_DATA (list->data)->button;

      child_allocation.width = child->requisition.width;
      if (direction == GTK_TEXT_DIR_RTL)
	child_allocation.x -= child_allocation.width;

      /* Check to see if we've don't have any more space to allocate buttons */
      if (need_sliders && direction == GTK_TEXT_DIR_RTL)
	{
	  if (child_allocation.x - path_bar->spacing - path_bar->slider_width < widget->allocation.x + border_width)
	    break;
	}
      else if (need_sliders && direction == GTK_TEXT_DIR_LTR)
	{
	  if (child_allocation.x + child_allocation.width + path_bar->spacing + path_bar->slider_width >
	      widget->allocation.x + border_width + allocation_width)
	    break;
	}

      gtk_widget_set_child_visible (BUTTON_DATA (list->data)->button, TRUE);
      gtk_widget_size_allocate (child, &child_allocation);

      if (direction == GTK_TEXT_DIR_RTL)
	{
	  child_allocation.x -= path_bar->spacing;
	  down_slider_offset = child_allocation.x - widget->allocation.x - path_bar->slider_width;
	  down_slider_offset = border_width;
	}
      else
	{
	  down_slider_offset = child_allocation.x - widget->allocation.x;
	  down_slider_offset = allocation->width - border_width - path_bar->slider_width;
	  child_allocation.x += child_allocation.width + path_bar->spacing;
	}
    }
  /* Now we go hide all the widgets that don't fit */
  while (list)
    {
      gtk_widget_set_child_visible (BUTTON_DATA (list->data)->button, FALSE);
      list = list->prev;
    }
  for (list = first_button->next; list; list = list->next)
    {
      gtk_widget_set_child_visible (BUTTON_DATA (list->data)->button, FALSE);
    }

  if (need_sliders || path_bar->fake_root)
    {
      child_allocation.width = path_bar->slider_width;
      child_allocation.x = up_slider_offset + allocation->x;
      gtk_widget_size_allocate (path_bar->up_slider_button, &child_allocation);

      gtk_widget_set_child_visible (path_bar->up_slider_button, TRUE);
      gtk_widget_show_all (path_bar->up_slider_button);
    }
  else
    gtk_widget_set_child_visible (path_bar->up_slider_button, FALSE);
      
  if (need_sliders)
    {
      child_allocation.width = path_bar->slider_width;
      child_allocation.x = down_slider_offset + allocation->x;
      gtk_widget_size_allocate (path_bar->down_slider_button, &child_allocation);

      gtk_widget_set_child_visible (path_bar->down_slider_button, TRUE);
      gtk_widget_show_all (path_bar->down_slider_button);
      bmp_path_bar_update_slider_buttons (path_bar);
    }
  else
    gtk_widget_set_child_visible (path_bar->down_slider_button, FALSE);
}

static void
bmp_path_bar_style_set (GtkWidget *widget,
			GtkStyle  *previous_style)
{
  if (GTK_WIDGET_CLASS (bmp_path_bar_parent_class)->style_set)
    GTK_WIDGET_CLASS (bmp_path_bar_parent_class)->style_set (widget, previous_style);

  bmp_path_bar_check_icon_theme (BMP_PATH_BAR (widget));
}

static void
bmp_path_bar_screen_changed (GtkWidget *widget,
			     GdkScreen *previous_screen)
{
  if (GTK_WIDGET_CLASS (bmp_path_bar_parent_class)->screen_changed)
    GTK_WIDGET_CLASS (bmp_path_bar_parent_class)->screen_changed (widget, previous_screen);

  /* We might nave a new settings, so we remove the old one */
  if (previous_screen)
    remove_settings_signal (BMP_PATH_BAR (widget), previous_screen);

  bmp_path_bar_check_icon_theme (BMP_PATH_BAR (widget));
}

static void
bmp_path_bar_add (GtkContainer *container,
		  GtkWidget    *widget)
{
  gtk_widget_set_parent (widget, GTK_WIDGET (container));
}

static void
bmp_path_bar_remove_1 (GtkContainer *container,
		       GtkWidget    *widget)
{
  gboolean was_visible = GTK_WIDGET_VISIBLE (widget);
  gtk_widget_unparent (widget);
  if (was_visible)
    gtk_widget_queue_resize (GTK_WIDGET (container));
}

static void
bmp_path_bar_remove (GtkContainer *container,
		     GtkWidget    *widget)
{
  BmpPathBar *path_bar;
  GList *children;

  path_bar = BMP_PATH_BAR (container);

  if (widget == path_bar->up_slider_button)
    {
      bmp_path_bar_remove_1 (container, widget);
      path_bar->up_slider_button = NULL;
      return;
    }

  if (widget == path_bar->down_slider_button)
    {
      bmp_path_bar_remove_1 (container, widget);
      path_bar->down_slider_button = NULL;
      return;
    }

  children = path_bar->button_list;
  while (children)
    {
      if (widget == BUTTON_DATA (children->data)->button)
	{
	  bmp_path_bar_remove_1 (container, widget);
	  path_bar->button_list = g_list_remove_link (path_bar->button_list, children);
	  g_list_free (children);
	  return;
	}
      
      children = children->next;
    }
}

static void
bmp_path_bar_forall (GtkContainer *container,
		     gboolean      include_internals,
		     GtkCallback   callback,
		     gpointer      callback_data)
{
  BmpPathBar *path_bar;
  GList *children;

  g_return_if_fail (callback != NULL);
  path_bar = BMP_PATH_BAR (container);

  children = path_bar->button_list;
  while (children)
    {
      GtkWidget *child;
      child = BUTTON_DATA (children->data)->button;
      children = children->next;

      (* callback) (child, callback_data);
    }

  if (path_bar->up_slider_button)
    (* callback) (path_bar->up_slider_button, callback_data);

  if (path_bar->down_slider_button)
    (* callback) (path_bar->down_slider_button, callback_data);
}

static void
bmp_path_bar_scroll_down (GtkWidget *button, BmpPathBar *path_bar)
{
  GList *list;
  GList *down_button = NULL;
  GList *up_button = NULL;
  gint space_available;
  gint space_needed;
  gint border_width;
  GtkTextDirection direction;

  if (path_bar->ignore_click)
    {
      path_bar->ignore_click = FALSE;
      return;   
    }

  gtk_widget_queue_resize (GTK_WIDGET (path_bar));

  border_width = GTK_CONTAINER (path_bar)->border_width;
  direction = gtk_widget_get_direction (GTK_WIDGET (path_bar));
  
  /* We find the button at the 'down' end that we have to make
   * visible */
  for (list = path_bar->button_list; list; list = list->next)
    {
      if (list->next && gtk_widget_get_child_visible (BUTTON_DATA (list->next->data)->button))
	{
	  down_button = list;
	  break;
	}
    }
  
  /* Find the last visible button on the 'up' end
   */
  for (list = g_list_last (path_bar->button_list); list; list = list->prev)
    {
      if (gtk_widget_get_child_visible (BUTTON_DATA (list->data)->button))
	{
	  up_button = list;
	  break;
	}
    }

  space_needed = BUTTON_DATA (down_button->data)->button->allocation.width + path_bar->spacing;
  if (direction == GTK_TEXT_DIR_RTL)
    space_available = path_bar->down_slider_button->allocation.x - GTK_WIDGET (path_bar)->allocation.x;
  else
    space_available = (GTK_WIDGET (path_bar)->allocation.x + GTK_WIDGET (path_bar)->allocation.width - border_width) -
      (path_bar->down_slider_button->allocation.x + path_bar->down_slider_button->allocation.width);

  /* We have space_available extra space that's not being used.  We
   * need space_needed space to make the button fit.  So we walk down
   * from the end, removing buttons until we get all the space we
   * need. */
  while (space_available < space_needed)
    {
      space_available += BUTTON_DATA (up_button->data)->button->allocation.width + path_bar->spacing;
      up_button = up_button->prev;
      path_bar->first_scrolled_button = up_button;
    }
}

static void
bmp_path_bar_scroll_up (GtkWidget *button, BmpPathBar *path_bar)
{
  GList *list;

  if (path_bar->ignore_click)
    {
      path_bar->ignore_click = FALSE;
      return;   
    }

  gtk_widget_queue_resize (GTK_WIDGET (path_bar));

  for (list = g_list_last (path_bar->button_list); list; list = list->prev)
    {
      if (list->prev && gtk_widget_get_child_visible (BUTTON_DATA (list->prev->data)->button))
	{
	  if (list->prev == path_bar->fake_root)
	    path_bar->fake_root = NULL;
	  path_bar->first_scrolled_button = list;
	  return;
	}
    }
}

static gboolean
bmp_path_bar_scroll_timeout (BmpPathBar *path_bar)
{
  gboolean retval = FALSE;

  GDK_THREADS_ENTER ();

  if (path_bar->timer)
    {
      if (path_bar->scrolling_up)
	bmp_path_bar_scroll_up (path_bar->up_slider_button, path_bar);
      else if (path_bar->scrolling_down)
	bmp_path_bar_scroll_down (path_bar->down_slider_button, path_bar);

      if (path_bar->need_timer) 
	{
	  path_bar->need_timer = FALSE;

	  path_bar->timer = g_timeout_add (SCROLL_TIMEOUT,
					   (GSourceFunc)bmp_path_bar_scroll_timeout,
					   path_bar);
	  
	}
      else
	retval = TRUE;
      
    }

  GDK_THREADS_LEAVE ();

  return retval;
}

static void 
bmp_path_bar_stop_scrolling (BmpPathBar *path_bar)
{
  if (path_bar->timer)
    {
      g_source_remove (path_bar->timer);
      path_bar->timer = 0;
      path_bar->need_timer = FALSE;
    }
}

static gboolean
bmp_path_bar_slider_button_press (GtkWidget      *widget, 
				  GdkEventButton *event,
				  BmpPathBar     *path_bar)
{
  if (event->type != GDK_BUTTON_PRESS || event->button != 1)
    return FALSE;

  path_bar->ignore_click = FALSE;

  if (widget == path_bar->up_slider_button)
    {
      path_bar->scrolling_down = FALSE;
      path_bar->scrolling_up = TRUE;
      bmp_path_bar_scroll_up (path_bar->up_slider_button, path_bar);
    }
  else if (widget == path_bar->down_slider_button) 
    {
      path_bar->scrolling_up = FALSE;
      path_bar->scrolling_down = TRUE;
      bmp_path_bar_scroll_down (path_bar->down_slider_button, path_bar);
    }

  if (!path_bar->timer)
    {
      path_bar->need_timer = TRUE;
      path_bar->timer = g_timeout_add (INITIAL_SCROLL_TIMEOUT,
				       (GSourceFunc)bmp_path_bar_scroll_timeout,
				       path_bar);
    }

  return FALSE;
}

static gboolean
bmp_path_bar_slider_button_release (GtkWidget      *widget, 
				    GdkEventButton *event,
				    BmpPathBar     *path_bar)
{
  if (event->type != GDK_BUTTON_RELEASE)
    return FALSE;

  path_bar->ignore_click = TRUE;
  bmp_path_bar_stop_scrolling (path_bar);

  return FALSE;
}

static void
bmp_path_bar_grab_notify (GtkWidget *widget,
			  gboolean   was_grabbed)
{
  if (!was_grabbed)
    bmp_path_bar_stop_scrolling (BMP_PATH_BAR (widget));
}

static void
bmp_path_bar_state_changed (GtkWidget    *widget,
			    GtkStateType  previous_state)
{
  if (!GTK_WIDGET_IS_SENSITIVE (widget)) 
    bmp_path_bar_stop_scrolling (BMP_PATH_BAR (widget));
}


/* Changes the icons wherever it is needed */
static void
reload_icons (BmpPathBar *path_bar)
{
  GList *list;

  if (path_bar->root_icon)
    {
      g_object_unref (path_bar->root_icon);
      path_bar->root_icon = NULL;
    }

  for (list = path_bar->button_list; list; list = list->next)
    {
      ButtonData *button_data;
      gboolean current_dir;

      button_data = BUTTON_DATA (list->data);
      if (button_data->type != NORMAL_BUTTON)
	{
	  current_dir = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button_data->button));
	  bmp_path_bar_update_button_appearance (path_bar, button_data, current_dir);
	}
    }
  
}

static void
change_icon_theme (BmpPathBar *path_bar)
{
  GtkSettings *settings;
  gint width, height;

  settings = gtk_settings_get_for_screen (gtk_widget_get_screen (GTK_WIDGET (path_bar)));

  if (gtk_icon_size_lookup_for_settings (settings, GTK_ICON_SIZE_MENU, &width, &height))
    path_bar->icon_size = MAX (width, height);
  else
    path_bar->icon_size = FALLBACK_ICON_SIZE;

  reload_icons (path_bar);
}
/* Callback used when a GtkSettings value changes */
static void
settings_notify_cb (GObject    *object,
		    GParamSpec *pspec,
		    BmpPathBar *path_bar)
{
  const char *name;

  name = g_param_spec_get_name (pspec);

  if (! strcmp (name, "gtk-icon-theme-name") ||
      ! strcmp (name, "gtk-icon-sizes"))
    change_icon_theme (path_bar);
}

static void
bmp_path_bar_check_icon_theme (BmpPathBar *path_bar)
{
  GtkSettings *settings;

  if (path_bar->settings_signal_id)
    return;

  settings = gtk_settings_get_for_screen (gtk_widget_get_screen (GTK_WIDGET (path_bar)));
  path_bar->settings_signal_id = g_signal_connect (settings, "notify", G_CALLBACK (settings_notify_cb), path_bar);

  change_icon_theme (path_bar);
}

/* Public functions and their helpers */
static void
bmp_path_bar_clear_buttons (BmpPathBar *path_bar)
{
  while (path_bar->button_list != NULL)
    {
      gtk_container_remove (GTK_CONTAINER (path_bar), BUTTON_DATA (path_bar->button_list->data)->button);
    }
  path_bar->first_scrolled_button = NULL;
  path_bar->fake_root = NULL;
}

static void
button_clicked_cb (GtkWidget *button,
		   gpointer   data)
{
  ButtonData *button_data;
  BmpPathBar *path_bar;
  GList	     *button_list;
  gboolean    child_is_hidden;
  gchar	     *child_path;

  button_data = BUTTON_DATA (data);
  if (button_data->ignore_changes)
    return;

  path_bar = BMP_PATH_BAR (button->parent);

  button_list = g_list_find (path_bar->button_list, button_data);
  g_assert (button_list != NULL);

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);

  if (button_list->prev)
    {
      ButtonData *child_data;

      child_data = BUTTON_DATA (button_list->prev->data);
      child_path = child_data->name;
      child_is_hidden = child_data->file_is_hidden;
    }
  else
    {
      child_path = NULL;
      child_is_hidden = FALSE;
    }

  g_signal_emit (path_bar, path_bar_signals [_CLICKED], 0,
		 button_data->name, child_path, child_is_hidden);
}

static GdkPixbuf *
get_button_image (BmpPathBar *path_bar,
		  ButtonType  button_type)
{
  switch (button_type)
    {
    case ROOT_BUTTON:

      if (path_bar->root_icon != NULL)
	return path_bar->root_icon;
      
      path_bar->root_icon = gdk_pixbuf_new_from_file (DATA_DIR "/images/icon-bookmark.png", NULL); 

      return path_bar->root_icon;

    default:
      return NULL;
    }
  
  return NULL;
}

static void
button_data_free (ButtonData *button_data)
{
  gtk_file_path_free (button_data->name);
  g_free (button_data->name);
  g_free (button_data);
}

static const char *
get_name (ButtonData *button_data)
{
  return button_data->name;
}

/* We always want to request the same size for the label, whether
 * or not the contents are bold
 */
static void
label_size_request_cb (GtkWidget      *widget,
		       GtkRequisition *requisition,
		       ButtonData     *button_data)
{
  const gchar *name = get_name (button_data);
  PangoLayout *layout = gtk_widget_create_pango_layout (button_data->label, name);
  gint bold_width, bold_height;
  gchar *markup;

  pango_layout_get_pixel_size (layout, &requisition->width, &requisition->height);
  
  markup = g_markup_printf_escaped ("<b>%s</b>", name);
  pango_layout_set_markup (layout, markup, -1);
  g_free (markup);

  pango_layout_get_pixel_size (layout, &bold_width, &bold_height);
  requisition->width = MAX (requisition->width, bold_width);
  requisition->height = MAX (requisition->height, bold_height);
  
  g_object_unref (layout);
}

static void
bmp_path_bar_update_button_appearance (BmpPathBar *path_bar,
				       ButtonData *button_data,
				       gboolean    current_dir)
{
  const gchar *name = get_name (button_data);

  if (button_data->label != NULL)
    {
      if (current_dir)
	{
	  char *markup;

	  markup = g_markup_printf_escaped ("<b>%s</b>", name);
	  gtk_label_set_markup (GTK_LABEL (button_data->label), markup);
	  g_free (markup);
	}
      else
	{
	  gtk_label_set_text (GTK_LABEL (button_data->label), name);
	}
    }

  if (button_data->image != NULL)
    {
      GdkPixbuf *pixbuf;
      pixbuf = get_button_image (path_bar, button_data->type);
      gtk_image_set_from_pixbuf (GTK_IMAGE (button_data->image), pixbuf);
    }

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button_data->button)) != current_dir)
    {
      button_data->ignore_changes = TRUE;
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button_data->button), current_dir);
      button_data->ignore_changes = FALSE;
    }
}

/**
 * _bmp_path_bar_up:
 * @path_bar: a #BmpPathBar
 * 
 * If the selected button in the pathbar is not the furthest button "up" (in the
 * root direction), act as if the user clicked on the next button up.
 **/
void
_bmp_path_bar_up (BmpPathBar *path_bar)
{
  GList *l;

  for (l = path_bar->button_list; l; l = l->next)
    {
      GtkWidget *button = BUTTON_DATA (l->data)->button;
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
	{
	  if (l->next)
	    {
	      GtkWidget *next_button = BUTTON_DATA (l->next->data)->button;
	      button_clicked_cb (next_button, l->next->data);
	    }
	  break;
	}
    }
}

/**
 * _bmp_path_bar_down:
 * @path_bar: a #BmpPathBar
 * 
 * If the selected button in the pathbar is not the furthest button "down" (in the
 * leaf direction), act as if the user clicked on the next button down.
 **/
void
_bmp_path_bar_down (BmpPathBar *path_bar)
{
  GList *l;

  for (l = path_bar->button_list; l; l = l->next)
    {
      GtkWidget *button = BUTTON_DATA (l->data)->button;
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
	{
	  if (l->prev)
	    {
	      GtkWidget *prev_button = BUTTON_DATA (l->prev->data)->button;
	      button_clicked_cb (prev_button, l->prev->data);
	    }
	  break;
	}
    }
}
