/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  The BMPx project hereby grants permission for non GPL-compatible GStreamer
 *  plugins to be used and distributed together with GStreamer and BMPx. This
 *  permission is above and beyond the permissions granted by the GPL license
 *  BMPx is covered by.
 */

#include <config.h>
#include <revision.h>
#include <build.h>

#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#else
#  include <locale.h>
#endif

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>
#include <stdlib.h>

#include <glib-object.h>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#ifndef HAVE_DBUS
#include "ping_socket.h"
#endif

#include "error.h"

#ifdef HAVE_GUI
#  include "ui.h"
#  include "ui_splash.h"
#  include "sm.h"
#  include "fam_monitor.h"
#  include "xdb.h"
#ifdef HAVE_STARTUP_NOTIFICATION
# define SN_API_NOT_YET_FROZEN
# include <libsn/sn-launchee.h>
# include <gdk/gdkx.h>
static SnLauncheeContext *sn_context = NULL;
static SnDisplay *sn_display = NULL;
#endif
#endif

#include "gc.h"
#include "logger.h"
#include "play.h"
#include "plmanager.h"
#include "resource_manager.h"
#include "signals.h"

#include <bmpx/sanity.h>
#include <bmpx/plugin.h>
#include <bmpx/vfs.h>
#include <bmpx/metadata.h>
#include <bmpx/paramspecs.h>
#include <bmpx/playlist.h>

#include <xcs/xcs.h>

#include "main.h"
#include "loader.h"

#include <gst/gst.h>

#ifdef HAVE_DBUS
#  include <dbus/dbus-gtype-specialized.h>
#endif

#include "goa/libgoa.h"
#include "bmpx/util.h"

gchar *bmp_paths[BMP_N_PATHS] = { NULL };

Xcs                *_xcs = NULL;

BmpResourceManager *_bmp_rm = NULL;
BmpPlaylist        *_bmp_playlist = NULL;
BmpPlay            *_bmp_play_engine = NULL;
BmpSystemControl   *_bmp_system_control = NULL;
FamMonitor         *_bmp_fam_monitor = NULL;
PLManager          *_bmp_pl_manager = NULL;
#ifdef HAVE_GUI
BmpUI              *_bmp_ui = NULL;

gboolean            progress_running = FALSE;
gboolean            use_xdb = FALSE;
#endif
BmpMetadataCache   *_bmp_metadata_cache = NULL;
NGC                *ngc = NULL;

#ifndef HAVE_GUI
GMainLoop          *mainloop;
#endif

#ifdef HAVE_HAL
BmpHal		   *_bmp_hal;
#endif

#ifdef HAVE_SM
static gchar       *session_id = NULL;
#endif

static gboolean show_version  = FALSE;
#ifdef HAVE_GUI
gboolean run_headless  = FALSE;
static gboolean forced_headless = FALSE;
#endif
static gboolean detach = FALSE;
static gboolean log_stdout    = FALSE;

#ifdef HAVE_DBUS
static gboolean no_remote     = FALSE;
#endif

static GOptionEntry bmp_options[] =
  {
    {"version", 'v', 0, G_OPTION_ARG_NONE, &show_version,
     N_("Display version"), NULL},
    {"no-log", 0, 0, G_OPTION_ARG_NONE, &log_stdout,
     N_("Disable logfile, log messages to shell"), NULL},
#ifdef HAVE_GUI
    {"use-xdb", 0, 0, G_OPTION_ARG_NONE, &use_xdb,
     N_("Use XDB (X Debugger), verbose X events logging"), NULL},
    {"no-ui", 0, 0, G_OPTION_ARG_NONE, &run_headless,
     N_("Disable UI"), NULL},
    {"detach", 0, 0, G_OPTION_ARG_NONE, &detach,
     N_("Detach from the tty"), NULL},
#else
    {"detach", 0, 0, G_OPTION_ARG_NONE, &detach,
     N_("Detach from the tty"), NULL},
#endif

#ifdef HAVE_DBUS
    {"no-remote", 0, 0, G_OPTION_ARG_NONE, &no_remote,
     N_("Disable DBUS remote interface"), NULL},
#endif
#ifdef HAVE_SM
    {"sm-client-id", 0, 0, G_OPTION_ARG_STRING, &session_id,
     N_("Specify session management ID"), N_("ID")},
#endif
    { NULL }
  };

static const gchar *features[] =
  {
#ifdef HAVE_SM
    N_("feature|X11 Session Management"),
#endif
#ifdef HAVE_DBUS
    N_("feature|D-BUS Support"),
#endif
#ifdef USE_HAL
    N_("feature|HAL support"),
#endif
    N_("feature|Audio backend: GStreamer"),
#ifdef USE_GCONF
    N_("feature|Configuration backend: Gconf"),
#else
    N_("feature|Configuration backend: XML file"),
#endif
#ifdef USE_AMAZON
    N_("feature|Amazon album cover support"),
#endif
#ifdef HAVE_GUI
    N_("feature|Core GUI (Winamp 2.x skinning engine)")
#endif
  };

static void
print_version (void)
{
  g_print (Q_("%s %s, Build: %s-R%s [%s] (%s)\n\n"), PACKAGE, VERSION, RV_LAST_CHANGED_DATE, RV_REVISION, BUILD_DATE, BUILD_BUILDUSER);

  if (G_N_ELEMENTS (features) > 0)
    {
      gint i;

      g_print (ngettext ("Feature built: \n", "Features built: \n",
                         G_N_ELEMENTS (features)));

      for (i = 0; i < G_N_ELEMENTS (features); i++)
        g_print ("* %s\n", Q_(features[i]));
    }
}

static gboolean
make_user_dir (const gchar *name)
{
  const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
  return g_mkdir (name, mode755);
}

static void
bmp_make_user_dirs (void)
{
  make_user_dir (BMP_PATH_USER_DIR);
  make_user_dir (BMP_PATH_USER_SKIN_DIR);
  make_user_dir (BMP_PATH_SKIN_THUMB_DIR);
  make_user_dir (BMP_PATH_COVER_THUMB_DIR);
  make_user_dir (BMP_PATH_PLAYLISTS);
}

static void
bmp_free_paths (void)
{
  int i;

  for (i = 0; i < BMP_N_PATHS; i++)
    {
      g_free (bmp_paths[i]);
      bmp_paths[i] = NULL;
    }
}

#define USER_PATH(path) \
    g_build_filename (BMP_PATH_USER_DIR, path, NULL);

static void
bmp_init_paths (void)
{
  BMP_PATH_USER_DIR = g_build_filename(g_get_home_dir(), BMP_RCPATH, NULL);
  BMP_PATH_USER_SKIN_DIR = USER_PATH(BMP_SKIN_DIR_BASENAME);
  BMP_PATH_SKIN_THUMB_DIR = USER_PATH(BMP_SKIN_THUMB_DIR_BASENAME);
  BMP_PATH_PLAYLISTS = USER_PATH(BMP_PLAYLISTS_DIR_BASENAME);
  BMP_PATH_LOG_FILE = USER_PATH(BMP_LOG_BASENAME);
  BMP_PATH_COVER_THUMB_DIR = USER_PATH(BMP_COVER_THUMB_DIR_BASENAME);
#if 0
  BMP_PATH_CONFIG_FILE = USER_PATH(BMP_CONFIG_BASENAME);
  BMP_PATH_ACCEL_FILE = USER_PATH(BMP_ACCEL_BASENAME);
#endif

  bmp_make_user_dirs ();
}

static void
setup_i18n (void)
{
#ifdef HAVE_GUI
  gtk_set_locale ();
#else
  setlocale (LC_ALL, "");
#endif

  bindtextdomain (PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (PACKAGE, "UTF-8");
  textdomain (PACKAGE);
}

#ifdef HAVE_STARTUP_NOTIFICATION
static void
sn_error_trap_push(SnDisplay *display, Display *xdisplay)
{
        gdk_error_trap_push();
}

static void
sn_error_trap_pop(SnDisplay *display, Display *xdisplay)
{
        gdk_error_trap_pop();
}

static void
startup_notification_complete(void)
{
        Display *xdisplay;

        xdisplay = GDK_DISPLAY();
        sn_display = sn_display_new(xdisplay,
                                    sn_error_trap_push,
                                    sn_error_trap_pop);
        sn_context =
                sn_launchee_context_new_from_environment(sn_display,
                                                         DefaultScreen(xdisplay));

        if (sn_context != NULL)
        {
                sn_launchee_context_complete(sn_context);
                sn_launchee_context_unref(sn_context);

                sn_display_unref(sn_display);
        }
}
#endif /* HAVE_STARTUP_NOTIFICATION */

gint
main (gint    argc,
      gchar** argv)
{
  GError *error = NULL;
  gchar  *sanity_error = NULL;
  GOptionContext *oc_bmp;
  gboolean new_configuration = FALSE;

  g_thread_init (NULL);

  setup_i18n ();

  sanity_error = sanity_check_libs ();
  if (sanity_error)
    {
      g_printerr ("%s\n", sanity_error);
      g_free (sanity_error);

      exit (EXIT_FAILURE);
    }

  /* Attempt to run X11-less if we are --daemon or --no-ui */
  oc_bmp = g_option_context_new(" - Run BMPx");
  g_option_context_add_main_entries (oc_bmp, bmp_options, PACKAGE);

#ifdef HAVE_GUI
  g_option_context_add_group (oc_bmp, gtk_get_option_group (FALSE));
#endif
  if (!g_option_context_parse (oc_bmp, &argc, &argv, &error))
    {
      g_printerr (Q_("Invalid argument: %s\n"), error->message);
      g_error_free (error);
      exit (EXIT_FAILURE);
    }

  if (detach)
    {
      if (!bmp_detach ())
        g_warning (_("Unable to detach, proceeding with normal startup."));
    }

  bmp_init_paths ();

  if ((!run_headless) && (!g_getenv("DISPLAY")))
    {
      run_headless = TRUE;
      forced_headless = TRUE;
    }

  if (show_version)
    {
      print_version ();
      goto early_exit;
    }

  if (!log_stdout)
    {
      bmp_logger_start (BMP_PATH_LOG_FILE);
    }
  else
    {
      g_message ("BMPx STARTED [%s] PID [%d] ", get_timestamp_str (), getpid ());
    }

#ifndef HAVE_DBUS
  /* NOTE: ping_socket() returns FALSE when BMPx really isn't present OR
     errors were present */
  if (ping_socket ())
    {
      g_printerr (Q_("BMPx is already running.\n"));
      goto early_exit;
    }

  ping_socket_init ();
#endif

#ifdef HAVE_GUI
  if (!run_headless)
   {
     gtk_init (&argc, &argv);
   }
#endif

  ngc = gc_new ();

  vfs_init ();
  gst_init (&argc, &argv);

  bmp_param_spec_types_init ();

  _bmp_rm = bmp_resource_manager_new ();
  _xcs = xcs_new (&new_configuration);

  if (new_configuration)
    {
#ifndef USE_GCONF
      xcs_key_set_string (xcs, "audio","sink", DEFAULT_SINK);
      xcs_key_set_string (xcs, "audio","device", DEFAULT_DEVICE);

      xcs_key_set_string (xcs, "bmp","font", BMP_DEFAULT_FONT);
      xcs_key_set_string (xcs, "bmp","titlestring-format","%{artist} - %{album} - %{title}");
      xcs_key_set_string (xcs, "bmp","icon-theme","red");
      xcs_key_set_string (xcs, "bmp","skin",
                          DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN);
      xcs_key_set_boolean (xcs, "bmp","use-custom-cursors", TRUE);
      xcs_key_set_boolean (xcs, "bmp","display-tooltips", TRUE);
      xcs_key_set_boolean (xcs, "bmp","display-tracklist-numbers", TRUE);
      xcs_key_set_boolean (xcs, "bmp","font-per-skin",FALSE);
      xcs_key_set_boolean (xcs, "bmp","shuffle", FALSE);
      xcs_key_set_boolean (xcs, "bmp","repeat", FALSE);
      xcs_key_set_boolean  (xcs, "tracklist","column_0_visible", TRUE);
      xcs_key_set_boolean  (xcs, "tracklist","column_1_visible", TRUE);
      xcs_key_set_boolean  (xcs, "tracklist","column_2_visible", TRUE);
      xcs_key_set_boolean  (xcs, "tracklist","column_3_visible", TRUE);
      xcs_key_set_boolean  (xcs, "tracklist","column_4_visible", FALSE);
      xcs_key_set_boolean  (xcs, "tracklist","column_5_visible", FALSE);
      xcs_key_set_boolean  (xcs, "tracklist","column_6_visible", TRUE);
#endif
      xcs_key_set_boolean (xcs, "bmp","keep-above", FALSE);
      xcs_key_set_boolean (xcs, "bmp","file-chooser-close-on-open", TRUE);
      xcs_key_set_boolean (xcs, "bmp","file-chooser-close-on-add", FALSE);
      xcs_key_set_string  (xcs, "bmp","file-chooser-path", g_get_home_dir());
      xcs_key_set_string  (xcs, "bmp","playlists-path", BMP_PATH_PLAYLISTS);
    }

  /* Initialize Plugins */
  plugins_initialize();

  /* Init signal handlers */
  signal_handlers_init ();

  /* FIXME: gtk_init_with_args() (above) will remove all recognized
     arguments, so the session manager will run BMPx with incomplete
     options when it restarts. */

#ifdef HAVE_GUI
  if (xcs_key_fetch_boolean (xcs, "bmp", "no-ui")) run_headless = TRUE;

  if (!run_headless)
    {
      session_start (argc, argv, session_id);
#ifdef HAVE_STARTUP_NOTIFICATION
      gtk_window_set_auto_startup_notification (FALSE);
#endif

      bmp_splash_window_show ();
      progress_running = TRUE;

      while (gtk_events_pending ())
        gtk_main_iteration ();
    }
#endif


#ifdef HAVE_GUI
  _bmp_fam_monitor = fam_monitor_new ();
#endif

#ifdef HAVE_DBUS
  if (xcs_key_fetch_boolean (xcs, "bmp", "no-remote"))
    no_remote = TRUE;
#endif

  _bmp_playlist = bmp_playlist_new ();
  _bmp_play_engine = bmp_play_new ();

#ifdef HAVE_DBUS
  if (!no_remote)
    dbus_g_type_specialized_init ();

  _bmp_system_control = bmp_system_control_new (no_remote);
#else
  _bmp_system_control = bmp_system_control_new (FALSE);
#endif

  if (bmp_system_control)
    {
#ifdef HAVE_HAL
      _bmp_hal = bmp_hal_new ();
#endif

      _bmp_metadata_cache = bmp_metadata_cache_new ();
      bmp_playlist_load_playlist (bmp_playlist);
      _bmp_pl_manager = pl_manager_new ();

#ifdef HAVE_GUI
      if (!run_headless)
        {
          bmp_system_control_ui_start (bmp_system_control, NULL, NULL);
        }
      else
        {
          if (!forced_headless)
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Not starting UI on request.");
            }
          else
            {
              g_printerr (Q_("No X11 server found (DISPLAY environment variable not set): Not starting UI\n"));
            }
        }
#endif

#ifdef HAVE_GUI
      if (!run_headless)
        {
#ifdef HAVE_STARTUP_NOTIFICATION
          startup_notification_complete ();
#endif
          bmp_splash_window_hide ();
          progress_running = FALSE;
        }

      g_idle_add ((GSourceFunc)bmp_system_control_send_startup_complete, bmp_system_control);
      gtk_main ();
#else
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, _("Starting mainloop."));
      mainloop = g_main_loop_new (NULL, FALSE);
      g_main_loop_run (mainloop);
#endif
    }
  else
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, _("BmpSystemControl not present, cannot proceed with startup sequence."));
    }

  vfs_stop();

  if ( bmp_play_engine && bmp_system_control)
    bmp_play_stop_stream (bmp_play_engine);

#ifdef HAVE_GUI
  if (_bmp_fam_monitor)
    {
      g_object_unref (_bmp_fam_monitor);
      _bmp_fam_monitor = NULL;
    }

  if (_bmp_ui)
    {
      g_object_unref (_bmp_ui);
      _bmp_ui = NULL;
    }
#endif

 if (_bmp_playlist)
    {
      g_object_unref (_bmp_playlist);
      _bmp_playlist = NULL;
    }

  if (_xcs)
    {
      g_object_unref (_xcs);
      _xcs = NULL;
    }

  if (_bmp_pl_manager)
    {
      g_object_unref (_bmp_pl_manager);
      _bmp_pl_manager = NULL;
    }

  if (_bmp_system_control)
    {
      bmp_system_control_send_shutdown_complete (bmp_system_control);
      g_object_unref (_bmp_system_control);
      _bmp_system_control = NULL;
    }

  if (_bmp_metadata_cache)
    {
      g_object_unref (_bmp_metadata_cache);
      _bmp_metadata_cache = NULL;
    }

#ifdef HAVE_HAL
  if (_bmp_hal)
    {
      g_object_unref (_bmp_hal);
      _bmp_hal = NULL;
    }
#endif

  if (_bmp_play_engine)
    {
      g_object_unref (_bmp_play_engine);
      _bmp_play_engine = NULL;
    }

#ifdef HAVE_GUI
  if (!run_headless)
    session_end ();
#endif

  /* Shutdown Plugins */
  plugins_shutdown();

  gc_destroy (ngc);

#ifndef HAVE_DBUS
  ping_socket_close ();
#endif

early_exit:

  if (!log_stdout)
    {
      bmp_logger_stop ();
    }
  else
    {
      g_message (" ");
      g_message ("BMPx STOPPED [%s]", get_timestamp_str());
    }

  bmp_free_paths ();

  exit (EXIT_SUCCESS);
}
