/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <config.h>
#include <bmpx/uri.h>

#include <gst/gst.h>
#include <gst/gstelement.h>
#include <gst/interfaces/mixer.h>
#include <gst/interfaces/mixertrack.h>
#include <gst/interfaces/mixeroptions.h>

#include <goa/libgoa.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>
#include <libhrel/relation.h>
#include <libhrel/relation_base.h>

#include "play.h"

#define BMP_PLAY_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), BMP_TYPE_PLAY, BmpPlayPrivate))

#define BMP_GST_BUFFER_TIME   ((gint64) 50000)
#define BMP_GST_LATENCY_TIME  ((gint64) 500)
#define BMP_GST_PLAY_TIMEOUT  ((gint64) 3000000000)

G_DEFINE_TYPE (BmpPlay, bmp_play, G_TYPE_OBJECT)

#include "main.h"

enum
{
  BMP_PLAY_SIGNAL_EOS,
  BMP_PLAY_SIGNAL_ERROR,
  BMP_PLAY_SIGNAL_BUFFER,
  BMP_PLAY_SIGNAL_SET_STREAM_TITLE,

  BMP_PLAY_N_SIGNALS
};

enum
{
  BMP_PLAY_PROP_STREAM = 1,
  BMP_PLAY_PROP_PLAYSTATUS = 2,
  BMP_PLAY_N_PROPS
};

enum
{
  P_FILE,
  P_HTTP,

  N_PIPES
};

enum
{
  PIPELINE,

  SRC,
  DECODER,
  CONVERT,
  SINK,
  VOLUME,

  N_ELEMENTS
};

typedef struct _BmpPlayPrivate BmpPlayPrivate;
struct _BmpPlayPrivate
{
  gboolean dispose_has_run;

  GstElement *elements[N_PIPES][N_ELEMENTS];

  gint	      current_pipe;

  gchar	     *mrl;
  gint	      playstatus;
  gdouble     volume;

  guint	      source;
  gboolean    sending;
};

static guint signals[BMP_PLAY_N_SIGNALS] = { 0, 0, 0, 0 };

static GObject *bmp_play_constructor (GType type,
                                      guint n_construct_properties,
                                      GObjectConstructParam *construct_properties);

static gboolean bmp_play_bus_watch (GstBus	*bus,
                                    GstMessage	*message,
                                    BmpPlay	*self);

static void
pipelines_create		   (BmpPlay	*self);

static void
send_stream_pos_stop (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GSource *source;

  if (private->source)
    {
	  source = g_main_context_find_source_by_id (NULL, private->source);
	  if (source)
	    {
	      g_source_destroy (source);
	      private->sending = FALSE;
	    }
    }
}

static gboolean
send_stream_pos (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint            stream_pos = bmp_play_get_time (self);

  if (!private->elements[private->current_pipe][PIPELINE])
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      private->sending = FALSE;
      return FALSE;
    }

  if ((stream_pos >= 0) && (GST_STATE(private->elements[private->current_pipe][PIPELINE]) >= GST_STATE_PLAYING))
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", stream_pos, NULL);
    }
  else
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      private->sending = FALSE;
      return FALSE;
    }

  return TRUE;
}

void
bmp_play_set_volume (BmpPlay *self,
                     gint     volume)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gdouble _volume;

  g_return_if_fail (BMP_IS_PLAY (self));

  _volume = volume / 100.;
  private->volume = _volume;
  g_object_set (BMP_PLAY_GET_PRIVATE (self)->elements[private->current_pipe][VOLUME], "volume", _volume, NULL);
}

gint
bmp_play_get_volume (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gdouble volume;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  g_object_get (private->elements[private->current_pipe][VOLUME], "volume", &volume, NULL);
  return (gint) (volume * 100);
}

static void
pipelines_destroy (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  g_return_if_fail (BMP_IS_PLAY (self));

  if (private->elements[P_FILE][PIPELINE])
    {
      gst_element_set_state (private->elements[P_FILE][PIPELINE], GST_STATE_NULL);
      gst_object_unref (private->elements[P_FILE][PIPELINE]);
    }

  if (private->elements[P_HTTP][PIPELINE])
    {
      gst_element_set_state (private->elements[P_HTTP][PIPELINE], GST_STATE_NULL);
      gst_object_unref (private->elements[P_HTTP][PIPELINE]);
    }
}

static void
get_stream_title (GObject *gobject,
                  GParamSpec *arg1,
                  gpointer user_data)
{
}

static gulong handler_id = 0;

static void
link_pad (GstElement *element,
	  GstPad     *pad,
	  gboolean    last,
	  gpointer    data)
{
  BmpPlay	  *self = PLAY (data);
  BmpPlayPrivate  *private = BMP_PLAY_GET_PRIVATE (self);
  GstPad	  *sinkpad;

  sinkpad = gst_element_get_pad (private->elements[private->current_pipe][CONVERT], "sink");
  gst_pad_link (pad, sinkpad);
}

static void
pipelines_create (BmpPlay	  *self)
{
  BmpPlayPrivate	    *private = BMP_PLAY_GET_PRIVATE (self);
  GError		    *error = NULL;
  gchar			    *device, *sink;
  GstStateChangeReturn	     ret;

  sink = xcs_key_fetch_string (xcs, "audio", "sink");
  device = xcs_key_fetch_string (xcs, "audio", "device");

  //Create FILE pipeline
  private->elements[P_FILE][SRC]	= gst_element_factory_make ("filesrc",	    "src0");
  private->elements[P_FILE][DECODER]	= gst_element_factory_make ("decodebin",    "decoder0");
  private->elements[P_FILE][CONVERT]	= gst_element_factory_make ("audioconvert", "audioconvert0");
  private->elements[P_FILE][VOLUME]	= gst_element_factory_make ("volume",	    "volume0");

  private->elements[P_FILE][SINK]	= gst_element_factory_make (sink,   "sink0");
  g_object_set (G_OBJECT(private->elements[P_FILE][SINK]), "device", device, NULL);
  
  private->elements[P_FILE][PIPELINE]	= gst_pipeline_new ("pipeline_file"); 

  gst_bin_add_many (GST_BIN (private->elements[P_FILE][PIPELINE]),
		    private->elements[P_FILE][SRC],	
		    private->elements[P_FILE][DECODER],	
		    private->elements[P_FILE][CONVERT],
		    private->elements[P_FILE][VOLUME],
		    private->elements[P_FILE][SINK],
		    NULL);

  gst_element_link_many (private->elements[P_FILE][SRC], 
			 private->elements[P_FILE][DECODER], NULL);

  gst_element_link_many (private->elements[P_FILE][CONVERT], 
			 private->elements[P_FILE][VOLUME], 
			 private->elements[P_FILE][SINK], NULL);

  handler_id = g_signal_connect (G_OBJECT(private->elements[P_FILE][DECODER]),
		    "new-decoded-pad", 
		    G_CALLBACK (link_pad),
		    self);

  g_object_set (G_OBJECT(private->elements[P_FILE][SINK]), "latency-time", (guint64)500, "buffer-time", (guint64)250, NULL);
  gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(private->elements[P_FILE][PIPELINE])), (GstBusFunc) bmp_play_bus_watch, self);

  //Create HTTP pipeline
  private->elements[P_HTTP][SRC]	= gst_element_factory_make ("neonhttpsrc",  "src1");

  if (!private->elements[P_HTTP][SRC])
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: No HTTP support available!", G_STRLOC);
      private->elements[P_HTTP][PIPELINE] = NULL;
    }
  else
    {
      private->elements[P_HTTP][DECODER]	= gst_element_factory_make ("decodebin",    "decoder1");
      private->elements[P_HTTP][CONVERT]	= gst_element_factory_make ("audioconvert", "audioconvert1");
      private->elements[P_HTTP][VOLUME]		= gst_element_factory_make ("volume",	    "volume1");

      private->elements[P_HTTP][SINK]	= gst_element_factory_make (sink, "sink1");
      g_object_set (G_OBJECT(private->elements[P_FILE][SINK]), "device", device, NULL);

      private->elements[P_HTTP][PIPELINE]	= gst_pipeline_new ("pipeline_http"); 

      gst_bin_add_many (GST_BIN (private->elements[P_HTTP][PIPELINE]),
		    private->elements[P_HTTP][SRC],	
		    private->elements[P_HTTP][DECODER],	
		    private->elements[P_HTTP][CONVERT],
		    private->elements[P_HTTP][VOLUME],
		    private->elements[P_HTTP][SINK],
		    NULL);

      gst_element_link_many (private->elements[P_HTTP][SRC], 
			 private->elements[P_HTTP][DECODER], NULL);

      gst_element_link_many (private->elements[P_HTTP][CONVERT], 
			 private->elements[P_HTTP][VOLUME], 
			 private->elements[P_HTTP][SINK], NULL);

      handler_id = g_signal_connect (G_OBJECT(private->elements[P_HTTP][DECODER]),
		    "new-decoded-pad", 
		    G_CALLBACK (link_pad),
		    self);

      g_object_set (G_OBJECT(private->elements[P_HTTP][SINK]), "latency-time", (guint64)500, "buffer-time", (guint64)250, NULL);
      gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(private->elements[P_HTTP][PIPELINE])), (GstBusFunc) bmp_play_bus_watch, self);
  }
  g_free (sink);
  g_free (device);
}

gboolean
bmp_play_stop_stream (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstPad *pad;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  if (!private->elements[private->current_pipe][PIPELINE]) return FALSE;

  gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_NULL);
  send_stream_pos_stop (self);

  pipelines_destroy (self);
  pipelines_create (self);

#if 0
  //Hack around for decodebin brokenness
  pad = gst_element_get_pad (private->elements[P_FILE][CONVERT], "sink"); 
  if (gst_pad_is_linked (pad))
	{
	  GstPad *peer_pad;
	  peer_pad = gst_pad_get_peer (pad);
	  gst_pad_unlink (peer_pad, pad);
	}
#endif

  return TRUE;
}

gboolean
bmp_play_play_stream (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstPad *pad;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
 
  if (g_uri_test (private->mrl, G_URI_TEST_PROTOCOL, PROTOCOL_HTTP) && (!private->elements[P_HTTP][PIPELINE]))
    {
	g_object_set (G_OBJECT(self), "playstatus", PLAYSTATUS_STOPPED, NULL);
	return FALSE;
    }

#if 0
  //Hack around for decodebin brokenness
  pad = gst_element_get_pad (private->elements[P_FILE][CONVERT], "sink"); 
  if (gst_pad_is_linked (pad))
	{
	  GstPad *peer_pad;
	  peer_pad = gst_pad_get_peer (pad);
	  gst_pad_unlink (peer_pad, pad);
	}
#endif

  if (gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE)
    {
      g_message ("%s: Error in changing state to PLAYING", G_STRFUNC);
      gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_READY);
      return FALSE;
    }

  return TRUE;
}

static GstElement *playbin_tags = NULL;

void
tag_foreach (const GstTagList *list,
             const gchar      *tag,
             HTuple           *tuple)
{
  GValue value = { 0, };

  gst_tag_list_copy_value (&value, list, tag);
  if (G_VALUE_HOLDS (&value, G_TYPE_STRING))
    {
      g_message ("%s: %s", tag, g_value_get_string(&value));

      h_tuple_insert (tuple,
			tag,	
			G_TYPE_STRING,
			(gpointer)g_value_get_string (&value));	
      return;
    }

  if (G_VALUE_HOLDS (&value, G_TYPE_INT))
    {
      g_message ("%s: %d", tag, g_value_get_int(&value));

      h_tuple_insert (tuple,
			tag,	
			G_TYPE_INT,
			GINT_TO_POINTER(g_value_get_int (&value)));	
      return;
    }

  if (G_VALUE_HOLDS (&value, G_TYPE_UINT))
    {
      gint _v_int;

      _v_int = g_value_get_uint (&value);

      g_message ("%s: %d", tag, g_value_get_int(&value));

      h_tuple_insert (tuple,
			tag,	
			G_TYPE_INT,
			GINT_TO_POINTER(_v_int));
      return;
    }
}

static gboolean
bus_watch_tags (GstBus     *bus,
                GstMessage *message,
                HTuple     *tuple)
{
  GstTagList *tag_list;

  switch (GST_MESSAGE_TYPE (message))
    {
    case GST_MESSAGE_TAG:
      gst_message_parse_tag (message, &tag_list);
      gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach,
                            tuple);
      break;

    case GST_MESSAGE_ERROR:
      {
        GError *error = NULL;

        gst_message_parse_error (message, &error, NULL);
        g_print ("%s ERROR: %s\n", G_STRLOC, error->message);
        g_error_free (error);

        gst_element_set_state (playbin_tags, GST_STATE_READY);
      }
      break;

    case GST_MESSAGE_WARNING:
      {
        GError *error = NULL;

        gst_message_parse_warning (message, &error, NULL);
        g_print ("%s WARNING: %s\n", G_STRLOC, error->message);
        g_error_free (error);

        /* FIXME: Not sure, should we do this here?: gst_element_set_state (playbin_tags, GST_STATE_READY); */
      }
      break;


    default:;
    }

  return TRUE;
}

HTuple*
bmp_play_get_metadata_gst (BmpPlay     *self,
                           const gchar *uri)
{
  HTuple  *tuple;

  g_return_val_if_fail (BMP_IS_PLAY (self), NULL);

  playbin_tags = gst_element_factory_make ("playbin", "playbin");

  tuple = bmp_metadata_tuple_create_default (uri);

  h_tuple_insert (tuple,
		  bmp_metadata_get_id_static (BMP_DATUM_TITLE),	
		  G_TYPE_STRING,
		  (gpointer)uri);	

  gst_bus_add_watch (gst_element_get_bus (GST_ELEMENT (playbin_tags)),
                     (GstBusFunc) bus_watch_tags, tuple);

  g_object_set (G_OBJECT (playbin_tags), "uri", uri, NULL);
  gst_element_set_state (playbin_tags, GST_STATE_PAUSED);
  gst_element_set_state (playbin_tags, GST_STATE_NULL);
  g_object_unref (G_OBJECT (playbin_tags));

  h_tuple_insert (tuple,
		  bmp_metadata_get_id_static (BMP_DATUM_LOCATION),	
		  G_TYPE_STRING,
		  (gpointer)uri);	

  return tuple;
}

gboolean
bmp_play_pause_stream (BmpPlay *self)
{
  BmpPlayPrivate *private;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  private = BMP_PLAY_GET_PRIVATE (self);

  switch (GST_STATE (private->elements[private->current_pipe][PIPELINE]))
    {
    case GST_STATE_PLAYING:
      return gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_PAUSED);

    case GST_STATE_PAUSED:
      return gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_PLAYING);

    default:
      return FALSE;
    }
}

gboolean
bmp_play_seek_stream (BmpPlay *self,
                      gint     offset)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstEvent *s_event;
  guint64   s_offset;
  gboolean  res = FALSE;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  s_offset = (offset * GST_SECOND) / 1000;
  s_event = gst_event_new_seek (1.0,
      GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, 
      s_offset, 
      GST_SEEK_TYPE_NONE, 0);

  res = gst_element_send_event (private->elements[private->current_pipe][SINK], s_event);

  return TRUE;
}

GstState
bmp_play_get_element_state (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  g_return_val_if_fail (BMP_IS_PLAY (self), GST_STATE_NULL);

  return GST_STATE (private->elements[private->current_pipe][PIPELINE]);
}

gint
bmp_play_get_time (BmpPlay *self)
{
  BmpPlayPrivate *private;
  GstQuery       *query;
  GstFormat       fmt = GST_FORMAT_TIME;
  gint64          time_in_nanoseconds;
  gint            time_in_seconds;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!(GST_STATE (private->elements[private->current_pipe][PIPELINE]) == GST_STATE_PLAYING))
      return 0;

  query = gst_query_new_position (fmt);
  gst_element_query (private->elements[private->current_pipe][PIPELINE], query);
  gst_query_parse_position (query, &fmt, &time_in_nanoseconds);

  time_in_seconds = time_in_nanoseconds / 1000000000;

  return time_in_seconds;
}

gint
bmp_play_get_length (BmpPlay *self)
{
  BmpPlayPrivate *private;
  GstQuery       *query;
  GstFormat       fmt = GST_FORMAT_TIME;
  gint64          length_in_nanoseconds;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->elements[private->current_pipe][PIPELINE])
      return 0;

  if (!(GST_STATE (private->elements[private->current_pipe][PIPELINE]) == GST_STATE_PLAYING))
      return 0;

  query = gst_query_new_duration (fmt);
  gst_element_query (private->elements[private->current_pipe][PIPELINE], query);
  gst_query_parse_duration (query, &fmt, &length_in_nanoseconds);

  length_in_nanoseconds = length_in_nanoseconds / 1000000000;

  return length_in_nanoseconds;
}

gint
bmp_play_get_status (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint status;

  g_return_val_if_fail (BMP_IS_PLAY (self), PLAYSTATUS_STOPPED);

  status = GST_STATE (private->elements[private->current_pipe][PIPELINE]);

  if (status == GST_STATE_READY)
    return PLAYSTATUS_STOPPED;

  if (status == GST_STATE_PLAYING)
    return PLAYSTATUS_PLAYING;

  return status;
}

static gboolean
playstatus_notify_cb (BmpPlay *self)
{
  gint status;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  g_object_get (self, "playstatus", &status, NULL);

  if (status & PLAYSTATUS_SEEKING)
    {
      bmp_play_pause_stream (self);
      return TRUE;
    }

  if (status & PLAYSTATUS_STOPPED)
    {
      bmp_play_stop_stream (self);
      return TRUE;
    }

  if (status & PLAYSTATUS_WAITING)
    {
      bmp_play_stop_stream (self);
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      return TRUE;
    }

  if (status & PLAYSTATUS_PAUSED)
    {
      bmp_play_pause_stream (self);
      return TRUE;
    }

  if (status & PLAYSTATUS_PLAYING)
    {
      bmp_play_play_stream (self);
      return TRUE;
    }

  return FALSE;
}

static void
bmp_play_set_property (GObject      *object,
                       guint         property_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
  BmpPlay        *self    = (BmpPlay *) object;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  switch (property_id)
  {
  case BMP_PLAY_PROP_PLAYSTATUS:
    {
      gint status = g_value_get_int (value);
      private->playstatus = status;
    }
    break;

  case BMP_PLAY_PROP_STREAM:
    {
      GstPad  *pad;
      gchar   *filename, 
	      *host;

      bmp_play_stop_stream (self);

      g_free (private->mrl);
      private->mrl = g_value_dup_string (value);

      if (g_uri_test (private->mrl, G_URI_TEST_PROTOCOL, PROTOCOL_FILE))
	{
	  private->current_pipe = P_FILE;
	  filename = g_filename_from_uri (private->mrl, &host, NULL);      
	  g_object_set (private->elements[private->current_pipe][SRC], "location", filename, NULL);

#if 0
	  gst_element_set_state (private->elements[private->current_pipe][DECODER], GST_STATE_NULL);
	  gst_element_set_state (private->elements[private->current_pipe][DECODER], GST_STATE_READY);
#endif

	  g_free (filename);
	}
      else if ((g_uri_test (private->mrl, G_URI_TEST_PROTOCOL, PROTOCOL_HTTP) && private->elements[P_HTTP][PIPELINE]))
	{
	  private->current_pipe = P_HTTP;

	  g_object_set (private->elements[private->current_pipe][SRC], "uri", private->mrl, NULL);

#if 0
	  gst_element_set_state (private->elements[private->current_pipe][DECODER], GST_STATE_NULL);
	  gst_element_set_state (private->elements[private->current_pipe][DECODER], GST_STATE_READY);
#endif
	}

    }
    break;

  default:
    g_assert (FALSE);
  }
}

static void
bmp_play_get_property (GObject    *object,
                       guint       property_id,
                       GValue     *value,
                       GParamSpec *pspec)
{
  BmpPlay        *self = (BmpPlay *) object;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  switch (property_id)
    {
    case BMP_PLAY_PROP_STREAM:
      g_value_set_string (value, private->mrl);
      break;

    case BMP_PLAY_PROP_PLAYSTATUS:
      g_value_set_int (value, private->playstatus);
      break;

    default:
      g_assert (FALSE);
    }
}

static void
bmp_play_dispose (GObject *obj)
{
  BmpPlay        *self    = (BmpPlay *) obj;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  if (private->dispose_has_run)
    return;

  private->dispose_has_run = TRUE;

  g_message ("stopped: %s", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_play_finalize (GObject *obj)
{
  /* nothing to do */
}

static void
bmp_play_class_init (BmpPlayClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (BmpPlayPrivate));

  gobject_class->set_property = bmp_play_set_property;
  gobject_class->get_property = bmp_play_get_property;
  gobject_class->dispose = bmp_play_dispose;
  gobject_class->finalize = bmp_play_finalize;
  gobject_class->constructor = bmp_play_constructor;


  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_STREAM,
                                   g_param_spec_string ("mrl",
                                                        "mrl",
                                                        "mrl",
                                                        "",
                                                        G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_PLAYSTATUS,
                                   g_param_spec_int ("playstatus",
                                                     "playstatus",
                                                     "player playing status",
                                                     0, 128, 0,
                                                     G_PARAM_READWRITE));

  signals[BMP_PLAY_SIGNAL_SET_STREAM_TITLE] =
   g_signal_new ("set_stream_title",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, set_stream_title),
                  NULL, NULL, g_cclosure_marshal_VOID__STRING,
                  G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[BMP_PLAY_SIGNAL_BUFFER] =
    g_signal_new ("buffer",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, buffer),
                  NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);

  signals[BMP_PLAY_SIGNAL_EOS] =
    g_signal_new ("eos",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, eos),
                  NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                  0);

  signals[BMP_PLAY_SIGNAL_ERROR] =
    g_signal_new ("error",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, error),
                  NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);
}

BmpPlay *
bmp_play_new (void)
{
    BmpPlay *play;

    play = g_object_new (BMP_TYPE_PLAY, NULL);
    return play;
}

static GObject *
bmp_play_constructor (GType type,
                      guint n_construct_properties,
                      GObjectConstructParam * construct_properties)
{
  GObject      *obj;
  BmpPlayClass *klass;
  GObjectClass *parent_class;

  klass = PLAY_CLASS (g_type_class_peek (BMP_TYPE_PLAY));

  parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));

  obj = parent_class->constructor (type,
                                   n_construct_properties,
                                   construct_properties);
  return obj;
}

static gboolean
_idle_emit_eos (BmpPlay * play)
{
  g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

  g_signal_emit (G_OBJECT (play), signals[BMP_PLAY_SIGNAL_EOS], 0, NULL);

  return FALSE;
}

static gboolean
bmp_play_bus_watch (GstBus     *bus,
                    GstMessage *message,
                    BmpPlay    *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstState old_state, new_state, pending_state;

  switch (GST_MESSAGE_TYPE (message))
    {
    case GST_MESSAGE_STATE_CHANGED:
      {
        gst_message_parse_state_changed (message, &old_state,
                                         &new_state, &pending_state);

	if ((old_state < GST_STATE_PLAYING) && (new_state == GST_STATE_PLAYING) && (!private->sending))
	  {
	      private->source = g_timeout_add_full (G_PRIORITY_DEFAULT, 250, (GSourceFunc) send_stream_pos, self, NULL);
	      private->sending = TRUE;
	  }
      }
      break;

    case GST_MESSAGE_ERROR:
      {
        GError  *error = NULL;
	gchar   *debug;

        gst_message_parse_error (message, &error, &debug);
        g_print ("%s ERROR: %s (%s)\n", G_STRLOC, error->message, debug);
        g_error_free (error);

	gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_NULL);
      }
    break;

    case GST_MESSAGE_WARNING:
      {
        GError  *error = NULL;
	gchar   *debug;

        gst_message_parse_warning (message, &error, &debug);
        g_print ("%s WARNING: %s (%s)\n", G_STRLOC, error->message, debug);
        g_error_free (error);
      }
    break;

    case GST_MESSAGE_EOS:
        {
            /* end-of-stream */
            g_idle_add ((GSourceFunc) _idle_emit_eos, self);
        }
        break;

    default:
        /* unhandled message */
        break;
    }

    return TRUE; 
}

static void
bmp_play_init (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint		  n;

  private->dispose_has_run = FALSE;
  private->mrl = NULL;
  private->sending = FALSE;
  private->source = 0;
  private->current_pipe = P_FILE;
  
  for (n = 0; n < N_ELEMENTS; n++) 
    {
	private->elements[P_FILE][n] = NULL;
	private->elements[P_HTTP][n] = NULL;
    }

  pipelines_create (self);

  g_signal_connect (G_OBJECT (self), "notify::playstatus",
                    G_CALLBACK (playstatus_notify_cb), NULL);
}
