/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 *
 */


#ifdef HAVE_GUI
#include <gtk/gtk.h>
#else
#include <glib.h>
#endif

#include "plmanager.h"

#include "main.h"
#include <bmpx/metadata.h>
#include <bmpx/metadata_common.h>
#include <bmpx/util.h>
#include <bmpx/rtfs.h>

G_DEFINE_TYPE(PLManager, pl_manager, G_TYPE_OBJECT)

//FIXME: How should we watch entries of the list? Proxied trough PLManager, or watch the HRelation
//       directly? (RFC?)

typedef enum {

	PL_MANAGER_SIGNAL_LIST_CREATED,
	PL_MANAGER_SIGNAL_LIST_DELETED,

	PL_MANAGER_MANAGER_N_SIGNALS,

} PLManagerSignals;

static guint signals[PL_MANAGER_MANAGER_N_SIGNALS] = {0};

typedef enum {
	PL_MANAGER_MANAGER_N_PROPERTIES
} PLManagerProperties;

#define UNSORTED_ATTR_ID -1

typedef struct _PLManagerPlaylist PLManagerPlaylist;

struct _PLManagerPlaylist {
    gchar	*name;
    HRelation	*relation;
    RTFS	*rtfs;

    gint	 sort_attr_id;
};

struct _PLManagerPrivate
{
        gboolean       dispose_has_run;
	GHashTable    *playlists;
};

// List creation/deletion
static void
pl_manager_playlist_free (PLManagerPlaylist *playlist)
{
    g_free (playlist->name);
    g_object_unref (playlist->relation);

    g_free (playlist); 
}

gboolean
pl_manager_list_delete (PLManager    *self,
			const gchar  *name)
{
    PLManagerPlaylist *playlist;

    g_return_val_if_fail (PL_IS_MANAGER(self), FALSE);

    if (!g_hash_table_lookup (self->private->playlists, name))
      return FALSE;

    playlist = g_hash_table_lookup (self->private->playlists, name);    

    pl_manager_playlist_free (playlist);
    g_hash_table_remove (self->private->playlists, name);

    g_signal_emit (G_OBJECT(self), signals[PL_MANAGER_SIGNAL_LIST_DELETED], 0, name);

    return TRUE;
}

gboolean
pl_manager_list_new (PLManager	  *self,
		     const gchar  *name,
		     GList	  *uri_list)
{
    PLManagerPlaylist *playlist;
    GList	      *iter_list = uri_list;

    g_return_val_if_fail (PL_IS_MANAGER(self), FALSE);

    if (g_hash_table_lookup (self->private->playlists, name))
      return FALSE;

    playlist = g_new0 (PLManagerPlaylist, 1);
    playlist->sort_attr_id = UNSORTED_ATTR_ID; 
    playlist->relation = create_relation_base_track ();
    playlist->name = g_strdup (name);

    playlist->rtfs = rtfs_open (playlist->relation, BMP_PATH_USER_DIR, name, FALSE); 

    for ( ; iter_list ; iter_list = g_list_next (iter_list))
      { 
	  HTuple *tuple = NULL;
   
	  if (iter_list->data) 
	    tuple = bmp_metadata_cache_get_metadata_noinsert (bmp_metadata_cache, (const gchar*)iter_list->data);

	  if (tuple)
	    h_relation_base_insert_tuple (H_RELATION_BASE(playlist->relation), tuple);
      }
    
    g_hash_table_insert (self->private->playlists, g_strdup (name), playlist);
    g_signal_emit (G_OBJECT(self), signals[PL_MANAGER_SIGNAL_LIST_CREATED], 0, name);

    return TRUE;
}

gboolean
pl_manager_list_new_from_uri (PLManager	   *self,
			      const gchar  *name,
			      const gchar  *uri)
{
    PLManagerPlaylist *playlist;
    VFSHandle	      *handle;
    GList	      *iter_list = NULL;

    g_return_val_if_fail (PL_IS_MANAGER(self), FALSE);

    if (g_hash_table_lookup (self->private->playlists, name))
      return FALSE;

    handle = vfs_handle_new ();  
    handle->uri = g_strdup (uri);

    if (!vfs_handle_read (handle))
      {
	  vfs_handle_free (handle);
	  return FALSE;
      }

    playlist = g_new0 (PLManagerPlaylist, 1);

    playlist->sort_attr_id = UNSORTED_ATTR_ID; 
    playlist->relation = create_relation_base_track ();
    playlist->name = g_strdup (name);
    playlist->rtfs = rtfs_open (playlist->relation, BMP_PATH_USER_DIR, name, FALSE); 

    iter_list = handle->uri_list;

    for ( ; iter_list ; iter_list = g_list_next (iter_list))
      { 
	  HTuple *tuple = NULL;
   
	  if (iter_list->data) 
	    tuple = bmp_metadata_cache_get_metadata_noinsert (bmp_metadata_cache, (const gchar*)iter_list->data);

	  if (tuple)
	    h_relation_base_insert_tuple (H_RELATION_BASE(playlist->relation), tuple);
      }
    
    g_hash_table_insert (self->private->playlists, g_strdup (name), playlist);

    vfs_handle_free (handle);

    return TRUE;
}


//List manipulation
gboolean
pl_manager_list_remove_uris (PLManager	  *self,
			     const gchar  *name,
			     GList	  *uri_list)
{
    PLManagerPlaylist *playlist;
    GList	      *iter_list = uri_list;

    g_return_val_if_fail (PL_IS_MANAGER(self), FALSE);

    if (!g_hash_table_lookup (self->private->playlists, name))
      return FALSE;

    playlist = g_hash_table_lookup (self->private->playlists, name); 

    for ( ; iter_list ; iter_list = g_list_next (iter_list))
      { 
	if (iter_list->data)
	  {
	      h_relation_base_delete_tuplev (H_RELATION_BASE(playlist->relation),
					     bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
					     iter_list->data); 
	  }
      }

    return TRUE;
}

//List accessors
const HRelation*
pl_manager_list_get_relation (PLManager	      *self,
			      const gchar     *name)
{
    PLManagerPlaylist *playlist;

    g_return_val_if_fail (PL_IS_MANAGER(self), FALSE);

    if (!g_hash_table_lookup (self->private->playlists, name))
      return FALSE;

    playlist = g_hash_table_lookup (self->private->playlists, name); 

    if (!playlist)
      return NULL;


    return (const HRelation*)playlist->relation;
}

static void
liststore_fill (HTuple		*tuple,
		gpointer	 data)
{
  HTuple       *tuple_dest;
  GtkListStore *liststore = GTK_LIST_STORE(data);
  GtkTreeIter	iter;

  tuple_dest = h_tuple_clone (tuple); 

  gtk_list_store_append (liststore, &iter);
  gtk_list_store_set (liststore, &iter, 0, tuple_dest, -1);
}

#ifdef HAVE_GUI
GtkListStore*
pl_manager_list_get_model (PLManager	*self,
			   const gchar  *name)
{
    PLManagerPlaylist *playlist;
    GtkListStore      *liststore;

    g_return_val_if_fail (PL_IS_MANAGER(self), FALSE);

    if (!g_hash_table_lookup (self->private->playlists, name))
      return FALSE;

    playlist = g_hash_table_lookup (self->private->playlists, name); 

    if (!playlist)
      return NULL;

    liststore = gtk_list_store_new (1, 
				    G_TYPE_POINTER);

    h_relation_for_each_tuple (playlist->relation, (HTupleFunc) liststore_fill, liststore);

    return liststore;
}
#endif


/* GObject */
static void
free_playlist (gpointer item)
{
}

static void
pl_manager_init (PLManager *self)
{
    GError    *error = NULL;
    GDir      *dir;
    gchar     *basepath; 
    gboolean   run = TRUE;

    self->private = g_new (PLManagerPrivate, 1);
    self->private->dispose_has_run = FALSE;

    self->private->playlists = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, free_playlist);

    basepath = g_build_filename (BMP_PATH_USER_DIR, BMP_PLAYLISTS_DIR_BASENAME, NULL);

    dir = g_dir_open (basepath,
                      0, 
                      &error); 

    if (error)
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Unable to open playlists path %s", basepath);
	g_error_free (error);
	return;
      }
   
    while (run) 
      {
	  const gchar *name;

	  name = g_dir_read_name (dir);
	  if (!name)
	    {
	      run = FALSE;
	      break;
	    }

#if 0
	  if (str_has_suffix_nocase (name, "map"))
	    {
	      gchar  *lcmfsbase;
	      gchar **fragments; 

	      fragments = g_strsplit (name, ".", -1);
  
	      lcmfsbase = g_build_filename (basepath, fragments[0], NULL);
	      list_new_from_lcmfsbase (self, fragments[0], lcmfsbase);
	    }
#endif
      }

    g_dir_close (dir);
}

static void
pl_manager_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
pl_manager_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{


        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
pl_manager_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                PLManagerClass *klass;
                GObjectClass *parent_class;
                klass = PL_MANAGER_CLASS (g_type_class_peek (PL_TYPE_MANAGER));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
close_rtfs (gpointer _k, gpointer _v, gpointer data)
{
    PLManagerPlaylist *playlist = (PLManagerPlaylist*)_v;
    rtfs_close (playlist->rtfs);
}

static void
pl_manager_dispose (GObject *obj)
{
        PLManager *self = (PLManager *)obj;

        if (self->private->dispose_has_run)
                return;

	g_hash_table_foreach (self->private->playlists, (GHFunc)close_rtfs, self); 

        self->private->dispose_has_run = TRUE;
}

static void
pl_manager_finalize (GObject *obj)
{
        PLManager *self = (PLManager *)obj;

        g_free (self->private);
}

PLManager*
pl_manager_new (void)
{
  PLManager *self;

  self = g_object_new (pl_manager_get_type (), NULL);
  return self;
}

static void
pl_manager_class_init (PLManagerClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

        gobject_class->set_property = pl_manager_set_property;
        gobject_class->get_property = pl_manager_get_property;
        gobject_class->dispose = pl_manager_dispose;
        gobject_class->finalize = pl_manager_finalize;
        gobject_class->constructor = pl_manager_constructor;

    signals[PL_MANAGER_SIGNAL_LIST_CREATED] =
      g_signal_new ("list-created",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (PLManagerClass, list_created),
		    NULL, NULL, g_cclosure_marshal_VOID__STRING,
		    G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[PL_MANAGER_SIGNAL_LIST_DELETED] =
      g_signal_new ("list-deleted",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (PLManagerClass, list_deleted),
		    NULL, NULL, g_cclosure_marshal_VOID__STRING,
		    G_TYPE_NONE, 1, G_TYPE_STRING);

}

static void
list_append (gpointer _k, gpointer _v, GList **list)
{
  *list = g_list_append (*list, _k);
}

GList*
pl_manager_get_lists (PLManager *self)
{
    GList *lists = NULL;

    g_hash_table_foreach (self->private->playlists, (GHFunc) list_append, &lists);

    return lists;
}
