/*  BMPx - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  Edward Brocklesby, Chong Kai Xiong,
 *            Milosz Derezynski, David Lau, Michiel Sikkes
 *
 *  ui.c/bmp_ui.h
 *  bmp_button.c/bmp_button.h
 *  bmp_button_toggle.c/bmp_button_toggle.h
 *  bmp_slider.c/bmp_slider.h
 *
 *  M.Derezynski
 *  Tim-Philipp Mueller
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "error.h"
#include "resource_manager.h"

#include "ui.h"

#include "ui_playlist.h"
#include "ui_main.h"

#include "ui_dialog_about.h"
#include "ui_dialog_jtt.h"
#include "ui_dialog_library.h"
#include "ui_dialog_preferences.h"
#include "ui_dialog_streams.h"
#include "ui_dialog_trackinfo.h"

#include "ui_util.h"
#include "ui_callbacks.h"

#include <libskinned/bmp_button.h>
#include <libskinned/bmp_button_toggle.h>
#include <libskinned/bmp_slider.h>
#include <libskinned/bmp_window.h>
#include <libskinned/dock.h>

#include <libchroma/chroma-list.h>

#include <bmpx/util.h>
#include <goa/libgoa.h>

#include "bmp_status_icon.h"
#include "wm.h"

G_DEFINE_TYPE (BmpUI, bmp_ui, G_TYPE_OBJECT)

#include "main.h"
#include "signal_handler_pool.h"

/* GObject stuff */
enum {
    BMP_UI_SIGNAL_0,

    BMP_UI_N_SIGNALS
};

#if 0
static guint signals[BMP_UI_N_SIGNALS] = { 0 };
#endif

enum {
    BMP_UI_PROP_TRAYICON_EMBEDDED = 1,

    BMP_PLAY_N_PROPS
};

struct _BmpUIPrivate {

    gboolean dispose_has_run;

    gboolean		   trayicon_embedded;

    BmpWindowMain         *window_main;
    BmpWindowPlaylist     *window_playlist;

    BmpPreferences	  *preferences;
    BmpTrackInfo          *trackinfo;
    BmpLibrary		  *library;
    BmpRadioStreams	  *streams;
    BmpJumpToTrack        *jtt;

    GdkBitmap             *window_main_mask;

    BmpStatusIcon	  *status_icon;
    GdkPixbuf		  *status_icon_image_default;
    GdkPixbuf		  *status_icon_image_paused;
    GdkPixbuf		  *status_icon_image_playing;
    GdkPixbuf		  *status_icon_image_exit;

    SignalHandlerPool	  *pool;
};

static gchar* bmp_skin_files[] = {
	"main.bmp",
	"shufrep.bmp",
	"cbuttons.bmp",
	"posbar.bmp",
	"volume.bmp",
	"titlebar.bmp",
	"balance.bmp",
	"pledit.bmp",
	"monoster.bmp",
	NULL,
	"playpaus.bmp"
};

static BmpSkinComponent bmp_skin_files_sequential[] = {
	SKIN_COMPONENT_MAIN,
	SKIN_COMPONENT_SHUFREP,
	SKIN_COMPONENT_CBUTTONS,
	SKIN_COMPONENT_TITLEBAR,
	SKIN_COMPONENT_PLEDIT,
	SKIN_COMPONENT_MONOSTER,
	SKIN_COMPONENT_PLAYPAUSE
};

static void
on_action_show_ui_toggled (GtkToggleAction *action, gpointer data);

static void
on_action_keep_above_toggled (GtkToggleAction *action, gpointer data);

static GtkToggleActionEntry bmp_toggle_actions_UI[] = {

  { BMP_TOGGLE_ACTION_SHOW_UI,
	NULL,
	N_("Show UI"),
	NULL,
	N_("Toggle BMPx UI Visiblity"),
	G_CALLBACK (on_action_show_ui_toggled),
	FALSE },

  { BMP_TOGGLE_ACTION_KEEP_ABOVE,
	NULL,
	N_("Keep BMPx on top"),
	NULL,
	N_("Keep BMPx on top"),
	G_CALLBACK (on_action_keep_above_toggled),
	FALSE }

};

/* 'extern' referenced */
GtkUIManager *bmp_ui_manager = NULL;
bmp_posbar_t *bmp_posbar;
bmp_volume_t *bmp_volume;
bmp_volume_t *bmp_balance;
bmp_skin_component_t *bmp_skin_component[N_SKIN_COMPONENTS];
bmp_skin_component_t *bmp_default_skin_component[N_SKIN_COMPONENTS];
GtkWidget    *c_tracklist;
GtkWidget    *c_tracklist_scrollbar;

/* Action groups */
GtkActionGroup *bmp_actions = NULL;
GtkActionGroup *bmp_toggle_actions = NULL;

static void
skin_components_free (BmpUI *self)
{
  gint n;

 /* Free memory */
  for (n = 0; n < N_SKIN_COMPONENTS; n++)
  {
     if (bmp_skin_component[n] != NULL)
     {
	g_object_unref (bmp_skin_component[n]->pixbuf);
	g_free (bmp_skin_component[n]);
	bmp_skin_component[n] = NULL;
     }
  }

  if (bmp_posbar)
	{
		g_object_unref (bmp_posbar->pixbuf);
		g_free (bmp_posbar);
		bmp_posbar = NULL;
	}

  if (bmp_volume)
	{
		g_object_unref (bmp_volume->pixbuf);
		g_free (bmp_volume);
		bmp_volume = NULL;
	}

  if (bmp_balance)
	{
		g_object_unref (bmp_balance->pixbuf);
		g_free (bmp_balance);
		bmp_balance = NULL;
	}
}

static void
load_skin_components(BmpUI *self, const gchar * archive_filename, gint type, bmp_skin_component_t * (*skin_components)[N_SKIN_COMPONENTS])
{
  GdkPixbuf *pixbuf;
  const gchar *archive_dirname;
  gchar *filename;
  gint n;

  /* Free memory */
  skin_components_free (self);
  bmp_cursors_free ();

  /* LOAD COMPONENTS */
  if (type == BMP_SKIN_ARCHIVE)
  	archive_dirname = archive_decompress (archive_filename);
  else
  	archive_dirname = archive_filename;

  bmp_cursors_set (archive_dirname);

  /* Buttons, Titlebars */
  for (n = 0; n < G_N_ELEMENTS(bmp_skin_files_sequential); n++)
    {
       filename = find_file_recursively(archive_dirname, bmp_skin_files[bmp_skin_files_sequential[n]], 1);
      if (!filename)
	{
        	g_message("%s not found", bmp_skin_files[bmp_skin_files_sequential[n]]);
	}
      else
        {
	      pixbuf = gdk_pixbuf_new_from_file(filename, NULL);

	      (*skin_components)[bmp_skin_files_sequential[n]] = g_malloc0(sizeof(bmp_skin_component_t));
	      (*skin_components)[bmp_skin_files_sequential[n]]->identifier = bmp_skin_files_sequential[n];
	      (*skin_components)[bmp_skin_files_sequential[n]]->pixbuf = pixbuf;

	      g_free(filename);
        }
    }

  /* Numbers */
  (*skin_components)[SKIN_COMPONENT_NUMBERS] = g_malloc0(sizeof(bmp_skin_component_t));
  filename = find_file_recursively(archive_dirname, "numbers.bmp", 1);
  if (!filename)
	filename = find_file_recursively(archive_dirname, "nums_ex.bmp", 1);

  if (!filename)
    {
	g_message("%s: Numbers not found", G_STRFUNC);
	g_free((*skin_components)[SKIN_COMPONENT_NUMBERS]);
    }
  else 
    {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	(*skin_components)[SKIN_COMPONENT_NUMBERS]->identifier = SKIN_COMPONENT_NUMBERS;
	(*skin_components)[SKIN_COMPONENT_NUMBERS]->pixbuf = pixbuf;

	if (gdk_pixbuf_get_width(pixbuf) >= 108)
	   (*skin_components)[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (12);
	else
	   (*skin_components)[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (10);

	g_free(filename);
    } 

  /* Volume, Balance and Seek slider*/
  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_POSBAR], 1);
  if (!filename)
    {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_POSBAR]);
    }
  else
   {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_posbar = g_malloc0(sizeof(bmp_posbar_t));
	bmp_posbar->pixbuf = pixbuf;
	g_free(filename);
   }

  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_VOLUME], 1);
  if (!filename)
    {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_VOLUME]);
    }
  else
    {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_volume = g_malloc0(sizeof(bmp_volume_t));
	bmp_volume->pixbuf = pixbuf;
	bmp_volume->height = gdk_pixbuf_get_height(pixbuf) - 422;

	if (gdk_pixbuf_get_height(pixbuf) > 422)
	    bmp_volume->has_button = TRUE;
	else
	    bmp_volume->has_button = FALSE;
	g_free(filename);
    }

  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_BALANCE], 1);
  if (!filename)
    {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_BALANCE]);
    }
  else
    {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_balance = g_malloc0(sizeof(bmp_volume_t));
	bmp_balance->pixbuf = pixbuf;
	bmp_balance->height = gdk_pixbuf_get_height(pixbuf) - 421;
	if (gdk_pixbuf_get_height(pixbuf) > 421)
	    bmp_balance->has_button = TRUE;
	else
	    bmp_balance->has_button = FALSE;
	g_free(filename);
    }

  bmp_colors_init(archive_dirname);

  /* FIXME: Move this into ui_main.c */
  self->private->window_main_mask = bmp_get_mask_window_main (archive_dirname);

  if (type == BMP_SKIN_ARCHIVE)
    {
  	del_directory(archive_dirname);
    } 
}

GdkBitmap*
bmp_ui_get_mask_window_main (BmpUI *self)
{
    g_return_val_if_fail (BMP_IS_UI (self), NULL);

    return self->private->window_main_mask;
}

void
bmp_ui_display_windows(BmpUI *self) {

    gint x,y,w,h;
    gboolean visible;

    x = xcs_key_fetch_int (xcs, "main-window", "pos_x");
    y = xcs_key_fetch_int (xcs, "main-window", "pos_y");
    if ((x > -1) && (y > -1))
    {
	gtk_window_move (GTK_WINDOW(self->private->window_main->window), x, y);
	BMP_WINDOW (self->private->window_main->window)->x = x;
	BMP_WINDOW (self->private->window_main->window)->y = y;
    }
    else
    {
	x = gdk_screen_get_width (gdk_screen_get_default());
	y = gdk_screen_get_height (gdk_screen_get_default());

	x /= 2;
	y /= 2;

	x -= BMP_WINDOW_MAIN_WIDTH/2;
	y -= BMP_WINDOW_MAIN_HEIGHT/2;

	gtk_window_move (GTK_WINDOW(self->private->window_main->window), x, y);
	BMP_WINDOW (self->private->window_main->window)->x = x;
	BMP_WINDOW (self->private->window_main->window)->y = y;
    }
    gtk_widget_show_all (BMP_WINDOW_MAIN(self->private->window_main)->window);

    x = xcs_key_fetch_int (xcs, "playlist-window", "pos_x");
    y = xcs_key_fetch_int (xcs, "playlist-window", "pos_y");

    if ((x > -1) && (y > -1))
    {
	gtk_window_move (GTK_WINDOW(self->private->window_playlist->window), x, y);
	BMP_WINDOW (self->private->window_playlist->window)->x = x;
	BMP_WINDOW (self->private->window_playlist->window)->y = y;
    }
    else
    {
	x = gdk_screen_get_width (gdk_screen_get_default());
	y = gdk_screen_get_height (gdk_screen_get_default());

	x /= 2;
	y /= 2;

	x -= BMP_WINDOW_MAIN_WIDTH;
	y -= BMP_WINDOW_MAIN_HEIGHT;

	gtk_window_move (GTK_WINDOW(self->private->window_playlist->window), x, y);
	BMP_WINDOW (self->private->window_playlist->window)->x = x;
	BMP_WINDOW (self->private->window_playlist->window)->y = y;
    }

    w = xcs_key_fetch_int (xcs, "playlist-window", "width");
    h = xcs_key_fetch_int (xcs, "playlist-window", "height");

    w = (w < 0) ? 10*25 : w;
    h = (h < 0) ? 10*29 : h;

#if 0
    gtk_widget_set_size_request (GTK_WINDOW(self->private->window_playlist->window), w, h); 
#endif
    gtk_window_set_default_size (GTK_WINDOW(self->private->window_playlist->window), w, h); 

    visible = xcs_key_fetch_boolean (xcs,"playlist-window","visible");

    if (visible)
	bmp_system_control_show_pl_window (bmp_system_control, TRUE);
}

static gint
load_skin (BmpUI *self, const gchar * filename, bmp_skin_component_t * (*skin_components)[N_SKIN_COMPONENTS])
{
    gint error_status;

    if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
	error_status = BMP_ERROR_FAILURE;
	return error_status;
    }

    bmp_colors_free();

    if (g_file_test(filename, G_FILE_TEST_IS_DIR)) {
        load_skin_components(self, filename, BMP_SKIN_DIR, skin_components);
	error_status = BMP_ERROR_SUCCESS;
    }
    else
    if (g_file_test(filename, G_FILE_TEST_IS_REGULAR)) {
        load_skin_components(self, filename, BMP_SKIN_ARCHIVE, skin_components);
	error_status = BMP_ERROR_SUCCESS;
    }
    else
    {
        error_status = BMP_ERROR_FAILURE;
	return error_status;
    }

    if (self->private->window_main)
        bmp_window_main_configure (BMP_WINDOW_MAIN (self->private->window_main));

    if (self->private->window_playlist)
	bmp_window_playlist_configure (BMP_WINDOW_PLAYLIST (self->private->window_playlist));

    return error_status;
}

static void
on_bmp_skin_changed (GObject *configuration, gpointer data)
{
    if (load_skin (BMP_UI(data), xcs_key_fetch (xcs, "bmp", "skin"), &bmp_skin_component) != BMP_ERROR_SUCCESS) 
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Unable to load skin '%s'", xcs_key_fetch (xcs, "bmp", "skin"));
      }
}

static void
on_system_control_set_title (BmpSystemControl *control, BmpUI *self)
{
     gchar *title, *window_title;

     g_return_if_fail (BMP_IS_UI(self));

     title = g_object_get_string (bmp_system_control, "current_title");

     if (!title)
       {
	window_title = g_strdup_printf ("BMPx");
       }
     else
       {
	window_title = g_strdup_printf ("BMPx: %s", title);
       }

     gtk_window_set_title (GTK_WINDOW(self->private->window_main->window), window_title);

     g_free (window_title);
     g_free (title);
}
   
void
bmp_ui_set_trayicon_exit (BmpUI *self)
{
    g_return_if_fail (BMP_IS_UI(self));

    bmp_status_icon_set_blinking (self->private->status_icon, TRUE);
}

static void
on_tray_icon_click (BmpStatusIcon *icon, gpointer data)
{
#if 0
    BmpUI *ui = BMP_UI(data);
#endif
    GtkToggleAction *action;
    gboolean active;
    
    action = GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_SHOW_UI));
    active = gtk_toggle_action_get_active (action);

#if 0
    if ((active) &&
	(gdk_window_get_state (GTK_WIDGET(ui->private->window_main->window)->window) & GDK_WINDOW_STATE_ICONIFIED))
      {
	gdk_window_show (GTK_WIDGET(ui->private->window_main->window)->window);
      }
    else
      {
#endif
	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(action), !active);
#if 0
      } 
#endif
}

static void
on_tray_icon_scroll_up (BmpStatusIcon *icon, gpointer data)
{
    gint volume;

    volume = g_object_get_int (bmp_system_control, "volume");
    volume += 4;
    volume = (volume > 100) ? 100 : volume;
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
}

static void
on_tray_icon_scroll_down (BmpStatusIcon *icon, gpointer data)
{
    gint volume;

    volume = g_object_get_int (bmp_system_control, "volume");
    volume -= 4;
    volume = (volume < 0) ? 0 : volume;
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
}


static void
on_tray_icon_popup_menu (BmpStatusIcon *icon, guint arg1, guint arg2, gpointer data)
{
    GtkWidget *menu;

    menu = ui_manager_get_popup (bmp_ui_manager, "/popup-tray/menu-tray");
    gtk_widget_realize (menu);
    gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, arg1, arg2);
}

static void
set_status_playstatus (BmpUI *self)
{
  gint status;

  status = g_object_get_int (bmp_play_engine, "playstatus");

  if ((status != PLAYSTATUS_PLAYING) && (status != PLAYSTATUS_PAUSED))
      {
	    bmp_status_icon_set_from_pixbuf (self->private->status_icon, 
					     self->private->status_icon_image_default);
      }
  else if (status == PLAYSTATUS_PAUSED)
      {
	    bmp_status_icon_set_from_pixbuf (self->private->status_icon, 
					     self->private->status_icon_image_paused);
      }
  else if (status == PLAYSTATUS_PLAYING)
      {
	    bmp_status_icon_set_from_pixbuf (self->private->status_icon, 
					     self->private->status_icon_image_playing);
      }
}

static void
on_system_control_set_playstatus (BmpSystemControl *system_control, gint status, gpointer data) 
{
    g_return_if_fail (BMP_IS_UI(data));

    set_status_playstatus (BMP_UI(data));
}

static void
on_action_keep_above_toggled (GtkToggleAction *action, gpointer data)
{
    wm_set_keep_above (gtk_toggle_action_get_active (action));
}

static void
on_action_show_ui_toggled (GtkToggleAction *action, gpointer data)
{
    gboolean active;
    static gboolean playlist_visible = FALSE;
    static gint x = -1 , y = -1;

    active = gtk_toggle_action_get_active (action);

    if (active)
      {
	if (x != -1) gtk_window_move (GTK_WINDOW(bmp_ui->private->window_main->window), x, y);

	gtk_widget_show (bmp_ui->private->window_main->window);

	if (playlist_visible) gtk_widget_show (bmp_ui->private->window_playlist->window);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             FALSE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             FALSE); 

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             FALSE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             FALSE); 
	bmp_window_playlist_resize_widgets (bmp_ui->private->window_playlist);
      }
    else
      {
	gtk_window_get_position (GTK_WINDOW(bmp_ui->private->window_main->window), &x, &y);
	gtk_widget_hide (GTK_WIDGET(bmp_ui->private->window_main->window));

	playlist_visible = GTK_WIDGET_VISIBLE(bmp_ui->private->window_playlist->window);
	if (playlist_visible)
	    gtk_widget_hide (bmp_ui->private->window_playlist->window);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             TRUE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             TRUE); 

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             TRUE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             TRUE); 
      }
}


static gint
ui_initialize (BmpUI *self)
{
  GError *error = NULL;
  const gchar *filename; 
  gchar *trayicon_ui;
  gint rv, n;

  if ((self->private->window_playlist != NULL) && (self->private->window_main != NULL)) return BMP_ERROR_SUCCESS;

	for (n = 0; n < N_SKIN_COMPONENTS; bmp_skin_component[n] = bmp_default_skin_component[n] = NULL, n++);

  bmp_cursors_init();

	filename = DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN;
  rv = load_skin (self, filename, &bmp_default_skin_component);
	if (rv != BMP_ERROR_SUCCESS) {
		g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Couldn't load default skin.", G_STRLOC);
		exit(-1);
	}

  filename = xcs_key_fetch(xcs, "bmp", "skin");
  rv = load_skin (self, filename, &bmp_skin_component);
	if (rv != BMP_ERROR_SUCCESS) {
	  int x;
	xcs_key_set_string (xcs, "bmp","skin", DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN);
	for (x = 0; x < N_SKIN_COMPONENTS; bmp_skin_component[x] = bmp_default_skin_component[x], x++);
  }


  if (rv == BMP_ERROR_SUCCESS)
    {

      self->private->pool = signal_handler_pool_new ("Ui");

      signal_handler_pool_add_connection (self->private->pool,
					  xcs,	
					  "changed::bmp-skin",	
					  G_CALLBACK (on_bmp_skin_changed),	
					  self,
					  0);	

      bmp_ui_manager = gtk_ui_manager_new();

      bmp_resource_manager_create_path (bmp_rm, NULL, "dialogs");
      bmp_resource_manager_create_path (bmp_rm, NULL, "windows");
      bmp_resource_manager_create_path (bmp_rm, NULL, "tooltips");

      bmp_actions = gtk_action_group_new(BMP_MAIN_ACTION_GROUP);
      bmp_toggle_actions = gtk_action_group_new(BMP_MAIN_ACTION_GROUP_TOGGLE);

      gtk_action_group_set_translate_func (bmp_actions,
		bmp_menu_translate,
		NULL,
		NULL);
      gtk_action_group_set_translate_func (bmp_toggle_actions,
		bmp_menu_translate,
		NULL,
		NULL);

      gtk_ui_manager_insert_action_group (bmp_ui_manager, bmp_actions, 0);
      gtk_ui_manager_insert_action_group (bmp_ui_manager, bmp_toggle_actions, 0);
      gtk_action_group_set_sensitive (bmp_actions, TRUE);
      gtk_action_group_set_sensitive (bmp_toggle_actions, TRUE);

      gtk_action_group_add_toggle_actions (bmp_toggle_actions, bmp_toggle_actions_UI, G_N_ELEMENTS (bmp_toggle_actions_UI), self);


      /* Dialogs */
      self->private->preferences = bmp_preferences_new ();
      RM_ITEM_NEW("dialogs", 
		"dialog-preferences", 
		 self->private->preferences,
		 NULL);

      self->private->trackinfo = bmp_trackinfo_new ();
      RM_ITEM_NEW("dialogs", 
		"dialog-trackinfo", 
		 self->private->trackinfo,
		 NULL);

      self->private->jtt = bmp_jtt_new ();
      RM_ITEM_NEW("dialogs", 
		"dialog-jtt", 
		 self->private->jtt,
		 NULL);

      self->private->library = bmp_library_new ();
      RM_ITEM_NEW("dialogs", 
		"dialog-library", 
		 self->private->library,
		 NULL);

      self->private->streams = bmp_streams_new ();
      RM_ITEM_NEW("dialogs", 
		"dialog-streams", 
		 self->private->streams,
		 NULL);


      /* Main windows */
      self->private->window_main = bmp_window_main_new();
      RM_ITEM_NEW("windows", 
		"window-main", 
		self->private->window_main,
		NULL);

      self->private->window_playlist = bmp_window_playlist_new();
      RM_ITEM_NEW("windows", 
	        "window-playlist", 
		self->private->window_playlist,
		NULL);

      /* Set up configuration bindings */
      xcs_bind_toggle_action (xcs, 
			    GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_PL)),
			    "playlist-window",
			    "visible");

      bmp_window_main_configure (self->private->window_main);
      bmp_window_playlist_configure (self->private->window_playlist);

      bmp_ui_display_windows (self);

      bmp_system_control_repeat_set (bmp_system_control, 
				   xcs_key_fetch_boolean (xcs, "bmp", "repeat"), 
				   NULL);

      bmp_system_control_shuffle_set (bmp_system_control, 
				    xcs_key_fetch_boolean (xcs, "bmp", "shuffle"), 
				    NULL);

      /* Init Tray Icon */
      self->private->status_icon_image_default =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-default.png",
				  NULL);
      self->private->status_icon_image_paused =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-paused.png",
				  NULL);
      self->private->status_icon_image_playing =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-playing.png",
				  NULL);
      self->private->status_icon_image_exit =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-exit.png",
				  NULL);
      self->private->status_icon = 
	bmp_status_icon_new_from_pixbuf (self->private->status_icon_image_default);
      bmp_status_icon_set_visible (self->private->status_icon, TRUE);

      g_object_connect (G_OBJECT (self->private->status_icon),
		    "signal::click",
		    G_CALLBACK(on_tray_icon_click),
		    self,
		    "signal::popup-menu",
		    G_CALLBACK(on_tray_icon_popup_menu),
		    self,
		    "signal::scroll-up",
		    G_CALLBACK(on_tray_icon_scroll_up),
		    self,
		    "signal::scroll-down",
		    G_CALLBACK(on_tray_icon_scroll_down),
		    self,
		    NULL);

      trayicon_ui = g_build_filename (DATA_DIR, BMP_UI_BASENAME, "trayicon.ui", NULL);

      if (!gtk_ui_manager_add_ui_from_file (bmp_ui_manager, trayicon_ui, &error))
	{
		g_message ("%s: Building menus %s failed: %s", G_STRLOC, trayicon_ui, error->message);
		g_error_free (error);
		error = NULL;
	}

      g_free (trayicon_ui);

      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(gtk_action_group_get_action 
						(bmp_toggle_actions, BMP_TOGGLE_ACTION_SHOW_UI)), TRUE);

      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(gtk_action_group_get_action 
						(bmp_toggle_actions, BMP_TOGGLE_ACTION_KEEP_ABOVE)),
						 xcs_key_fetch_boolean (xcs,"bmp","keep-above"));

      set_status_playstatus (self);

      signal_handler_pool_add_connection (self->private->pool,
					  bmp_system_control,	
					  "set-playstatus",
					  G_CALLBACK (on_system_control_set_playstatus),	
					  self,
					  0);	

      signal_handler_pool_add_connection (self->private->pool,
					  bmp_system_control,	
					  "set-title",
					  G_CALLBACK (on_system_control_set_title),	
					  self,
					  0);	
    }

  dialog_about_setup ();

  self->private->trayicon_embedded = bmp_status_icon_is_embedded (self->private->status_icon);

  return rv;
}

/* GObject stuff */
static void
bmp_ui_dispose (GObject *obj)
{
    BmpUI *self = BMP_UI (obj);

    if (self->private->dispose_has_run) return;

    self->private->dispose_has_run = TRUE;

#if 0
    lv_bmp_stop ();
#endif

#if 0
    bookmarks_save ();
#endif

    skin_components_free (self);
    dock_window_clear_list ();

    g_object_unref (self->private->window_main);
    g_object_unref (self->private->window_playlist);
    g_object_unref (self->private->preferences);
    g_object_unref (self->private->library);
    g_object_unref (self->private->trackinfo);
    g_object_unref (self->private->jtt);
    g_object_unref (self->private->streams);
    g_object_unref (self->private->status_icon);

    g_object_unref (bmp_ui_manager);

    dialog_about_destroy ();

    signal_handler_pool_free (self->private->pool);

/*    gdk_display_close (gdk_display_get_default()); */

    g_message("stopped: %s", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_ui_finalize (GObject *obj)
{
}
 
static void
bmp_ui_get_property (GObject    *object,
                     guint       property_id,
                     GValue     *value,
                     GParamSpec *pspec)
{
  BmpUI *self = (BmpUI*) object;

  switch (property_id)
    {
	case BMP_UI_PROP_TRAYICON_EMBEDDED:
	      g_value_set_boolean (value, self->private->trayicon_embedded);
	      break;

	default:
	      g_assert (FALSE);
    }
}


static void
bmp_ui_class_init (BmpUIClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->get_property = bmp_ui_get_property;
#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->constructor = bmp_play_constructor;
#endif

    g_object_class_install_property (gobject_class,
				     BMP_UI_PROP_TRAYICON_EMBEDDED,
				     g_param_spec_boolean ("trayicon-embedded",
                                                           "trayicon-embedded",
                                                           "Whether tray icon is embedded or not", FALSE,
							    G_PARAM_READABLE));

    gobject_class->dispose  = bmp_ui_dispose;
    gobject_class->finalize = bmp_ui_finalize;
}

static void
bmp_ui_init (BmpUI *self)
{
    self->private = g_new0 (BmpUIPrivate, 1);

    self->private->window_main = NULL;
    self->private->window_playlist = NULL;

    self->private->preferences = NULL;
    self->private->trackinfo = NULL;

    self->private->window_main_mask = NULL;
}

BmpUI*
bmp_ui_new (void)
{
    BmpUI *self;

    GtkSettings *settings;
    const gchar *name;

    self = g_object_new (BMP_TYPE_UI, NULL);

    /* Hack-around for no xsettings manager available that would set the cursor theme */
    settings = gtk_settings_get_for_screen (gdk_screen_get_default());
    name = g_object_get_string (settings, "gtk-cursor-theme-name"); 
    if (!name)
      {
	  name = g_getenv("XCURSOR_THEME");
	  g_object_set (settings, "gtk-cursor-theme-name", name, NULL);
      }

    return self;
}

gint
bmp_ui_finalize_init (BmpUI *self)
{
    gint rv;

    rv = ui_initialize(self);

    if (rv == BMP_ERROR_SUCCESS)
	    g_message("running: %s", G_OBJECT_TYPE_NAME(self));
    else
	    g_message("%s Not initalized: Skin couldn't be loaded", __FILE__);

    return rv;
}

void
bmp_ui_set_trayicon_text (BmpUI *self, const gchar *text)
{
    const gchar *stock_id;
    gint playstatus;

    g_return_if_fail (BMP_IS_UI(self));

    playstatus = g_object_get_int (bmp_play_engine, "playstatus");

    switch (playstatus)
     {
	case PLAYSTATUS_STOPPED: stock_id = GTK_STOCK_MEDIA_STOP; break;
	case PLAYSTATUS_PAUSED:  stock_id = GTK_STOCK_MEDIA_PAUSE; break;
	case PLAYSTATUS_PLAYING: stock_id = GTK_STOCK_MEDIA_PLAY; break;
	default: stock_id = GTK_STOCK_MEDIA_STOP; break;
     }

    bmp_status_icon_set_tooltip (self->private->status_icon, text, stock_id);    
}

void
bmp_ui_jtt_show		    (BmpUI *self)
{
    bmp_jtt_show      (self->private->jtt);
}

void
bmp_ui_library_show	    (BmpUI *self)
{
    bmp_library_show  (self->private->library);
}

void
bmp_ui_streams_show	    (BmpUI *self)
{
    bmp_streams_show  (self->private->streams);
}

void
bmp_ui_raise_windows	    (BmpUI *self)
{
    GtkToggleAction *action;
    
    action = GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_SHOW_UI));
    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(action), TRUE); 

    gtk_window_present (GTK_WINDOW(self->private->window_playlist->window));
    gtk_window_present (GTK_WINDOW(self->private->window_main->window));
}
