/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <stdlib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>
#include <string.h>

#include "system_control.h"
#include "resource_manager.h"

#include "ui.h"
#include "ui_util.h"

#include "main.h"
#include <bmpx/dbus.h>
#include <bmpx/util.h>
#include <bmpx/uri.h>
#include <bmpx/playlist.h>
#include <bmpx/metadata.h>

#include "xcs/xml.h"
#include "xcs/xcs.h"
#include "goa/libgoa.h"

#include "glade.h"
#include "loader.h"
#include "dnd.h"
#include "wm.h"

#include "embedded-images/media-pixbufs.h"

#include "ui_dialog_library.h"
#include "signal_handler_pool.h"

/* GObject stuff */
G_DEFINE_TYPE (BmpLibrary, bmp_library, G_TYPE_OBJECT)

struct _BmpLibraryPrivate {

    gboolean	       dispose_has_run;

    GladeXML	      *xml;
    SignalHandlerPool *pool;

    GtkTreeModel      *tree_store_browse_filter; 
    GtkTreeModel      *tree_store_browse; 
    GtkTreeModel      *tree_store_tags;

    GHashTable	      *map_root_browse;
    GHashTable	      *map_root_tags;
    GHashTable	      *map_root_playlists; 

    GdkPixbuf	      *pb[BMP_N_TAGS];
    GdkPixbuf	      *pb_starred[BMP_N_TAGS];

    gboolean	       shown; 
    gboolean	       running;

    gboolean clear;
    gboolean start_playback;
    gint     start_playback_track;
};

static void
window_library_create (BmpLibrary *self);

BmpLibrary*
bmp_library_new (void)
{
    BmpLibrary *self;

    self = g_object_new (BMP_TYPE_LIBRARY, NULL);
    return self;
}

static GObject*
bmp_library_constructor (GType type,
	                   guint n_construct_properties,
         	           GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        BmpLibraryClass *klass;
        GObjectClass *parent_class;
        klass = BMP_LIBRARY_CLASS (g_type_class_peek (BMP_TYPE_LIBRARY));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    return obj;
}

static void
bmp_library_dispose (GObject *object)
{
    BmpLibrary *self = (BmpLibrary *)object;

    if (self->private->tree_store_browse)
      g_object_unref (self->private->tree_store_browse);

    if (self->private->tree_store_browse_filter)
      g_object_unref (self->private->tree_store_browse_filter);

    if (self->private->tree_store_tags)
      g_object_unref (self->private->tree_store_tags);

    if (self->private->map_root_browse)
      g_hash_table_destroy (self->private->map_root_browse);

    if (self->private->map_root_tags)
      g_hash_table_destroy (self->private->map_root_tags);

    if (self->private->map_root_playlists)
      g_hash_table_destroy (self->private->map_root_playlists); 

    signal_handler_pool_free (self->private->pool);
    g_object_unref (self->private->xml);
    gtk_widget_destroy (self->window);
}

static void
bmp_library_class_init (BmpLibraryClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
#endif
    gobject_class->dispose	= bmp_library_dispose;
    gobject_class->constructor	= bmp_library_constructor;

}

static void
bmp_library_init (BmpLibrary *self)
{
    self->private = g_new (BmpLibraryPrivate, 1);

    self->private->dispose_has_run = FALSE;
    self->private->pool = signal_handler_pool_new ("UiLibrary");

    self->private->tree_store_browse	    = NULL; 
    self->private->tree_store_browse_filter = NULL; 
    self->private->map_root_browse	    = NULL;

    self->private->tree_store_tags	    = NULL;
    self->private->map_root_tags	    = NULL;

    self->private->map_root_playlists	    = NULL; 

    self->private->shown		= FALSE; 
    self->private->running		= FALSE; 

    self->private->clear		= FALSE;
    self->private->start_playback	= FALSE;
    self->private->start_playback_track = 0; 

    self->private->pb[0] = gdk_pixbuf_new_from_inline (-1, icon_artist, FALSE, NULL);
    self->private->pb[1] = gdk_pixbuf_new_from_inline (-1, icon_album,  FALSE, NULL);
    self->private->pb[2] = gdk_pixbuf_new_from_inline (-1, icon_title,  FALSE, NULL);
    self->private->pb_starred[0] = gdk_pixbuf_new_from_inline (-1, icon_artist_starred, FALSE, NULL);
    self->private->pb_starred[1] = gdk_pixbuf_new_from_inline (-1, icon_album_starred,  FALSE, NULL);
    self->private->pb_starred[2] = gdk_pixbuf_new_from_inline (-1, icon_title_starred,  FALSE, NULL);

    window_library_create (self);
}

static GtkTargetEntry dnd_entries[] = {
    {"application/x-bmpx-db-query", 0, BMP_DROP_QUERY},
};

enum {
      BROWSE_COL_IMG,
      BROWSE_COL_KEY,
      BROWSE_COL_TUPLE,
      BROWSE_COL_MAP,
      BROWSE_COL_TRACK,
      BROWSE_COL_LENGTH,
      BROWSE_COL_GENRE,
      BROWSE_COL_BITRATE,
};

static gint browse_levels[] = {
    BMP_DATUM_ARTIST,
    BMP_DATUM_ALBUM,
    BMP_DATUM_TITLE
};

enum {
    TAG_COL_PIXBUF, 
    TAG_COL_NAME,
    TAG_COL_MAP,
    TAG_COL_TUPLE,
    TAG_COL_DATUM
};

static void
ui_dialog_library_playlist_new_from_file (GtkAction	  *action,
					  BmpLibrary	  *self);

static GtkActionEntry bmp_actions_LIBRARY[] = {
  { "dummy-plmanager", NULL, "dummy" },

  { BMP_ACTION_LIBRARY_PLAYLIST_NEW_FROM_FILE,
    GTK_STOCK_OPEN,
    N_("New Playlist from File"), "",
    N_("New Playlist from File"),
    G_CALLBACK(ui_dialog_library_playlist_new_from_file)},

  { BMP_ACTION_LIBRARY_PLAYLIST_NEW_FROM_LIBRARY,
    GTK_STOCK_FIND,
    N_("New Playlist from Library"), "",
    N_("New Playlist from Library"),
    G_CALLBACK(NULL)}
};

static gboolean
tree_store_browse_visible_func (GtkTreeModel	*model,
                                GtkTreeIter	*iter,
                                BmpLibrary	*self);
	
static void
browse_view_insert_tuple (BmpMetadataCache	  *cache,
			  HTuple		  *tuple,
			  BmpLibrary		  *self)
{
    HTupleElem	  *elem;
    GtkTreeIter	   iter[BMP_N_TAGS];
    GHashTable*	   map[BMP_N_TAGS];
    GHashTable*	   map_c;
    gint	   n;
    gchar	  *key;
    gchar	  *filename;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

    if (!g_uri_test (elem->value, G_URI_TEST_PROTOCOL, PROTOCOL_FILE)) return;

    filename = g_filename_from_uri (elem->value, NULL, NULL);
    if (!g_file_test (filename, G_FILE_TEST_EXISTS))
	{
	    g_free (filename);
	    return;
	}
    g_free (filename);

    map[0] = self->private->map_root_browse;

    // FIXME: We have to get rid of the semi-hardcoding
    map_c = map[0];

    for (n = 0; n < BMP_N_TAGS; n++)
      {
	elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[n])); 

	if ((elem->value) && (strlen(elem->value)>0))
	  {
	    key = g_strdup (elem->value);
	    g_strstrip (key);
	  }
	else
	  {
	    key = g_strdup (bmp_metadata_get_default_char (n));
	  }

	//FIXME: We need to distinguish by location here, but the again it's information the user doesn't see (RFC)
	if (!attr_row_map_mapping_exists (map_c, key))
          {	
	      if (n == 0)
		  gtk_tree_store_append (GTK_TREE_STORE(self->private->tree_store_browse), &iter[n], NULL); 
	      else
		  gtk_tree_store_append (GTK_TREE_STORE(self->private->tree_store_browse), &iter[n], &iter[n-1]); 

	      gtk_tree_store_set (GTK_TREE_STORE(self->private->tree_store_browse), 
			     &iter[n], 
			     BROWSE_COL_IMG, 
			     self->private->pb[n], 
			     BROWSE_COL_KEY, 
			     key, 
			     BROWSE_COL_TUPLE, 
			     h_tuple_clone (tuple), 
			     BROWSE_COL_MAP, 
			     attr_row_map_new (),
			     -1); 

	      attr_row_map_mapping_insert (map_c, key, GTK_TREE_MODEL(self->private->tree_store_browse), &iter[n]);
	  }
	else if (!attr_row_map_mapping_get (map_c, key, GTK_TREE_MODEL(self->private->tree_store_browse), &iter[n])) 
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Mapping for key '%p' disappeared", G_STRLOC, key); 
	    g_free (key);
	    break;
	  }

	g_free (key);
	gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_browse), &iter[n], BROWSE_COL_MAP, &map_c, -1); 
    }
}

G_LOCK_DEFINE(DELETE_TUPLE);

static void
browse_view_delete_tuple (BmpMetadataCache	  *cache,
			  HTuple		  *tuple,
			  BmpLibrary		  *self)
{
    HTupleElem	  *elem;
    GtkTreeIter	   iter[BMP_N_TAGS];
    GHashTable*	   map[BMP_N_TAGS];
    gchar	  *key[BMP_N_TAGS];

    G_LOCK (DELETE_TUPLE);

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

    if (g_ascii_strncasecmp (elem->value, "file:", 5))
	{
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    map[0] = self->private->map_root_browse;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[0]));
    key[0] = g_strdup (elem->value);
    g_strstrip (key[0]);
    if (!attr_row_map_mapping_get (map[0], key[0], self->private->tree_store_browse, &iter[0]))
	{
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_browse), &iter[0], BROWSE_COL_MAP, &map[1], -1); 

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[1]));
    key[1] = g_strdup (elem->value);
    g_strstrip (key[1]);

    if (!attr_row_map_mapping_get (map[1], key[1], self->private->tree_store_browse, &iter[1]))
	{
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_browse), &iter[1], BROWSE_COL_MAP, &map[2], -1); 

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[2]));
    key[2] = g_strdup (elem->value);
    g_strstrip (key[2]);

    if (!attr_row_map_mapping_get (map[2], key[2], self->private->tree_store_browse, &iter[2]))
	{
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    gtk_tree_store_remove (GTK_TREE_STORE(self->private->tree_store_browse), &iter[2]);
    attr_row_map_mapping_remove_nodestroy (map[2], key[2]);

    if (gtk_tree_model_iter_n_children (self->private->tree_store_browse, &iter[1]) == 0)
      {
	attr_row_map_mapping_remove_nodestroy (map[1], key[1]);
	gtk_tree_store_remove (GTK_TREE_STORE(self->private->tree_store_browse), &iter[1]);
      }

    if (gtk_tree_model_iter_n_children (GTK_TREE_MODEL(self->private->tree_store_browse), &iter[0]) == 0)
      {
	attr_row_map_mapping_remove_nodestroy (map[0], key[0]);
	gtk_tree_store_remove (GTK_TREE_STORE(self->private->tree_store_browse), &iter[0]);
      }

    G_UNLOCK (DELETE_TUPLE);
}

static void
browse_view_insert_tuple_proxy_func (HTuple		  *tuple,
			       BmpLibrary	  *self)	      
{
    browse_view_insert_tuple (bmp_metadata_cache, tuple, self);
    while (gtk_events_pending()) gtk_main_iteration ();
}

static gint
tree_store_browse_sort_func (GtkTreeModel *model,
			     GtkTreeIter  *a,
			     GtkTreeIter  *b,
			     gpointer      user_data)
{
    HTuple *tuple_a, *tuple_b;
    GtkTreePath *path_a, *path_b;
    gchar *key_a, *_key_a;
    gchar *key_b, *_key_b;
    gint cmp_result = 0;

    gtk_tree_model_get (model, a, BROWSE_COL_KEY, &_key_a, BROWSE_COL_TUPLE, &tuple_a, -1);
    gtk_tree_model_get (model, b, BROWSE_COL_KEY, &_key_b, BROWSE_COL_TUPLE, &tuple_b, -1);

    key_a = g_utf8_casefold (_key_a, strlen (_key_a));
    key_b = g_utf8_casefold (_key_b, strlen (_key_b));

    path_a =
      gtk_tree_model_get_path (model, a);

    path_b =
      gtk_tree_model_get_path (model, b);

    if (path_b && (gtk_tree_path_get_depth (path_b) == 3))
      {
	HTupleElem *elem_a, *elem_b;
	gint track_a, track_b;

	if ((!tuple_a) || (!tuple_b))
	  return 0;

	elem_a = h_tuple_get (tuple_a, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	elem_b = h_tuple_get (tuple_b, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	track_a = GPOINTER_TO_INT (elem_a->value);
	track_b = GPOINTER_TO_INT (elem_b->value);

	gtk_tree_path_free (path_a);
	gtk_tree_path_free (path_b);

	if (track_a  < track_b)
	  cmp_result = -1;
  
	if (track_a == track_b)
	  cmp_result =  0;

	if (track_a >  track_b)
	  cmp_result =  1;

//	if ((track_a != 0) && (track_b != 0))
	  goto done;
      }
    else 
      {
	if (path_a)
	  gtk_tree_path_free (path_a);

	if (path_b)
	  gtk_tree_path_free (path_b);
      }

    if ((key_a != NULL) && (key_b == NULL))
      {
	cmp_result = 1;
	goto done;
      }

    if ((key_a == NULL) && (key_b == NULL))
      {
	cmp_result = 0;
	goto done;
      }

    if ((key_a == NULL) && (key_b != NULL))
      {
	cmp_result = -1;
	goto done;
      }
    
    cmp_result = g_utf8_collate (key_a, key_b);

    done:

#if 0
    g_free (_key_a);
    g_free (_key_b);

    g_free (key_a);
    g_free (key_b);
#endif

    return cmp_result;
}

static void
on_tag_added (BmpMetadataCache *cache,
	      gchar	       *tagname,
	      BmpLibrary       *self)
{
    GladeXML	  *xml = self->private->xml; 

    GtkComboBox	  *cbox_tag = GTK_COMBO_BOX(glade_xml_get_widget_warn(xml, "cbox_tag"));
    GtkTreeIter	   iter, iter_child;
    GdkPixbuf	  *pb_bookmark;
    GHashTable	  *mapping_sub;
    gint	   n;

    gtk_combo_box_append_text (cbox_tag, tagname);
    pb_bookmark = gdk_pixbuf_new_from_file (DATA_DIR "/images/icon-bookmark.png", NULL);

    gtk_tree_store_append (GTK_TREE_STORE(self->private->tree_store_tags), &iter, NULL);
    attr_row_map_mapping_insert (self->private->map_root_tags, tagname, GTK_TREE_MODEL(self->private->tree_store_tags), &iter);

    mapping_sub = attr_row_map_new (); 
    gtk_tree_store_set (GTK_TREE_STORE(self->private->tree_store_tags), &iter, 0, pb_bookmark, 1, tagname, 2, mapping_sub, -1);

    for (n = 0; n < BMP_N_TAGS; n++) 
      {
	GHashTable  *mapping_values;
	gchar	    *name;

	gtk_tree_store_append (GTK_TREE_STORE(self->private->tree_store_tags), &iter_child, &iter);
	name = bmp_metadata_get_title_plural (bmp_metadata_tag_id_to_datum(n));
	attr_row_map_mapping_insert (mapping_sub, name, GTK_TREE_MODEL(self->private->tree_store_tags), &iter_child);

	mapping_values = attr_row_map_new ();
	gtk_tree_store_set (GTK_TREE_STORE(self->private->tree_store_tags), &iter_child, 
			      TAG_COL_PIXBUF, 
			      self->private->pb_starred[n], 
			      TAG_COL_NAME,
			      name, 
			      TAG_COL_MAP,
			      mapping_values, 
			      -1); 
      }
}

static void
on_tag_deleted (BmpMetadataCache *cache,
	        gchar	         *tagname,
	        BmpLibrary	 *self)
{
    GladeXML	  *xml;
    GtkComboBox	  *cbox_tag;
    GtkTreeModel  *model_cbox;
    GtkTreeIter	   iter;
    gboolean	   proceed = TRUE;

    attr_row_map_mapping_get (self->private->map_root_tags, tagname, GTK_TREE_MODEL(self->private->tree_store_tags), &iter);
    if (!gtk_tree_store_remove (GTK_TREE_STORE(self->private->tree_store_tags), &iter))
	{
	    //FIXME: Add some error handling here.. shouldn't probably happen ever anyway
	}

    xml = self->private->xml; 

    cbox_tag = GTK_COMBO_BOX(glade_xml_get_widget_warn(xml, "cbox_tag"));
    model_cbox = gtk_combo_box_get_model (GTK_COMBO_BOX(cbox_tag));

    if (!gtk_tree_model_get_iter_first (model_cbox, &iter)) return;
    gtk_tree_model_iter_next (model_cbox, &iter);

    while (proceed)
      {
	gchar *value;

	gtk_tree_model_get (model_cbox, &iter, 0, &value, -1);

	if (!strcmp (tagname, value))
	  {
	    if (!gtk_list_store_remove (GTK_LIST_STORE(model_cbox), &iter))
	      {
		  //FIXME: Add some error handling here.. shouldn't probably happen ever anyway
	      }
	    break;
	  }

	proceed = gtk_tree_model_iter_next (model_cbox, &iter);
      }

    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER(self->private->tree_store_browse_filter));
}

static void
on_tag_value_insert (BmpMetadataCache	*cache,
		     const gchar	*tagname,
		     BmpMetadatumTagId   metadatum_tag_id,
		     TagDataTuple	*data_tuple, 
		     BmpLibrary		*self)
{
    GtkTreeIter	   iter, iter_child;
    GHashTable	  *mapping_sub, *mapping_tags;
    gchar	  *name;

    if (!attr_row_map_mapping_get (self->private->map_root_tags, tagname, GTK_TREE_MODEL(self->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag with name '%s' does not exist", G_STRLOC, tagname);
	return;
      }

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_tags), &iter, 2, &mapping_sub, -1);

    name = bmp_metadata_get_title_plural (bmp_metadata_tag_id_to_datum(metadatum_tag_id));

    if (!attr_row_map_mapping_get (mapping_sub, name, GTK_TREE_MODEL(self->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag Id %d does not exist", G_STRLOC, metadatum_tag_id);
	g_free (name);
	return;
      }
    g_free (name);

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_tags), &iter, TAG_COL_MAP, &mapping_tags, -1); 

    gtk_tree_store_append (GTK_TREE_STORE(self->private->tree_store_tags), &iter_child, &iter);
    attr_row_map_mapping_insert (mapping_tags, data_tuple->tag[metadatum_tag_id], GTK_TREE_MODEL(self->private->tree_store_tags), &iter_child);

    gtk_tree_store_set (GTK_TREE_STORE(self->private->tree_store_tags), &iter_child, 
			TAG_COL_PIXBUF, 
			self->private->pb[metadatum_tag_id], 
			TAG_COL_NAME, 
			data_tuple->tag[metadatum_tag_id], 
			TAG_COL_MAP, 
			NULL, 
			TAG_COL_TUPLE, 
			data_tuple,
			TAG_COL_DATUM,
			metadatum_tag_id,
			-1); 
}

static void
on_tag_value_remove (BmpMetadataCache	*cache,
		     const gchar	*tagname,
		     BmpMetadatumTagId   metadatum_tag_id,
		     TagDataTuple	*data_tuple, 
		     BmpLibrary		*self)
{
    GtkTreeIter	   iter, iter_child;
    GHashTable	  *mapping_sub, *mapping_tags;
    gchar	  *name;

    if (!attr_row_map_mapping_get (self->private->map_root_tags, tagname, GTK_TREE_MODEL(self->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag with name '%s' does not exist", G_STRLOC, tagname);
	return;
      }

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_tags), &iter, 2, &mapping_sub, -1);

    name = bmp_metadata_get_title_plural (bmp_metadata_tag_id_to_datum(metadatum_tag_id));

    if (!attr_row_map_mapping_get (mapping_sub, name, GTK_TREE_MODEL(self->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag Id %d does not exist", G_STRLOC, metadatum_tag_id);
	g_free (name);
	return;
      }
    g_free (name);

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->tree_store_tags), &iter, TAG_COL_MAP, &mapping_tags, -1);

    if (attr_row_map_mapping_get (mapping_tags, data_tuple->tag[metadatum_tag_id], GTK_TREE_MODEL(self->private->tree_store_tags), &iter_child))
      gtk_tree_store_remove (GTK_TREE_STORE(self->private->tree_store_tags), &iter_child);
}

static void
on_library_add_path_clicked (GtkButton	    *button,
			     BmpLibrary	    *self) 
{
    GladeXML	  *xml;
    GtkWidget	  *dialog, *notebook_library_add, *window;
    GList	  *items = NULL;
    GError	  *error = NULL;
    gchar	  *folder;
    gint	   rv;

    g_return_if_fail (GTK_IS_BUTTON(button));

    xml = self->private->xml; 
    window = glade_xml_get_widget_warn (xml, "dialog_library");
    dialog = gtk_file_chooser_dialog_new (_("Select Path To Add"),
                                          GTK_WINDOW (glade_xml_get_widget_warn (xml, "dialog_library")),
                                          GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                          GTK_STOCK_CANCEL,
                                          GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_ADD,
                                          GTK_RESPONSE_OK,
                                          NULL);

    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),  
					 xcs_key_fetch (xcs, 
					 "bmp", 
					 "file-chooser-path"));

    rv = gtk_dialog_run (GTK_DIALOG (dialog));

    if (rv == GTK_RESPONSE_CANCEL)
      {
	gtk_widget_destroy (dialog);
	return;
      }

    folder = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy (dialog);

    notebook_library_add = glade_xml_get_widget_warn (xml, "notebook_library_add");
    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 1);
    bmp_window_set_busy (GTK_WIDGET(window));

    dir_foreach (folder, (DirForeachFunc) collect_path, &items, &error);

    for ( ; items ; items = items->next)
      {
	gchar *uri;
	 
	uri = g_filename_to_uri (items->data, NULL, NULL);
	bmp_metadata_cache_get_metadata (bmp_metadata_cache, uri);
	g_free (uri);

	while (gtk_events_pending()) gtk_main_iteration();
      }

    g_list_foreach (g_list_first(items), (GFunc)g_free, NULL);
    g_free (folder);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 0);
    bmp_window_set_idle (GTK_WIDGET(window));
}

enum {
    TAB_BROWSE,
    TAB_TAGS
};

static GtkTreeViewColumn* column_browse[4]; /* 0 for the main stuff, 1: BROWSE_COL_GENRE, 2: BROWSE_COL_BITRATE */
static GtkCellRenderer*   renderer_browse[6]; /* 0-2 for the main stuff, 3: BROWSE_COL_GENRE, 4: BROWSE_COL */

static GtkTreeViewColumn* column_tags[2]; 
static GtkCellRenderer*   renderer_tags[2];

static GtkTreeViewColumn* column_modify[1]; 
static GtkCellRenderer*   renderer_modify[2];

static void 
query_create (GtkTreeModel *store, GtkTreePath *path, gchar **_name, gchar **_uri)
{
    GtkTreeIter  iter,
		 iter_parent;

    *_name = NULL;
    *_uri  = NULL;

    gint depth = gtk_tree_path_get_depth (path);
    gtk_tree_model_get_iter (store, &iter, path); 

    switch (depth)
	{
	    case 1:
	      {
			    HTuple     *tuple;
			    HTupleElem *elem;
			    GURI       *uri;
			    GString    *query_uri;

			    query_uri = g_string_new ("query://");
			    g_string_append (query_uri, BMP_DBUS_INTERFACE); 
			    g_string_append (query_uri, "/"); 

			    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1); 
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-1]));

			    g_string_append_printf (query_uri, "?%s=", bmp_metadata_get_id_static (browse_levels[depth-1])); 
			    g_string_append (query_uri, elem->value);

			    uri = bmp_uri_new (query_uri->str);
			    g_string_free (query_uri, TRUE);

			    bmp_uri_escape (uri);

			    *_name = g_strdup (elem->value);
			    *_uri  = bmp_uri_get_string (uri);
      
			    bmp_uri_delete (uri);
			    return;
	    } break;
 
	  case 2:
	    {
			    HTuple     *tuple;
			    HTupleElem *elem;
			    GString    *query_uri;
			    GURI       *uri;
			    gchar      *l1, *l2;

			    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1); 

			    query_uri = g_string_new ("query://");
			    g_string_append (query_uri, BMP_DBUS_INTERFACE); 
			    g_string_append (query_uri, "/"); 

			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-1]));
			    l2 = g_strdup (elem->value);

			    gtk_tree_model_iter_parent (store, &iter_parent, &iter);
			    gtk_tree_model_get (store, &iter_parent, BROWSE_COL_TUPLE, &tuple, -1); 
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-2]));
			    l1 = g_strdup (elem->value);

			    g_string_append_printf (query_uri, "?%s=", bmp_metadata_get_id_static (browse_levels[depth-2])); 
			    g_string_append (query_uri, l1);
			    g_string_append_printf (query_uri, "&%s=", bmp_metadata_get_id_static (browse_levels[depth-1])); 
			    g_string_append (query_uri, l2); 

			    uri = bmp_uri_new (query_uri->str);
			    g_string_free (query_uri, TRUE);

			    bmp_uri_escape (uri);

			    *_name = g_strconcat (l1, " / ", l2, NULL);
			    *_uri  = bmp_uri_get_string (uri); 

			    bmp_uri_delete (uri);
			    return;
	    }
	 break;

	case 3:
	    {
			    HTuple     *tuple;
			    HTupleElem *elem;
			    GString    *query_uri;
			    GURI       *uri;
			    gchar      *l1, *l2, *l3;

			    query_uri = g_string_new ("query://");
			    g_string_append (query_uri, BMP_DBUS_INTERFACE); 
			    g_string_append (query_uri, "/"); 

			    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1); 
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-1]));
			    l3 = g_strdup (elem->value);

			    gtk_tree_model_iter_parent (store, &iter_parent, &iter);
			    gtk_tree_model_get (store, &iter_parent, BROWSE_COL_TUPLE, &tuple, -1); 
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-2]));
			    l2 = g_strdup (elem->value);

			    gtk_tree_model_iter_parent (store, &iter, &iter_parent);
			    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1); 
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-3]));
			    l1 = g_strdup (elem->value);

			    g_string_append_printf (query_uri, "?%s=", bmp_metadata_get_id_static (browse_levels[depth-3])); 
			    g_string_append (query_uri, l1);
			    g_string_append_printf (query_uri, "&%s=", bmp_metadata_get_id_static (browse_levels[depth-2])); 
			    g_string_append (query_uri, l2); 
			    g_string_append_printf (query_uri, "&%s=", bmp_metadata_get_id_static (browse_levels[depth-1])); 
			    g_string_append (query_uri, l3); 

			    uri = bmp_uri_new (query_uri->str);
			    g_string_free (query_uri, TRUE);

			    bmp_uri_escape (uri);

			    *_name = g_strconcat (l1, " / ", l2, " / ", l3, NULL);
			    *_uri  = bmp_uri_get_string (uri);
    
			    bmp_uri_delete (uri);
			    return;
	      } break;
			
	      default: break; 
	}
}

static void
on_button_tag_clicked (GtkButton	*button, 
		       BmpLibrary	*self)
{
    GladeXML	      *xml;
    GtkWidget	      *image_type,
		      *label_name,
		      *dialog_set_tags,
		      *library_notebook;

    GtkWidget*	       tree_view;
    GtkTreeSelection  *selection;
  
    GtkTreeModel      *store;
    GtkTreeIter	       iter;

    HTuple	      *tuple;
    HTupleElem	      *elem;

    GList	      *tags = NULL,
		      *paths = NULL;

    BmpMetadatumId     metadatum_id;
    BmpMetadatumTagId  metadatum_tag_id;

    gint	       current_page = -1;
    gint	       depth;

    gboolean	       proceed = TRUE;

    xml = self->private->xml;
    library_notebook = glade_xml_get_widget_warn (xml, "library_notebook");
    dialog_set_tags = glade_xml_get_widget_warn (xml, "dialog_set_tags");

    current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK(library_notebook));

    if (current_page != TAB_BROWSE)
	return; //Shouldn't happen

    tree_view =	glade_xml_get_widget_warn (xml, "tree_view_browse");
    store = gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
    paths = gtk_tree_selection_get_selected_rows (selection, NULL);

    if (!paths)
	return; //Shouldn't happen.

    depth = gtk_tree_path_get_depth (paths->data);

    gtk_tree_model_get_iter (store, &iter, paths->data);
    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1);

    image_type = glade_xml_get_widget (xml, "image_type");
    gtk_image_set_from_pixbuf (GTK_IMAGE(image_type), self->private->pb[depth-1]);

    tree_view =	glade_xml_get_widget_warn (xml, "tree_view_modify_tags");
    store = GTK_TREE_MODEL(gtk_list_store_new (2, G_TYPE_BOOLEAN, G_TYPE_STRING));
    tags = bmp_metadata_cache_tags_list (bmp_metadata_cache);

    tags = g_list_sort (tags, (GCompareFunc)g_utf8_collate);

    metadatum_id = browse_levels[depth-1];
    metadatum_tag_id = bmp_metadata_datum_to_tag_id (metadatum_id);

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));
    label_name = glade_xml_get_widget (xml, "label_name");
    gtk_label_set_text (GTK_LABEL(label_name), elem->value);

    for ( ; tags ; tags = tags->next)
      {
	  gboolean tagged;

	  gtk_list_store_append (GTK_LIST_STORE(store), &iter);

	  tagged = bmp_metadata_cache_tuple_matches_tag (bmp_metadata_cache,
							 tuple,
							 metadatum_tag_id,
							 (const gchar*)tags->data);

	  gtk_list_store_set (GTK_LIST_STORE(store), &iter, 0, tagged, 1, (gchar*)tags->data, -1);
      }

    gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view), store);
    gtk_dialog_run (GTK_DIALOG(dialog_set_tags));

    proceed = gtk_tree_model_get_iter_first (store, &iter); 

    while (proceed)
      { 
	  GError   *error = NULL;
	  gchar	   *id;
	  gboolean  checked;

	  gtk_tree_model_get (store, &iter, 0, &checked, 1, &id, -1); 

	  if (checked)
	    {
		if (!bmp_metadata_cache_tag_tuple (bmp_metadata_cache,
						   tuple,
						   metadatum_tag_id,
						   id,
						   &error))
		  {
		    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: %s", G_STRLOC, error->message);
		    g_error_free (error);
		    error = NULL;
		  }
	    }
	  else
	    {
		if (!bmp_metadata_cache_untag_tuple (bmp_metadata_cache,
						   tuple,
						   metadatum_tag_id,
						   id,
						   &error))
		  {
		    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: %s", G_STRLOC, error->message);
		    g_error_free (error);
		    error = NULL;
		  }
	    }

	  proceed = gtk_tree_model_iter_next (store, &iter);
      }

    gtk_widget_hide (dialog_set_tags);
    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER(self->private->tree_store_browse_filter));
}

static gchar*
get_query_sublevel (GtkTreeModel *store, GtkTreeIter *iter)
{
  GString	      *query_uri;
  GURI		      *uri;
  gchar		      *uri_str;
  TagDataTuple	      *data_tuple;
  BmpMetadatumTagId    metadatum_tag_id;

  gtk_tree_model_get (store, iter, TAG_COL_TUPLE, &data_tuple, TAG_COL_DATUM, &metadatum_tag_id, -1); 

  query_uri = g_string_new ("query://");

  g_string_append (query_uri, BMP_DBUS_INTERFACE);
  g_string_append (query_uri, "/"); 

  g_string_append_printf (query_uri, "?%s=%s", bmp_metadata_get_id_static (BMP_DATUM_ARTIST), data_tuple->tag[BMP_TAG_ARTIST]); 

  if (metadatum_tag_id > BMP_TAG_ARTIST)
      g_string_append_printf (query_uri, "&%s=%s", bmp_metadata_get_id_static (BMP_DATUM_ALBUM), data_tuple->tag[BMP_TAG_ALBUM]); 
  if (metadatum_tag_id > BMP_TAG_ALBUM)
      g_string_append_printf (query_uri, "&%s=%s", bmp_metadata_get_id_static (BMP_DATUM_TITLE), data_tuple->tag[BMP_TAG_TITLE]); 

  uri = bmp_uri_new (query_uri->str);
  bmp_uri_escape (uri);
  uri_str = bmp_uri_get_string (uri); 

  bmp_uri_delete (uri); 
  g_string_free (query_uri, TRUE);

  return uri_str; 
}

static void
append_to_uniquelist (HTuple	       *tuple,
		      BmpUniqueList   **ulist)
{
  HTupleElem *elem;

  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_LOCATION));
  bmp_unique_list_insert (*ulist, g_strdup(elem->value));
}

static void
on_button_enqueue_clicked (GtkButton	  *button, 
			   BmpLibrary	  *self)
{
    GladeXML	      *xml;
    GtkWidget 	      *window;
    GtkNotebook       *library_notebook;
    GtkWidget	      *tree_view;
    GtkTreeModel      *store;
    GtkTreeSelection  *selection;
    GList	      *paths = NULL;
    GList	      *uris = NULL;
    gchar	     **uri_list;
    GtkTreeIter	       iter, iter_children, iter_children_children;
    gint	       new_ctr = 0, current_page = -1;

    xml = self->private->xml;
    window = glade_xml_get_widget_warn (xml, "dialog_library");
    bmp_window_set_busy (GTK_WIDGET(window));

    library_notebook = GTK_NOTEBOOK (glade_xml_get_widget_warn (xml, "library_notebook"));
    current_page = gtk_notebook_get_current_page (library_notebook);

    switch (current_page)
      {
	case TAB_TAGS:
	  {
	    GtkTreePath		*tree_path;

	    tree_view =
		glade_xml_get_widget_warn (xml, "tree_view_tags");
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));

	    if (!gtk_tree_selection_get_selected (selection, &store, &iter))
	      goto quit_enqueue;

	    tree_path = gtk_tree_model_get_path (store, &iter);

	    switch (gtk_tree_path_get_depth(tree_path))
		{
		    case 1:
			{
			    // Here we cant use get_query_sublevel() because we must avoid dupes:
			    // An album might be basically tagged twice, once as the album explicitely,
			    // and secondly implicitly inside the artist of this album.
			    BmpUniqueList   *ulist;
			    gboolean	     proceed;
  
			    ulist = bmp_unique_list_new (G_TYPE_STRING);

			    //First descend into all subs
			    proceed = gtk_tree_model_iter_children (store, &iter_children, &iter);
			    while (proceed)
			      {
				    gboolean proceed_s = gtk_tree_model_iter_children (store, &iter_children_children, &iter_children);
				    while (proceed_s)
				      {
					  TagDataTuple	      *data_tuple;
					  Query		      *query;
					  HRelation	      *result;
					  BmpMetadatumTagId    metadatum_tag_id;

					  gtk_tree_model_get (store, &iter_children_children, TAG_COL_TUPLE, &data_tuple, TAG_COL_DATUM, &metadatum_tag_id, -1); 

					  query = bmp_metadata_cache_query_initialize (bmp_metadata_cache);
					  bmp_metadata_cache_query_append_string (bmp_metadata_cache,
										  query,
										  bmp_metadata_tag_id_to_datum (metadatum_tag_id),
										  TRUE,
										  data_tuple->tag[metadatum_tag_id]);

					  result = bmp_metadata_cache_query_run (bmp_metadata_cache,
										 query);   


				result = h_relation_sort (
					 h_relation_sort (
                                         h_relation_sort (result,
                                                          (GCompareDataFunc) h_tuple_compare_elem,
                                                          (gpointer)bmp_metadata_get_id_static (BMP_DATUM_TRACK)),
                                                 (GCompareDataFunc) h_tuple_compare_elem,
                                                 (gpointer)bmp_metadata_get_id_static (BMP_DATUM_ARTIST)),

                              (GCompareDataFunc) h_tuple_compare_elem,
                              (gpointer)bmp_metadata_get_id_static (BMP_DATUM_ALBUM));



					  h_relation_for_each_tuple (result, (HTupleFunc)append_to_uniquelist, &ulist);

					  h_relation_unref (result);
					  bmp_metadata_cache_query_free (bmp_metadata_cache, query);
					  
					  proceed_s = gtk_tree_model_iter_next (store, &iter_children_children);
				      }

				    proceed = gtk_tree_model_iter_next (store, &iter_children);
			      }
			    uri_list = glist_to_strv (bmp_unique_list_to_glist(ulist));
			    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    self->private->clear,
					    self->private->start_playback,
					    self->private->start_playback_track,
					    &new_ctr,
					    NULL);
			    bmp_unique_list_free (ulist);
			}
			break;

		    case 2:
			{
			    gboolean proceed = TRUE;
  
			    proceed = gtk_tree_model_iter_children (store, &iter_children, &iter);

			    while (proceed)
			      {
				  uris = g_list_append (uris, get_query_sublevel (store, &iter_children)); 
				  proceed = gtk_tree_model_iter_next (store, &iter_children);
			      }

			    uri_list = glist_to_strv (uris);
			    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    self->private->clear,
					    self->private->start_playback,
					    self->private->start_playback_track,
					    &new_ctr,
					    NULL);
			}
			break;

		    case 3:
			{
			    uris = g_list_append (uris, get_query_sublevel (store, &iter)); 
			    uri_list = glist_to_strv (uris);
			    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    self->private->clear,
					    self->private->start_playback,
					    self->private->start_playback_track,
					    &new_ctr,
					    NULL);
			}
			break;
  
		    default: break;
		}

		g_list_foreach (uris, (GFunc)g_free, NULL);
		g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
		g_list_free (paths);
		g_list_free (uris);

	  }
	break;

        case TAB_BROWSE: 
	  {
	    tree_view =
		glade_xml_get_widget_warn (xml, "tree_view_browse");
	    store =
		gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
	    paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);

	    if (!paths)
	      {
		gtk_tree_selection_select_all (selection);
		paths =
		    gtk_tree_selection_get_selected_rows (selection, NULL);
	      }

	    while (paths)
	      {
		gtk_tree_model_get_iter (store, &iter, paths->data);

	      switch (gtk_tree_path_get_depth (paths->data))
		  {
			// FIXME: Yay wow man is this efficient, can't hold onto my chair 
			case 1:
			  {
			    gboolean run_1 = TRUE, run_2 = TRUE;

			    run_1 = 
			      gtk_tree_model_iter_children (store, &iter_children, &iter);

			    while (run_1)
			      {
				run_2 = 
				  gtk_tree_model_iter_children (store, &iter_children_children, &iter_children);

				while (run_2)
				  {
				    HTuple *tuple;
				    HTupleElem *elem;

				    gtk_tree_model_get (store, &iter_children_children, BROWSE_COL_TUPLE, &tuple, -1); 

				    elem = 
				      h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
				    uris = 
				      g_list_append (uris, g_strdup (elem->value));
				    run_2 = 
				      gtk_tree_model_iter_next (store, &iter_children_children);
				  }

				  run_1 = 
				    gtk_tree_model_iter_next (store, &iter_children);
			      }

			  } break;
  
			case 2:
			  {
			    gboolean run = TRUE;

			    run = 
			      gtk_tree_model_iter_children (store, &iter_children, &iter);

			    while (run)
			      {
				HTuple *tuple;
				HTupleElem *elem;

				gtk_tree_model_get (store, &iter_children, BROWSE_COL_TUPLE, &tuple, -1); 

				elem = 
				  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
				uris = 
				  g_list_append (uris, g_strdup (elem->value));
				run = 
				  gtk_tree_model_iter_next (store, &iter_children);
			      }

			  } break;
	
			case 3:
			  {
			    HTuple *tuple;
			    HTupleElem *elem;

			    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1); 

			    elem = 
			      h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
			    uris = 
			      g_list_append (uris, g_strdup (elem->value));
			  } break;
			
			default: /* Uhm... */ break; 
		  }

	      paths = paths->next;
	  }

	uri_list = glist_to_strv (uris);
	bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    self->private->clear,
					    self->private->start_playback,
					    self->private->start_playback_track,
					    &new_ctr,
					    NULL);

	g_list_foreach (uris, (GFunc)g_free, NULL);
	g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	g_list_free (paths);
	g_list_free (uris);
    }
    break;

    default: break;
  }

  quit_enqueue:

  self->private->clear = FALSE;
  self->private->start_playback = FALSE;
  self->private->start_playback_track = 0;

  bmp_window_set_idle (GTK_WIDGET(window));
}

static void
on_button_play_clicked (GtkButton	    *button,
			BmpLibrary	    *self)
{
    GladeXML  *xml = self->private->xml;

    self->private->clear = TRUE;
    self->private->start_playback = TRUE;
    self->private->start_playback_track = 0;

    on_button_enqueue_clicked (GTK_BUTTON(glade_xml_get_widget_warn (xml, "button_enqueue")), self);
}

static gboolean
tree_store_browse_visible_func (GtkTreeModel  *model,
                                GtkTreeIter   *iter,
                                BmpLibrary    *self) 
{
    GtkWidget	      *cbox_metadatum_id,
		      *cbox_tag,
		      *entry_value;

    GladeXML	      *xml = self->private->xml; 
    BmpMetadatumId     metadatum_id;
    HTuple	      *tuple;
    HTupleElem	      *elem;
#ifdef HAVE_HAL
    gboolean	       is_present;
#endif

    cbox_metadatum_id = glade_xml_get_widget_warn (xml, "cbox_metadatum_id");
    cbox_tag = glade_xml_get_widget_warn (xml, "cbox_tag");
    entry_value = glade_xml_get_widget_warn (xml, "entry_value");

    metadatum_id = gtk_combo_box_get_active (GTK_COMBO_BOX(cbox_metadatum_id));

    gtk_tree_model_get (model, iter, BROWSE_COL_TUPLE, &tuple, -1);

    if (!tuple) return TRUE;

#ifdef HAVE_HAL
    if (!bmp_metadata_cache_check_tuple_volume	(bmp_metadata_cache,
						 tuple,
						 &is_present))
      {
	//FIXME: Error
      }
    else
      {
	if (!is_present) return FALSE;
      }
#endif

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));

    if (!elem->value)
	return TRUE;

    if (match_keys (elem->value, gtk_entry_get_text (GTK_ENTRY(entry_value))))
      {
	//Ok, we got a match, so let's check for the tag
	if ((metadatum_id == BMP_DATUM_ARTIST) || (metadatum_id == BMP_DATUM_ALBUM) || (metadatum_id == BMP_DATUM_TITLE))
	  {
	      gchar   *value;
	      gint     tag;
	      gboolean matches;

	      tag = gtk_combo_box_get_active (GTK_COMBO_BOX(cbox_tag)); 
	      if (!tag) return TRUE; //No tag selected, so we ignore it

	      value = gtk_combo_box_get_active_text (GTK_COMBO_BOX(cbox_tag));
	      if (!value) return TRUE;

              matches = bmp_metadata_cache_tuple_matches_tag (bmp_metadata_cache,
							      tuple,
							      metadatum_id,
							      value); 

	      g_free (value);
	      return matches;
	  }
	else
	      return TRUE;
      }
    else
      {
	//If it doesn't match the keys, then we return FALSE regardless of the tag
	return FALSE;
      }

    g_assert_not_reached ();
}

static gboolean
on_dialog_library_key_press (GtkWidget	  *widget,
			     GdkEventKey  *event,
			     gpointer	   data)
{
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->keyval == GDK_Escape) gtk_widget_hide (widget);
    return FALSE;
}

void
length_renderer_func (GtkTreeViewColumn *tree_column,
                      GtkCellRenderer *cell,
                      GtkTreeModel *tree_model,
                      GtkTreeIter *iter,
                      gpointer data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	gint ttime, minutes, seconds;
	g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_TUPLE, &tuple, -1);

	elem = 
	  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_TIME));
	ttime = GPOINTER_TO_INT(elem->value);

	minutes = ttime / 60;
	seconds = ttime % 60;

	cell_data = 
	  g_strdup_printf ("%d:%-2.2d", minutes, seconds); 

	g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	g_free (cell_data);
      }
}

void
tracknumber_renderer_func          (GtkTreeViewColumn *tree_column,
                                             GtkCellRenderer *cell,
                                             GtkTreeModel *tree_model,
                                             GtkTreeIter *iter,
                                             gpointer data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_TUPLE, &tuple, -1);

	elem = 
	  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	cell_data = 
	  g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));

	g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	g_free (cell_data);
      }
}

void
bitrate_renderer_func (GtkTreeViewColumn *tree_column,
                       GtkCellRenderer *cell,
                       GtkTreeModel *tree_model,
                       GtkTreeIter *iter,
                       gpointer data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_TUPLE, &tuple, -1);

	elem = 
	  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_BITRATE));
	cell_data = 
	  g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));

	g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	g_free (cell_data);
      }
}


void
genre_renderer_func          (GtkTreeViewColumn *tree_column,
                              GtkCellRenderer *cell,
                              GtkTreeModel *tree_model,
                              GtkTreeIter *iter,
                              gpointer data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_TUPLE, &tuple, -1);

	elem = 
	  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_GENRE));

	g_object_set (G_OBJECT(cell), "text", elem->value, NULL);
      }
}


static gboolean
on_dialog_library_delete (GtkWidget *widget,
			  GdkEvent *event,
                          gpointer user_data)
{
  gtk_widget_hide (widget);
  return TRUE;
}


static void
on_library_notebook_switch_page (GtkNotebook	  *notebook,
				 GtkNotebookPage  *notebook_page,
				 guint		   page,
				 BmpLibrary	  *self)
{
    GtkTreeView	      *tree_view;
    GtkTreeSelection  *tree_selection;

    GladeXML	      *xml;

    GtkWidget	      *b_tag, 
		      *b_play, 
		      *b_enqueue,
		      *b_delete;

    GList	      *paths;

    static gchar* page_names[] = {
	"tree_view_browse",
	"",
	"tree_view_tags"
    };
		
    g_return_if_fail (GTK_IS_NOTEBOOK(notebook));

    xml = self->private->xml; 

    b_tag = 
	glade_xml_get_widget_warn (xml, "button_tag");
    b_play = 
	glade_xml_get_widget_warn (xml, "button_play");
    b_enqueue = 
	glade_xml_get_widget_warn (xml, "button_enqueue");
    b_delete = 
	glade_xml_get_widget_warn (xml, "button_tag_delete");

    if (!strlen(page_names[page]))
      {
	  gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 
	  gtk_widget_set_sensitive (GTK_WIDGET(b_play), FALSE); 
	  gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), FALSE); 
	  return;
      }

    tree_view = GTK_TREE_VIEW(glade_xml_get_widget_warn (xml, page_names[page])); 
    tree_selection = gtk_tree_view_get_selection (tree_view);  

    paths = gtk_tree_selection_get_selected_rows (tree_selection, NULL);

    if (page == TAB_TAGS) 
      {
	  if (paths && (gtk_tree_path_get_depth(paths->data)>1))
	    gtk_widget_set_sensitive (b_delete, FALSE); 
	  else
	    gtk_widget_set_sensitive (b_delete, (gtk_tree_selection_count_selected_rows (tree_selection) != 0) ? TRUE:FALSE);

	  gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 
      }
    else
      {
	  if (g_list_length (paths)==1)
	      gtk_widget_set_sensitive (GTK_WIDGET(b_tag), TRUE); 
	  else
	      gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 
      }

#if 0
    if (paths && (page == TAB_TAGS) && (gtk_tree_path_get_depth(paths->data)<3))
      {
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), FALSE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), FALSE); 
      }
    else
      {
#endif
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
#if 0
      }      
#endif

    g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
}

static void
on_tree_selection_changed (GtkTreeSelection *tree_selection,
			   BmpLibrary	    *self)
{
    GladeXML	      *xml;

    GtkWidget	      *b_tag, 
		      *b_play, 
		      *b_enqueue;

    GtkWidget	      *library_notebook;
    GList	      *paths;

    xml = self->private->xml; 

    library_notebook =
	glade_xml_get_widget_warn (xml, "library_notebook");

    b_tag = 
	glade_xml_get_widget_warn (xml, "button_tag");
    b_play = 
	glade_xml_get_widget_warn (xml, "button_play");
    b_enqueue = 
	glade_xml_get_widget_warn (xml, "button_enqueue");
  
    paths = gtk_tree_selection_get_selected_rows (tree_selection, NULL);

    if (g_list_length(paths)==1)
	gtk_widget_set_sensitive (GTK_WIDGET(b_tag), TRUE); 
    else
	gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 

    gtk_widget_set_sensitive (GTK_WIDGET(b_play), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
    gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);

    g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
}

//FIXME:
static DNDQueryTuple *qtuple = NULL;

/* DnD API implemenations */
static GdkPixmap *
tv_browse_create_row_drag_icon (GtkWidget   *tree_view,
                                BmpLibrary  *self)
{
    GladeXML	      *xml;
    GtkTreeModel      *tree_model;
    GtkTreeSelection  *tree_selection;
    GtkTreePath	      *tree_path;

    PangoLayout	      *layout;
    GdkDrawable	      *drawable;
    cairo_t	      *cr;
    gint	       width;
    gint	       height;
    gint	       depth;
    gint	       tree_path_depth;
    gchar	      *name,
		      *uri;
    GList	      *paths;


    xml = self->private->xml;
    depth = gdk_visual_get_best_depth ();

    tree_selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
    paths = gtk_tree_selection_get_selected_rows (tree_selection, &tree_model);
    
    if (!paths)
	return NULL;

    //FIXME: This is a little whackish because we assume there is only one selected path, _AND_ that it is the one being dragged
    tree_path = (GtkTreePath*)paths->data; 
    tree_path_depth = gtk_tree_path_get_depth (tree_path);
  
    query_create (tree_model, tree_path, &name, &uri);
 
    qtuple = g_new0 (DNDQueryTuple,1);
    qtuple->name = g_strdup(name);
    qtuple->uri  = g_strdup(uri);

    layout = gtk_widget_create_pango_layout (GTK_WIDGET(tree_view), name);
    pango_layout_get_pixel_size (layout, &width, &height);
    
    height = 28;
    width += 16 + 16 + 8;
    drawable = gdk_pixmap_new (NULL, width, height, depth);

    gtk_paint_box                   (GTK_WIDGET(tree_view)->style,
                                             drawable, 
                                             GTK_STATE_NORMAL, 
                                             GTK_SHADOW_OUT, 
                                             NULL, 
                                             GTK_WIDGET(tree_view), 
                                             "", 
                                             0, 
                                             0, 
                                             width,
                                             height);


    gtk_paint_layout                (GTK_WIDGET(tree_view)->style,
                                             drawable, 
                                             GTK_STATE_NORMAL, 
                                             TRUE, 
                                             NULL, 
                                             GTK_WIDGET(tree_view), 
                                             "", 
                                             32, 
                                             6, 
                                             layout); 

    cr = gdk_cairo_create (drawable);

    gdk_cairo_set_source_pixbuf     (cr,
                                     self->private->pb[tree_path_depth-1], 
                                     10, 
				     6); 

    cairo_paint (cr);
    cairo_destroy (cr);

    return drawable;
}

static void
tv_browse_drag_begin (GtkWidget	      *widget,
		      GdkDragContext  *ctx,
                      BmpLibrary      *self) 
{
    GdkPixmap *pixmap;
    gint       w, 
	       h;

    pixmap = tv_browse_create_row_drag_icon (widget, self); 
    
    if (!pixmap) gtk_drag_finish (ctx, FALSE, FALSE, 0L);
    gdk_drawable_get_size (pixmap, &w, &h);

    gtk_drag_set_icon_pixmap (ctx,
                              gdk_drawable_get_colormap (pixmap),
                              pixmap, 
			      NULL, 
			      0, 0); 
}

static void
drag_end (GtkWidget      *widget,
          GdkDragContext *drag_context,
          gpointer        data) 
{
    if (qtuple)
      {
	  g_free (qtuple->name);
	  g_free (qtuple->uri);
	  g_free (qtuple);
      }
}

static void
tv_browse_drag_data_get (GtkWidget        *widget,
                         GdkDragContext   *drag_context,
                         GtkSelectionData *data,
                         guint             info,
                         guint             time,
                         BmpLibrary       *self)
{
    gtk_selection_data_set          (data,
                                     gdk_atom_intern("application/x-bmpx-db-query", FALSE), 
                                     8, 
                                     (gpointer) &qtuple, 
                                     sizeof(gpointer)); 
}

static void
on_button_tag_new_clicked (GtkWidget	*widget,
			   BmpLibrary	*self)
{
    GError	  *error = NULL;
    GladeXML	  *xml = self->private->xml;
    GtkEntry	  *entry_tagname = GTK_ENTRY(glade_xml_get_widget_warn(xml, "entry_tagname"));
    const gchar	  *text = gtk_entry_get_text (entry_tagname);

    if ((!text) || (!strlen(text)))
      {
	GtkWidget   *message_dialog;

	message_dialog = gtk_message_dialog_new (GTK_WINDOW(glade_xml_get_widget_warn (xml, "dialog_library")),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR, 
						 GTK_BUTTONS_OK, 
						 _("Please supply a name for the new tag!")); 
	gtk_dialog_run (GTK_DIALOG(message_dialog));
	gtk_widget_destroy (message_dialog);
	return;
      }

    if (!bmp_metadata_cache_tag_new (bmp_metadata_cache,
				     text,
				     &error))
      {
	 if (g_error_matches (error, 
			      g_quark_from_static_string ("BmpMetadataTags"),
			      BMP_TAGS_EEXISTS)) 
	    {
	  	GtkWidget   *message_dialog;

		message_dialog = gtk_message_dialog_new (GTK_WINDOW(glade_xml_get_widget_warn (xml, "dialog_library")),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR, 
						 GTK_BUTTONS_OK, 
						 _("A tag with this name already exists. Please supply a different name.")); 
		gtk_dialog_run (GTK_DIALOG(message_dialog));
		gtk_widget_destroy (message_dialog);
		return;
	    }
  
	  // What could have happened here?!?! uh.
	  return;
      }
}

static void
on_button_tag_delete_clicked (GtkWidget	    *widget,
			      BmpLibrary    *self)
{
    GError	      *error = NULL;
    GladeXML	      *xml = self->private->xml; 
    GtkTreeView	      *tree_view_tags = GTK_TREE_VIEW(glade_xml_get_widget_warn(xml, "tree_view_tags"));
    GtkTreeSelection  *tree_selection = gtk_tree_view_get_selection (tree_view_tags);
    GtkTreeModel      *model_tags;
    GtkTreeIter	       iter;
    gchar	      *text;

    if (!gtk_tree_selection_get_selected (tree_selection, &model_tags, &iter))
      return;

    gtk_tree_model_get (model_tags, &iter, 1, &text, -1); 

    if (!bmp_metadata_cache_tag_delete (bmp_metadata_cache,
				     text,
				     &error))
      {
	 if (g_error_matches (error, 
			      g_quark_from_static_string ("BmpMetadataTags"),
			      BMP_TAGS_ENOEXIST)) 
	    {
	  	GtkWidget   *message_dialog;

		message_dialog = gtk_message_dialog_new (GTK_WINDOW(glade_xml_get_widget_warn (xml, "dialog_library")),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR, 
						 GTK_BUTTONS_OK, 
						 _("A tag with this name does not exist")); 
		gtk_dialog_run (GTK_DIALOG(message_dialog));
		gtk_widget_destroy (message_dialog);
		return;
	    }
  
	  // What could have happened here?!?! uh.
	  return;
      }
}

static void
on_tree_view_tags_selection_changed (GtkTreeSelection	*tree_selection,
				     BmpLibrary		*self)
{
    GladeXML	      *xml = self->private->xml; 

    GtkWidget	      *b_tag, 
		      *b_play, 
		      *b_enqueue,
		      *b_delete;

    GList	      *paths;
    
    paths = gtk_tree_selection_get_selected_rows (tree_selection, NULL);
    if (!paths)
      return;

    b_tag = 
	glade_xml_get_widget_warn (xml, "button_tag");
    b_play = 
	glade_xml_get_widget_warn (xml, "button_play");
    b_enqueue = 
	glade_xml_get_widget_warn (xml, "button_enqueue");
    b_delete = 
	glade_xml_get_widget_warn (xml, "button_tag_delete");

    if (gtk_tree_path_get_depth(paths->data)>1)
      {
	  gtk_widget_set_sensitive (b_delete, FALSE); 
      }
    else
      {
	  gtk_widget_set_sensitive (b_delete, (gtk_tree_selection_count_selected_rows (tree_selection) != 0) ? TRUE:FALSE);
      }

    gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 

#if 0
    if (gtk_tree_path_get_depth(paths->data)<3)
      {
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), FALSE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), FALSE); 
      }
    else
      {
#endif
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
#if 0
      }      
#endif
}

static void
on_renderer_modify_toggled (GtkCellRendererToggle *cell_renderer,
                            gchar                 *path,
                            BmpLibrary		  *self)
{
    GladeXML	      *xml = self->private->xml; 
    GtkTreeView	      *tree_view_modify_tags = GTK_TREE_VIEW(glade_xml_get_widget_warn (xml, "tree_view_modify_tags"));
    GtkTreeModel      *model = gtk_tree_view_get_model (tree_view_modify_tags);
    GtkTreePath	      *tree_path = gtk_tree_path_new_from_string (path);
    GtkTreeIter	       iter;
    gboolean	       value;

    if (!tree_path)
      return;

    if (!gtk_tree_model_get_iter (model, &iter, tree_path))
      {
	gtk_tree_path_free (tree_path);
	return;
      }

    gtk_tree_model_get (model, &iter, 0, &value, -1); 

    value = value ? FALSE : TRUE;

    gtk_list_store_set (GTK_LIST_STORE(model), &iter, 0, value, -1);
}

static void
on_entry_tagname_activate (GtkWidget	*widget,
			   BmpLibrary	*self)
{
    GladeXML	  *xml = self->private->xml; 
    GtkWidget	  *button_tag_new = glade_xml_get_widget_warn (xml, "button_tag_new");

    gtk_button_clicked (GTK_BUTTON(button_tag_new));
    gtk_entry_set_text (GTK_ENTRY(widget), "");
}

static void
on_cbox_metadatum_id_changed (GtkWidget	    *widget,
			      BmpLibrary    *self)
{
    GladeXML	      *xml = self->private->xml; 
    GtkWidget	      *cbox_tag = glade_xml_get_widget_warn (xml, "cbox_tag");
    BmpMetadatumId     metadatum_id = gtk_combo_box_get_active (GTK_COMBO_BOX(widget));

   
    if ( (metadatum_id == BMP_DATUM_ARTIST) || (metadatum_id == BMP_DATUM_ALBUM) || (metadatum_id == BMP_DATUM_TITLE) )
	gtk_widget_set_sensitive (cbox_tag, TRUE);
    else
	gtk_widget_set_sensitive (cbox_tag, FALSE);

    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER(self->private->tree_store_browse_filter));

}

#if 0
static gboolean
on_tb_playlist_new_toggled (GtkToggleButton *button,
			    BmpPlaylist	    *self)
{
  GtkWidget	  *menu;
  GdkWindow	  *parent;
  GtkRequisition   req;

  gint x_ro, y_ro;
  gint x_widget, y_widget;
  gint x_size, y_size;

  switch (gtk_toggle_button_get_active (button))
    {
	case TRUE:
	  {
	      parent = gtk_widget_get_parent_window(GTK_WIDGET(button));
	      gdk_drawable_get_size(parent, &x_size, &y_size);	 
	      gdk_drawable_get_size(GTK_WIDGET(button)->window, &x_size, &y_size);	 
	      gdk_window_get_root_origin(GTK_WIDGET(button)->window, &x_ro, &y_ro); 
	      gdk_window_get_position(GTK_WIDGET(button)->window, &x_widget, &y_widget);
 
	      menu = ui_manager_get_popup (bmp_ui_manager, "/popup-plmanager/menu-plmanager-new");
	      gtk_widget_size_request (menu, &req);
	      bmp_menu_popup (GTK_MENU(menu), 
			      x_ro + GTK_WIDGET(button)->allocation.x + 2, 
			      y_ro + y_size - GTK_WIDGET(button)->allocation.height - 6, 
			      1, 0L); 
	  }
	  break;

#if 0
	case FALSE:
	  {
	      menu = ui_manager_get_popup (bmp_ui_manager, "/popup-plmanager/menu-plmanager-new");
	      gtk_widget_hide (menu);
	  }
	  break;
#endif

	default: break;
    }

  return FALSE;
}

static void
disable_tb_playlist_new (GtkMenuShell	*shell,
		         BmpLibrary     *self)
{
  GladeXML    *xml = self->private->xml; 
  GtkWidget   *tb;

  tb = glade_xml_get_widget_warn (xml, "tb_playlist_new");

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(tb), FALSE);
}

static void
relation_tuple_cell_data_func (GtkTreeViewColumn *tree_column,
                               GtkCellRenderer *cell,
                               GtkTreeModel *tree_model,
                               GtkTreeIter *iter,
                               gpointer data)
{
  HTuple         *tuple = NULL;
  HTupleElem	 *elem = NULL;
  gchar	         *v_string = NULL;
  BmpMetadatumId  metadatum_id = GPOINTER_TO_INT(data);

  gtk_tree_model_get (tree_model, iter, 0 /* this isn't safe */, &tuple, -1);
  g_assert (tuple);

  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));
  g_assert (elem);

  if (elem->type == G_TYPE_INT)
    {
      v_string = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value)); 
    }
  else
    {
      v_string = g_strdup ((gchar*)elem->value);
    }

  g_object_set (G_OBJECT(cell), "text", v_string, NULL);

  g_free (v_string);
}

static void
on_tree_view_playlist_cursor_changed (GtkTreeView     *tree_view_p,
                                      BmpLibrary      *self)
{
    GladeXML	      *xml = self->private->xml; 
    GtkWidget	      *tree_view_t;

    GtkTreeModel      *model,
		      *model_list;

    GtkTreeSelection  *selection;

    GtkTreeIter	       iter;

    gchar	      *name;

    tree_view_t = glade_xml_get_widget_warn (xml, "tree_view_tracks"); 
    selection = gtk_tree_view_get_selection (tree_view_p);

    if (!gtk_tree_selection_get_selected (selection, &model, &iter))
      return;

    gtk_tree_model_get (model, &iter, 0, &name, -1);
    model_list = GTK_TREE_MODEL(pl_manager_list_get_model (bmp_pl_manager,
					    name));

    gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view_t), model_list);
}
#endif

static void
window_library_create (BmpLibrary *self)
{
    GError		*error = NULL;
    GtkWidget		*widget;
    static const gchar  *glade_file = DATA_DIR "/glade/dialog_library.glade";
    gint		 n;

    self->private->xml = glade_xml_new_or_die(N_("Library Dialog"), glade_file, NULL, NULL);
    if (!self->private->xml) return;
	
    glade_xml_signal_autoconnect (self->private->xml);

    /* Add actions */
    gtk_action_group_add_actions (bmp_actions, bmp_actions_LIBRARY, G_N_ELEMENTS (bmp_actions_LIBRARY), self->private->xml);
    if (!gtk_ui_manager_add_ui_from_file (bmp_ui_manager, DATA_DIR G_DIR_SEPARATOR_S BMP_UI_BASENAME G_DIR_SEPARATOR_S "plmanager.ui", &error))
      {
		g_message ("building menus failed: %s", error->message);
		g_error_free (error);
		error = NULL;
      }

    self->window = glade_xml_get_widget_warn (self->private->xml, "dialog_library");

    RM_ITEM_NEW("windows", "window-dialog-library", self->window, NULL);

    gtk_widget_realize (self->window);
    g_object_connect (G_OBJECT(self->window), 
			  "signal::delete-event", 
			  G_CALLBACK(on_dialog_library_delete),
			  self,
			  NULL);

    bmp_window_set_icon_list (GTK_WIDGET(self->window), "library");

    gtk_window_set_position (GTK_WINDOW(self->window), GTK_WIN_POS_CENTER);
    g_object_connect (G_OBJECT(self->window), 
			  "signal::key-press-event", 
			  G_CALLBACK(on_dialog_library_key_press), 
			  self, 
			  NULL);
    
    /* Various */
    widget = glade_xml_get_widget_warn (self->private->xml, "library_notebook");
    g_object_connect (G_OBJECT(widget),
			  "signal::switch-page",
			  G_CALLBACK(on_library_notebook_switch_page),
			  self, 
			  NULL);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(widget), 0);

    widget = glade_xml_get_widget_warn (self->private->xml, "image_library");
    gtk_image_set_from_file (GTK_IMAGE (widget), DATA_DIR "/images/header-library.png");

    widget = glade_xml_get_widget_warn (self->private->xml, "image_wait_adding");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget_warn (self->private->xml, "image_wait_library_inner");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget_warn (self->private->xml, "image_tag");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/icon-bookmark.png");


    widget = glade_xml_get_widget_warn (self->private->xml, "library_add_path");	
    g_object_connect (G_OBJECT(widget), 
			  "signal::clicked", 
			  G_CALLBACK(on_library_add_path_clicked), 
			  self, 
			  NULL);

    /* Setup Tags tab */
    widget = glade_xml_get_widget_warn (self->private->xml, "image_tag1");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/icon-bookmark.png");

    widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_tags");
    g_object_connect (G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(widget))),
			  "signal::changed",
			  G_CALLBACK(on_tree_view_tags_selection_changed),
			  self, 
			  NULL);

    column_tags[0] = gtk_tree_view_column_new ();

    gtk_tree_view_column_set_title (column_tags[0], _("Tag Name"));
    gtk_tree_view_column_set_sort_column_id (column_tags[0], 1);

    renderer_tags[0] =
		gtk_cell_renderer_pixbuf_new ();
    renderer_tags[1] =
		gtk_cell_renderer_text_new ();

    gtk_cell_renderer_set_fixed_size (renderer_tags[0], 16, 16);

    g_object_set (G_OBJECT(renderer_tags[0]), "xalign", 0.0, NULL);
    g_object_set (G_OBJECT(renderer_tags[1]), "xpad", 6, NULL);

    gtk_tree_view_column_pack_start
		(column_tags[0],
		 renderer_tags[0],
		 FALSE);

    gtk_tree_view_column_pack_start
		(column_tags[0],
		 renderer_tags[1],
		 TRUE);


    gtk_tree_view_column_set_attributes
		(column_tags[0],
		 renderer_tags[0],
		 "pixbuf",
		  0,
		  NULL);

    gtk_tree_view_column_set_attributes
		(column_tags[0],
		 renderer_tags[1],
		 "text",
		  1,
		  NULL);

    gtk_tree_view_column_set_expand
		(column_tags[0],
		 FALSE);

    gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_tags[0]));


	/* Setup BROWSE Tab */

	/* Filter combobox */
   	widget = glade_xml_get_widget_warn (self->private->xml, "cbox_metadatum_id");
	gtk_combo_box_remove_text (GTK_COMBO_BOX(widget), 0); //Remove dummy
	for (n = 0; n < BMP_DATUM_NONE; 
	    gtk_combo_box_append_text (GTK_COMBO_BOX(widget), bmp_metadata_get_title(n++)));
	gtk_combo_box_set_active (GTK_COMBO_BOX(widget), BMP_DATUM_ARTIST);

	g_object_connect(G_OBJECT(widget), 
			 "signal::changed", 
			 G_CALLBACK(on_cbox_metadatum_id_changed), 
			 self, 
			 NULL);


	widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_browse");
	g_object_connect(G_OBJECT(widget), 
			 "swapped-signal::row-activated", 
			 G_CALLBACK(gtk_window_activate_default), 
			 GTK_WINDOW(self->window), 
			 NULL);
	g_object_connect(G_OBJECT(gtk_tree_view_get_selection (GTK_TREE_VIEW(widget))),
			 "signal::changed",
			 G_CALLBACK(on_tree_selection_changed),
			 self, 
			 NULL);

	gtk_drag_source_set
			    (GTK_WIDGET(widget),
			     GDK_MODIFIER_MASK, 
			     dnd_entries, 
			     G_N_ELEMENTS(dnd_entries), 
                             GDK_ACTION_COPY); 

	g_object_connect (G_OBJECT(widget),
			  "signal::drag-begin",
			  G_CALLBACK(tv_browse_drag_begin),
			  self, 
			  "signal::drag-end",
			  G_CALLBACK(drag_end),
			  self, 
			  "signal::drag-data-get",
			  G_CALLBACK(tv_browse_drag_data_get),
			  self, 
			  NULL);

	gtk_tree_selection_set_mode
	    (gtk_tree_view_get_selection (GTK_TREE_VIEW(widget)),
	     GTK_SELECTION_MULTIPLE);

	/* Main column */ 
	column_browse[0] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[0], _("Artist / Album / Title"));
	gtk_tree_view_column_set_resizable (column_browse[0], TRUE);
	gtk_tree_view_column_set_min_width (column_browse[0],300);

	/* The cell renderer for the artist/album/title icon */
	renderer_browse[0] =
		gtk_cell_renderer_pixbuf_new ();

	g_object_set (renderer_browse[0], "width", 16, NULL);
	g_object_set (renderer_browse[0], "height", 16, NULL);

	gtk_tree_view_column_pack_start
		(column_browse[0],
		 renderer_browse[0],
		 FALSE);

	gtk_tree_view_column_set_attributes
		(column_browse[0],
		 renderer_browse[0],
		 "pixbuf",
		  0,
		  NULL);

	/* The cell renderer for the actual text */
	renderer_browse[1] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(renderer_browse[1]), 
		      "xpad", 4,
		      "ellipsize", PANGO_ELLIPSIZE_END,
		      NULL);

	gtk_tree_view_column_pack_start
		(column_browse[0],
		 renderer_browse[1],
		 TRUE);

	gtk_tree_view_column_set_attributes
		(column_browse[0],
		 renderer_browse[1],
		 "text",
		  1,
		  NULL);

	gtk_tree_view_column_set_expand
		(column_browse[0],
		 FALSE);

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[0]));

	/* Track number column */ 
	column_browse[1] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[1], _("Track"));

	renderer_browse[2] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(renderer_browse[2]),
	              "xalign", 1.0,
		      NULL);

	gtk_tree_view_column_pack_start
		(column_browse[1],
		 renderer_browse[2],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[1],
		 FALSE);

	gtk_tree_view_column_set_cell_data_func
		(column_browse[1],
		 renderer_browse[2],
		 (GtkTreeCellDataFunc) tracknumber_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[1]));

	/* Length */ 
	column_browse[2] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[2], _("Length"));

	renderer_browse[3] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(renderer_browse[2]),
	              "xalign", 0.5,
		      NULL);

	gtk_tree_view_column_pack_start
		(column_browse[2],
		 renderer_browse[3],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[2],
		 FALSE);

	gtk_tree_view_column_set_cell_data_func
		(column_browse[2],
		 renderer_browse[3],
		 (GtkTreeCellDataFunc) length_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[2]));

	/* Genre Column */
	column_browse[3] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[3], _("Genre"));

	renderer_browse[4] =
		gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column_browse[3],
		 renderer_browse[4],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[3],
		 FALSE);

      	gtk_tree_view_column_set_cell_data_func
		(column_browse[3],
		 renderer_browse[4],
		 (GtkTreeCellDataFunc) genre_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[3]));

	/* Bitrate Column */
	column_browse[4] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[4], _("Bitrate"));

	renderer_browse[5] =
		gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column_browse[4],
		 renderer_browse[5],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[4],
		 FALSE);

      	gtk_tree_view_column_set_cell_data_func
		(column_browse[4],
		 renderer_browse[5],
		 (GtkTreeCellDataFunc) bitrate_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[4]));

#if 0
	/* Setup Playlists tab */
	widget = glade_xml_get_widget_warn (self->private->xml, "image_playlists");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/icon-playlist.png");

	widget = glade_xml_get_widget_warn (self->private->xml, "tb_playlist_new");
	g_object_connect (G_OBJECT(widget),
			  "signal::toggled",
			  G_CALLBACK(on_tb_playlist_new_toggled),
			  self, 
			  NULL);

	widget = ui_manager_get_popup (bmp_ui_manager, "/popup-plmanager/menu-plmanager-new");
	g_object_connect (G_OBJECT(widget),
			  "signal::cancel",
			  G_CALLBACK(disable_tb_playlist_new),
			  self, 
			  "signal::selection-done",
			  G_CALLBACK(disable_tb_playlist_new),
			  self, 
			  NULL);


	{	
	  GtkTreeViewColumn   *column;
	  GtkCellRenderer     *cell;
	  gint		       n;

	  widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_playlists"); 
	  column = gtk_tree_view_column_new (); 
	  gtk_tree_view_column_set_title (column, _("Playlist Name")); 
	  g_object_connect (G_OBJECT(widget),
			    "signal::cursor-changed",
			    G_CALLBACK(on_tree_view_playlist_cursor_changed),
			    self,
			    NULL);

	  self->private->map_root_playlists = attr_row_map_new ();

	  cell = gtk_cell_renderer_text_new ();
	  gtk_tree_view_column_pack_start
		(column,
		 cell, 
		 FALSE);

	  gtk_tree_view_column_set_expand
		(column,
		 FALSE);

	  gtk_tree_view_column_set_attributes
		(column,
		 cell, 
		 "text",
		  0,
		  NULL);

	  gtk_tree_view_column_set_sort_column_id (column, 0);
	  gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
      
	  { 
	      GtkListStore    *liststore;
	      GtkTreeIter      iter;
	      GList	      *playlists;

	      liststore = gtk_list_store_new (1, G_TYPE_STRING);
	      playlists = pl_manager_get_lists (bmp_pl_manager);

	      for ( ; playlists ; playlists = playlists->next)
		{
		    gtk_list_store_append (liststore, &iter);
		    gtk_list_store_set (liststore, &iter, 0, playlists->data, -1);
		    attr_row_map_mapping_insert (self->private->map_root_playlists, playlists->data, GTK_TREE_MODEL(liststore), &iter);
		}
	      
	      gtk_tree_view_set_model (GTK_TREE_VIEW(widget), GTK_TREE_MODEL(liststore)); 
	  }

	  widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_tracks"); 
	  for (n = 0; n < BMP_DATUM_NONE; n++)
	    {

		column = gtk_tree_view_column_new (); 
		gtk_tree_view_column_set_title (column, gettext(bmp_metadata_get_title_static (n)));

		cell = gtk_cell_renderer_text_new ();
		gtk_tree_view_column_pack_start
		(column,
		 cell, 
		 FALSE);

		gtk_tree_view_column_set_expand
		(column,
		 FALSE);

		gtk_tree_view_column_set_cell_data_func
                                            (column,
                                             cell, 
                                             (GtkTreeCellDataFunc) relation_tuple_cell_data_func,
                                             GINT_TO_POINTER(n), 
                                             NULL);

		gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
	    }

	    gtk_tree_view_set_model (GTK_TREE_VIEW(widget), NULL); 
	}
#endif

	/* Setup modify tags dialog */
	widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_modify_tags");
	
	column_modify[0] =
	      gtk_tree_view_column_new (); 

	gtk_tree_view_column_set_title (column_modify[0], _("Tag Name"));

	renderer_modify[0] =
	      gtk_cell_renderer_toggle_new ();
	renderer_modify[1] =
	      gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column_modify[0],
		 renderer_modify[0],
		 FALSE);

	gtk_tree_view_column_pack_start
		(column_modify[0],
		 renderer_modify[1],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_modify[0],
		 FALSE);

	gtk_tree_view_column_set_attributes
		(column_modify[0],
		 renderer_modify[0],
		 "active",
		  0,
		  NULL);

	gtk_tree_view_column_set_attributes
		(column_modify[0],
		 renderer_modify[1],
		 "text",
		  1,
		  NULL);

	gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column_modify[0]);

	g_object_connect (G_OBJECT(renderer_modify[0]),
			  "signal::toggled",
			  G_CALLBACK(on_renderer_modify_toggled),
			  self, 
			  NULL);

	/* Common widgets */
	widget = glade_xml_get_widget_warn (self->private->xml, "button_tag");
	g_object_connect(G_OBJECT(widget),
			 "signal::clicked", 
			 G_CALLBACK(on_button_tag_clicked),
			 self, 
			 NULL);

	widget = glade_xml_get_widget_warn (self->private->xml, "button_enqueue");
	g_object_connect(G_OBJECT(widget),
			 "signal::clicked",
		         G_CALLBACK(on_button_enqueue_clicked),
			 self, 
			 NULL);

	widget = glade_xml_get_widget_warn (self->private->xml, "button_play");
	g_object_connect(G_OBJECT(widget),
			 "signal::clicked",
			 G_CALLBACK(on_button_play_clicked),
			 self, 
			 NULL);

	widget = glade_xml_get_widget_warn (self->private->xml, "entry_tagname");
	g_object_connect (G_OBJECT(widget),
			  "signal::activate",
			  G_CALLBACK(on_entry_tagname_activate),
			  self, 
			  NULL);
}

static gboolean
init_window (BmpLibrary *self)
{
    GtkWidget *widget;

    self->private->tree_store_browse = GTK_TREE_MODEL(gtk_tree_store_new (5, 
					    GDK_TYPE_PIXBUF, 
					    G_TYPE_STRING, 
					    H_TYPE_TUPLE, 
					    G_TYPE_POINTER,
					    G_TYPE_INT));

    gtk_tree_sortable_set_sort_func
                                            (GTK_TREE_SORTABLE (self->private->tree_store_browse),
					     1,
                                             (GtkTreeIterCompareFunc) tree_store_browse_sort_func,
                                             NULL, 
                                             NULL); 

    gtk_tree_sortable_set_sort_column_id 
					    (GTK_TREE_SORTABLE (self->private->tree_store_browse),
					     1, 
					     GTK_SORT_ASCENDING);


    self->private->tree_store_browse_filter = gtk_tree_model_filter_new (GTK_TREE_MODEL(self->private->tree_store_browse), NULL);
    self->private->map_root_browse = attr_row_map_new ();

    bmp_metadata_cache_load_rtfs (bmp_metadata_cache);
    bmp_metadata_cache_for_each_tuple (bmp_metadata_cache, (HTupleFunc)browse_view_insert_tuple_proxy_func, self);

    signal_handler_pool_add_connection (self->private->pool,
					bmp_metadata_cache,
					"insert-tuple",
					G_CALLBACK(browse_view_insert_tuple),
				        self,
					0);

    signal_handler_pool_add_connection (self->private->pool,
					bmp_metadata_cache,
					"delete-tuple",
					G_CALLBACK(browse_view_delete_tuple),
				        self,
					0);

    self->private->tree_store_tags = GTK_TREE_MODEL(gtk_tree_store_new (5, 
					  GDK_TYPE_PIXBUF,
					  G_TYPE_STRING, 
					  G_TYPE_POINTER,
					  G_TYPE_POINTER,
					  G_TYPE_INT)); 

    self->private->map_root_tags = attr_row_map_new ();

    signal_handler_pool_add_connection (self->private->pool,
					bmp_metadata_cache,
					"tag-added",
					G_CALLBACK(on_tag_added),
				        self,
					0);

    signal_handler_pool_add_connection (self->private->pool,
					bmp_metadata_cache,
					"tag-deleted",
					G_CALLBACK(on_tag_deleted),
				        self,
					0);

    signal_handler_pool_add_connection (self->private->pool,
					bmp_metadata_cache,
					"tag-value-insert",
					G_CALLBACK(on_tag_value_insert),
				        self,
					0);

    signal_handler_pool_add_connection (self->private->pool,
					bmp_metadata_cache,
					"tag-value-remove",
					G_CALLBACK(on_tag_value_remove),
				        self,
					0);

    widget = glade_xml_get_widget_warn (self->private->xml, "cbox_tag");
    gtk_combo_box_remove_text (GTK_COMBO_BOX(widget), 0); //Remove dummy
    gtk_combo_box_append_text (GTK_COMBO_BOX(widget), _("<not specified>")); 

    bmp_metadata_cache_load_tags (bmp_metadata_cache);

    /* Tags common setup */
    widget = glade_xml_get_widget_warn (self->private->xml, "cbox_tag");
    g_object_connect (G_OBJECT(widget), 
			 "swapped-signal::changed", 
			 G_CALLBACK(gtk_tree_model_filter_refilter), 
			 self->private->tree_store_browse_filter, 
			 NULL);

    // New tag
    widget = glade_xml_get_widget_warn (self->private->xml, "button_tag_new");
    g_object_connect (G_OBJECT(widget),
			 "signal::clicked",
			 G_CALLBACK(on_button_tag_new_clicked),
			 self, 
			 NULL);

    // Delete a tag
    widget = glade_xml_get_widget_warn (self->private->xml, "button_tag_delete");
    g_object_connect (G_OBJECT(widget),
			 "signal::clicked",
			 G_CALLBACK(on_button_tag_delete_clicked),
			 self, 
			 NULL);

    widget = glade_xml_get_widget_warn (self->private->xml, "cbox_tag");
    gtk_combo_box_set_active  (GTK_COMBO_BOX(widget), 0); 

    widget = glade_xml_get_widget_warn (self->private->xml, "entry_value");
    g_object_connect(G_OBJECT(widget), 
		     "swapped-signal::changed", 
		     G_CALLBACK(gtk_tree_model_filter_refilter), 
		     self->private->tree_store_browse_filter, 
		     NULL);

    widget = glade_xml_get_widget_warn (self->private->xml, "notebook_library_inner");
    gtk_notebook_set_current_page (GTK_NOTEBOOK (widget), 1);

    widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_browse");
    gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER(self->private->tree_store_browse_filter),
						(GtkTreeModelFilterVisibleFunc)tree_store_browse_visible_func,
						self,	
						NULL);

#ifdef HAVE_HAL
    g_object_connect(G_OBJECT(bmp_hal), 
		     "swapped-signal::volume-added", 
		     G_CALLBACK(gtk_tree_model_filter_refilter), 
		     self->private->tree_store_browse_filter, 
		     NULL);

    g_object_connect(G_OBJECT(bmp_hal), 
		     "swapped-signal::volume-removed", 
		     G_CALLBACK(gtk_tree_model_filter_refilter), 
		     self->private->tree_store_browse_filter, 
		     NULL);
#endif

    gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL(self->private->tree_store_browse_filter));
    gtk_widget_show_all (GTK_WIDGET (widget));

    widget = glade_xml_get_widget_warn (self->private->xml, "tree_view_tags");
    gtk_tree_view_set_model (GTK_TREE_VIEW(widget), GTK_TREE_MODEL(self->private->tree_store_tags));
    gtk_widget_show_all (GTK_WIDGET (widget));

    widget = glade_xml_get_widget_warn (self->private->xml, "library_notebook");
    gtk_widget_set_sensitive (widget, TRUE);

    self->private->shown = TRUE;
    self->private->running = FALSE;

    return FALSE;
}

void
bmp_library_show (BmpLibrary *self)
{
    if ((!self->private->shown) && (!self->private->running))
      {
	GtkWidget *widget;

	widget = glade_xml_get_widget_warn (self->private->xml, "library_notebook");
	gtk_widget_set_sensitive (widget, FALSE);

	self->private->running = TRUE;

	g_idle_add ((GSourceFunc) init_window, self);
      }

    gtk_window_present (GTK_WINDOW(self->window));
}

static void
ui_dialog_library_playlist_new_from_file (GtkAction	*action,
					  BmpLibrary    *self)
{
    GladeXML	  *xml = self->private->xml; 
    GtkWidget	  *chooser,
		  *parent;

    GtkFileFilter *file_filter;

    GList	  *containers = NULL;

    gint	   rv = 0;

    parent = glade_xml_get_widget_warn (xml, "dialog_library");
    chooser = gtk_file_chooser_dialog_new (_("Select playlist to load... - BMP"),
					   GTK_WINDOW(parent),
					   GTK_FILE_CHOOSER_ACTION_OPEN,
					   GTK_STOCK_OPEN,
					   GTK_RESPONSE_OK,
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_CANCEL,
					   NULL);

    gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER(chooser), FALSE);

    /* Set Path */
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(chooser), xcs_key_fetch (xcs, "bmp", "file-chooser-path"));

    containers = plugins_find_containers (); 

    for ( ; containers; containers=containers->next)
      {
	  ContainerHTuple *tuple = (ContainerHTuple*)containers->data;
	  gchar *glob, *glob_suffix;

	  glob = create_glob (tuple->suffix);
	  glob_suffix = g_strconcat ("*.", glob, NULL);
	  g_free (glob);

	  file_filter = gtk_file_filter_new();
	  gtk_file_filter_set_name (file_filter, tuple->desc);
	  gtk_file_filter_add_pattern(file_filter, glob_suffix); 
	  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	  g_free (glob_suffix);
      }


    rv = gtk_dialog_run (GTK_DIALOG(chooser));

    switch (rv)
      {
	  case GTK_RESPONSE_CANCEL:
	    {
		gtk_widget_destroy (chooser);
		return;
	    }

	  case GTK_RESPONSE_OK:
	    {
		GtkWidget    *widget = glade_xml_get_widget_warn (xml, "tree_view_tracks"); 
		GtkListStore *store;
		GtkTreeIter   iter;

		gchar  *uri,
		       *filename,
		       *basename;

		gchar **fragments;

		uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER(chooser));
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(chooser));
		basename = g_path_get_basename (filename);
		fragments = g_strsplit (basename, ".", -1);
		g_free (basename);

		//FIXME: PLManager should make use of GError
		if (!pl_manager_list_new_from_uri (bmp_pl_manager,
						   fragments[0], 
						   uri)) 
		  {
		    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Unable to create playlist from URI '%s'", uri);
		  }

		store = pl_manager_list_get_model (bmp_pl_manager, 
						   fragments[0]);

		g_free (uri);
		g_free (filename);
		gtk_widget_destroy (GTK_WIDGET(chooser));

		widget = glade_xml_get_widget_warn (xml, "tree_view_playlists"); 
		gtk_list_store_append (GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(widget))), &iter);
		gtk_list_store_set (GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(widget))), &iter, 0, fragments[0], -1);
		attr_row_map_mapping_insert (self->private->map_root_playlists, fragments[0], gtk_tree_view_get_model(GTK_TREE_VIEW(widget)), &iter);
	
		g_strfreev (fragments);	

		return;
	    }

	  default: break;
      }

    gtk_widget_destroy (GTK_WIDGET(chooser));
}
